
(* Equivalence decision procedure based on the 
   Partial Derivatives for Kat Expressions *)

open Batteries_uni              (* Non-threaded version *)
open Set

open Module_KatExpressions      (* a module for kat exps *)
open Module_LinearKatExps       (* a module for kat exps in the linear form *)
open Module_PairsKatExps        (* a module for pairs of kat exps *)
open Module_Utils
open Lexer
open LexerPolish
open Parser
open ParserPolish

exception ERROR of string

let doTest = ref true           (* This option turned false allows to simulate that the expressions *)
                                (* since the E test will be ignored *)
let opt2a = ref true            (* UP2Congr first version *)
let opt2b = ref false           (* UP2Congr second version - compares the two nf concurrently *)
let debugging = ref false

let opts = 
  [ 
    (opt2a, true, "UP2Congr first version");
    (opt2b, false, "UP2Congr second version - compares the two nf concurrently")
  ]

(* Function Aliases *)
let lsRemoveRepet = MUtils.lsRemoveRepetitions

let rec fe1 at t' e1 exp =
  match exp with
      MKE.Prog p           ->  
	MLinearKE.setLinearKEfromList [? List: (atom,p,MKE.(Bexp One)) | atom <- List: at; MKE.e atom t' e1 ?]
    | MKE.Bexp b           ->  SetLinearKE.empty
    | MKE.Star exp'        ->  MLinearKE.map (MLinearKE.concat3rd exp) (fe1 at t' e1 exp')
    | MKE.Conc (exp1::ls)  ->  let exp2 = MKE.buildConcat ls
			       in let set1 = MLinearKE.map (MLinearKE.concat3rd exp2) (fe1 at t' e1 exp1)
				  in let set2 = fe1 at t' (MKE.Conc [exp1;e1]) exp2
				     in SetLinearKE.union set1 set2
    | MKE.Conc []          ->  raise (ERROR "function fe1 received an unexpected pattern: Conc []")
    | MKE.Or set           ->  MLinearKE.unions (List.map (fe1 at t' e1) (SetKE.elements set))

let rec f at t' exp =
  match exp with
      MKE.Prog p           ->  
	MLinearKE.setLinearKEfromList [? List: (atom,p,MKE.(Bexp One)) | atom <- List : at ?]
    | MKE.Bexp b           ->  SetLinearKE.empty
    | MKE.Star exp'        ->  MLinearKE.map (MLinearKE.concat3rd exp) (f at t' exp')
    | MKE.Conc (exp1::ls)  ->  let exp2 = MKE.buildConcat ls
			       in let set1 = MLinearKE.map (MLinearKE.concat3rd exp2) (f at t' exp1)
				  in let set2 = fe1 at t' exp1 exp2
				     in SetLinearKE.union set1 set2
    | MKE.Conc []          ->  raise (ERROR "function f received an unexpected pattern: Conc []")
    | MKE.Or set           ->  MLinearKE.unions (List.map (f at t') (SetKE.elements set))
      
let gAux atom prog ls = [? List: exp | (atom',prog',exp) <- List : ls; atom=atom'; prog=prog' ?] 
let g at t' exp = 
  let exp_lin = SetLinearKE.elements (f at t' exp) in
  MLinearKE.setLinearKEfromList 
    [? List : (atom,prog,MKE.buildOr (gAux atom prog exp_lin)) | (atom,prog,_) <- List: exp_lin ?] 
    
let rec der atom prog lst =
  match lst with
      (atom',prog',e)::ls  ->  if atom'=atom & prog'=prog then e else der atom prog ls
    | []                   ->  MKE.(Bexp Zero)
      
let derivatives lf_set1 lf_set2 = 
  let ls_exp1 = SetLinearKE.elements lf_set1 in
  let ls_exp2 = SetLinearKE.elements lf_set2 in
  let heads_exp1 = [? List : (atom,prog) | (atom,prog,_) <- List : ls_exp1 ?] in
  let heads_disj = heads_exp1 @ ([? List : (atom,prog) | (atom,prog,_) <- List : ls_exp2 ; 
				  not(List.mem (atom,prog) heads_exp1) ?]) in
  [? List: (atom, prog, der atom prog ls_exp1, der atom prog ls_exp2) | (atom,prog) <- List: heads_disj ?]
    
let rec inCongrb e1 e2 rel = 
  let form1 = MKE.NonFinal (SetKE.empty, [e1]) in
  let form2 = MKE.NonFinal (SetKE.empty, [e2]) in
  foo1 (expand form1 rel) (expand form2 rel) rel
and foo1 form1 form2 rel = 
  match (form1,form2) with
    (MKE.Final nf1, MKE.Final nf2)                  ->  SetKE.equal nf1 nf2
  | (MKE.NonFinal (s1,ls1), MKE.Final nf2)          ->  
    if (SetKE.equal nf2 s1) then true else foo1 (expand form1 rel) form2 rel
  | (MKE.Final nf1, MKE.NonFinal (s2,ls2))          ->  
    if (SetKE.equal nf1 s2) then true else foo1 form1 (expand form2 rel) rel
  | (MKE.NonFinal (s1,lsx), MKE.NonFinal (s2,lsy))  ->  
    if (s1=s2) then true else foo1 (expand form1 rel) (expand form2 rel) rel
and expand form rel = 
  match form with
    (MKE.NonFinal (s,h::tl))  ->
      let newSet = SetKE.add h s in
      let newList = 
	match h with
	  | (MKE.Or set) ->  
	    let lst = [? List: x | h1 <- List: (h::(SetKE.fold ff set []));   (* ie, h1 is a subexp of h *)
		       not(MKE.keEq x h); 
		       x <- List: buildLst h1 newSet rel ?] in    (* ie, x is related to h1 in rel *)
	    lsRemoveRepet (tl@lst)
	  | otherwise    ->  let lst = buildLst h newSet rel in 
			     lsRemoveRepet (tl@lst)
      in if newList=[] then (MKE.Final newSet) else (MKE.NonFinal (newSet,newList))
    | _                       ->  raise (ERROR "expand received something other than NonFinal (s,h::tl)")
and buildLst h newSet rel =    
  (* Returns a list of all x s.t. x is related to h in rel and is not in the nf already *)
  List.append 
    [? List: x | (x,h') <- List: (SetPairsKE.elements rel); h'=h; 
     not((SetKE.mem x newSet) || MKE.keEq x MKE.(Bexp Zero)) ?] 
    [? List: x | (h',x) <- List: (SetPairsKE.elements rel); h'=h; 
     not((SetKE.mem x newSet) || MKE.keEq x MKE.(Bexp Zero))?]
and ff x ls = 
  match x with
    (MKE.Or set)  ->  SetKE.fold ff set ls 
  | otherwise     ->  (x::ls)

let rec inCongra e1 e2 rel = 
  let (exp,ls) as pair' = expand (SetKE.empty, [e1]) rel in
  if (ls=[]) then false    (* this means e1 has no correspondence in R *)
  else match (bar e2 pair' rel) with
    None           ->  true
  | Some (s1,ls1)  ->  let (s2,ls2) = expand_completely (SetKE.empty,[e2]) rel in
		       equal s1 s2
and bar e2 pair rel =
  match pair with
      (_,[])     ->  Some pair           (* means they are not equiv and _ is the complete nf of e1 *)
    | otherwise  ->  let pair' = expand pair rel in
		     if member e2 pair' then None      (* means they are equiv *)
		     else bar e2 pair' rel
and expand (s,ls) rel = 
  if ls=[] then (s,[]) else
    let h = List.hd ls in
    let tl = List.tl ls in
    match (h,tl) with
      | ((MKE.Or set),_) ->  let newSet = SetKE.add h s in
			     let lst = 
			       [? List: x | h1 <- List: (h::(SetKE.fold ff set [])) (* ie,h1 is a subexp of h *)
			       ; not(MKE.keEq x h) 
			       ; x <- List: buildLst h1 newSet rel ?] in  (* ie, x is related to h1 in rel *)
			     let newList = lsRemoveRepet (tl@lst) in
			     (newSet, newList)
      | (h,tl)           ->  let newSet = SetKE.add h s in
			     let lst = buildLst h newSet rel in
			     let newList = lsRemoveRepet (tl@lst) in
			     (newSet, newList)
and buildLst h newSet rel =    
  (* returns a list of all x s.t. x is related to h in rel and is not in the nf already *)
  List.append 
    [? List : x | (x,h') <- List : (SetPairsKE.elements rel) ; h'=h ; 
     not((SetKE.mem x newSet) || MKE.keEq x MKE.(Bexp Zero)) ?] 
    [? List : x | (h',x) <- List : (SetPairsKE.elements rel) ; h'=h ; 
     not((SetKE.mem x newSet) || MKE.keEq x MKE.(Bexp Zero))?]
and ff x ls = 
  match x with
    (MKE.Or set)   ->   SetKE.fold ff set ls 
  | otherwise      ->   (x::ls)
and expand_completely pair rel = 
  let (e',ls') as pair' = expand pair rel in 
  if ls'=[] then pair' 
  else expand_completely pair' rel
and member e2 (nf_set, nf_lst) =
  match nf_lst with
    (h::tl)  ->  
      if MKE.keEq e2 (MKE.buildOr (SetKE.elements (SetKE.add h nf_set))) then true else member e2 (nf_set,tl)
  | []       ->  false 
and equal nf1 nf2 = MKE.keEq (MKE.buildOr (SetKE.elements nf1)) (MKE.buildOr (SetKE.elements nf2))

let rec equivP set h = 
  if (SetPairsKE.is_empty set) then (true, h)
  else
    let (exp1,exp2) = (SetPairsKE.min_elt set) in
    let (at2,t'2) = MKE.getAtAndTp exp1 (if !doTest then exp2 else exp1) in
    if (MKE.eAll ~at:at2 ~t':t'2 exp1 (if !doTest then exp2 else exp1))=false then (false, h)
    else 
      let h' = SetPairsKE.add (exp1,exp2) h in
      let s' = 
	if !opt2a then
	  let sets = (SetPairsKE.union h' (SetPairsKE.remove (exp1,exp2) set)) in
	  MPairsKE.fromList 
	    [? List : (der1,der2) | (_,_,der1,der2) <- List:(derivatives (g at2 t'2 exp1) (g at2 t'2 exp2)); 
	     not(SetPairsKE.mem (der1,der2) sets) ; not(inCongra der1 der2 sets) ?]
	else if !opt2b then
	  let sets = (SetPairsKE.union h' (SetPairsKE.remove (exp1,exp2) set)) in
	  MPairsKE.fromList 
	    [? List : (der1,der2) | (_,_,der1,der2) <- List:(derivatives (g at2 t'2 exp1) (g at2 t'2 exp2)); 
	     not(SetPairsKE.mem (der1,der2) sets) ; not(inCongrb der1 der2 sets) ?]
	else
	  MPairsKE.fromList 
	    [? List : (der1,der2) | (_,_,der1,der2) <- List:(derivatives (g at2 t'2 exp1) (g at2 t'2 exp2)); 
	     (SetPairsKE.mem (der1,der2) h')==false ?] in
      let s' = SetPairsKE.union (SetPairsKE.remove (exp1,exp2) set) s' in
      equivP s' h'

let parse str =
  let linebuf = Lexing.from_string (str ^ "\nEOF") in
  List.hd (Parser.data Lexer.token linebuf)

let parsePolish str =
  let linebuf = Lexing.from_string (str ^ "\nEOF") in
  List.hd (ParserPolish.data LexerPolish.token linebuf)

let main exp1 exp2 = 
  let exp1_flatten = MKE.flattenConcats exp1 in
  let exp2_flatten = MKE.flattenConcats exp2 in
  let (ans,h) = equivP (SetPairsKE.singleton (exp1_flatten,exp2_flatten)) SetPairsKE.empty in
  (ans,h)
