 
(* This file presents an algorithm incl that attempts to build a simulation
   between two Non-Deterministic Finite Automata on Guarded Strings. These algorithms
   deal with AGS of type (Q, <o#,t#>), with 
     o#: P(Q) ---> B
     t#: P(Q) ---> (P(Q))^{BxΣ},
   where B is the free Boolean algebra.  *)

open Batteries_uni                (* Non-threaded version *)
open Set

open Module_KatExpressions        (* a module for kat exps *)
open Module_KatExpressionsPos
open Module_PairsActionBexp       (* a module for pairs of (kat exp,bexp) *)
open Module_PairsPairs
open Module_NFAGS
open Module_SetBuilder
open Module_Up2CongruenceSearch
open Module_Utils
open Lexer
open LexerPolish
open Parser
open ParserPolish
open RegstabBridge

exception ERROR of string

type state=int

type intSet = SetInt.t
module Up2Congr = Up2Congr (T)
module SetPairsIntSet = Up2Congr.R
type rel = SetPairsIntSet.t 
let inCongr x y rel = Up2Congr.inCongr x y rel

(* pointers *)
(* Transitions: each transition is a quad of the form 
   ((departingState,symbol),(destinationState,be)) *)
let transX = ref [((0,0),(0,MKE.Zero))]
let transY = ref !transX

(* Outputs: each output is a pair of the form (state,be) *)
let outputsX = ref [0,MKE.Zero]
let outputsY = ref !outputsX

let debugging = ref false

(* running options from the command-line *)
let doTest = ref true           (* This option turned false allows to simulate that the expressions *)
                                (* since the outputs test will be ignored *)
let optRSubset = ref true
let opts =    (* running options *)
  [
    (optRSubset,   true,  "RSubset - when testing (X,Y) always checks if some (X',Y'), with Y' a subset of Y and X a subset of X', is in R already")
  ]

(* Type Aliases *)
type atom = MKE.atom
type automaton = NFAGS.automatonBSKE
type be = MKE.be
type formula = RegstabBridge.formula
type ke = MKE.ke
type witness = RegstabBridge.witness
type stateSet = intSet

(* Function Aliases *)
let areEquiv = MKE.areEquiv
let top = RegstabBridge.top
let bottomSt = NFAGS.bottomSt
let bottomStSet = NFAGS.bottomStSet
let buildAnd = RegstabBridge.buildAnd
let buildAnds = RegstabBridge.buildAnds
let buildConj = MKE.buildConj
let buildConc = MKE.buildConcat
let buildConcPos = MKEPos.buildConcat
let buildDisj = MKE.buildDisj
let buildImplic = RegstabBridge.buildImplic
let buildOpposite = RegstabBridge.buildOpposite
let buildOr = MKE.buildOr
let eBexp = MKE.eBexp
let evaluate = MKE.evaluate
let findAWitness = RegstabBridge.findAWitness
let fromAtom2String = MKE.atom2String
let fromBexp2Formula = MKE.bexp2RegstabFormula
let fromBexp2String = MKE.bexp2String
let fromFormula2String = RegstabBridge.fromFormula2String
let fromIntList2IntSet = MInt.setIntFromList 
let getAtAndTpFromBexpList = MKE.getAtAndTpFromBexpList
let implies = MKE.implies
let isATautology = RegstabBridge.isATautology
let isSat = RegstabBridge.isSat
let emptyR = SetPairsIntSet.empty
let addToR = SetPairsIntSet.add 
let lsDiff = MUtils.lsDiff
let lsRemoveRepet = MUtils.lsRemoveRepetitions
let unionR = SetPairsIntSet.union 
let memR = MPairsIntSet.mem 
let print str = print_endline str ; flush stdout
let print_str str = print_string str ; flush stdout
let printSetInt = MInt.printSetInt
let setInt2String = MInt.setInt2String

(* Helpful Debugging Printers *)
let printIntList ls = 
  print_str "["; (List.iter (fun i -> print_str ((string_of_int i)^",")) ls); print_str "]"
let printTransByP (ls:(be*(int list)) list) =
  print_str "["; List.iter (fun (b,l) -> MKE.printBexp_nnl b; print_str " "; printIntList l) ls; print_str "]"
let printSetsSts setSt1 setSt2 = 
  print_str " "; printSetInt setSt1; print_str " and "; printSetInt setSt2 ; flush stdout
let printOutputs ls = 
  List.iter (fun (state,be) -> print_str (string_of_int state) ; MKE.printBexp_nnl be) ls ; flush stdout
let printTransitionQuads trans = 
  List.iter (fun ((x,y),(z,w)) -> print_str ("from " ^ string_of_int x ^ " by "); 
    print_str (string_of_int y ^ ","); print_str (" to " ^ string_of_int z ^ " if "); 
    MKE.printBexp_nnl w) trans; flush stdout
let debugAutomata (outputs,trans,i) = 
  (print ("Automaton" ^ string_of_int i ^ ": \noutputs:") ; (printOutputs outputs) ; 
   print ("\ntransitions:") ; printTransitionQuads trans)
let debugCurrentStates (x,y) = 
  print_str ("\nI'm looking at the following sets of states: ") ; printSetsSts x y; 
  print_str "\n"
let debugStatesAlreadyInR (x,y) = 
  print_str ("the following pair is already in R: ") ; printSetsSts x y ; 
  print_str "\n"
let debugOutputsImplies (x,y,answer) = 
  print_str "I'm going to test the implies of the outputs of the states "; printSetsSts x y; 
  print_str " using outputs1=" ; printOutputs (!outputsX) ; 
  print_str " and outputsy=" ; printOutputs (!outputsY) ;
  print_str " and the answer was " ; print (string_of_bool answer)
let printTransAllSymb trans1 trans2 = printTransitionQuads trans1; print_str " and "; printTransitionQuads trans2
let printTransitionPair (statef,be) = 
  print_str ("(" ^ (string_of_int statef) ^ ","); MKE.printBexp_nnl be; print_string ") "
let printTransitionPairs ls = if ls=[] then print_string "[]" else List.iter printTransitionPair ls
let printTransp trans1 trans2 = printTransitionPairs trans1; print_string " and "; printTransitionPairs trans2
let printCombination (x,y,z,a,b,c) = 
  MUtils.lsIntPrint x; print_str "statesf:{"; printSetInt y; print_str "}"; print_str " if "; 
  MKE.printBexp_nnl z; MUtils.lsIntPrint a; print_str "statesf:{"; printSetInt b; print_str "}"; 
  print_str " if "; MKE.printBexp_nnl c; print_str "\n"
let printCombinations ls = List.iter printCombination ls
let printBool_nl bool = print (string_of_bool bool)
let debugFooEntryPoint (x,transX,y,r) =
  print_str "I've entered foo x="; printSetInt x ; 
  print_str " transX=" ; printTransitionQuads transX ;
  print_str " y=" ; printSetInt y ; 
  print_str " r=" ; MPairsIntSet.printSet r; 
  print_str "\n"
let debugBarEntryPoint (p,destStX,beX,(atBeX,t'),y,r) = 
  print_str ("I've entered bar p=" ^ string_of_int p) ;
  print_str (" destStX=" ^ string_of_int destStX) ;
  print_str " beX=" ; MKE.printBexp_nnl beX ;
  print_str " atBeX=" ; MKE.debugAt atBeX ; 
  print_str " t'=" ; MKE.debugT' t' ;
  print_str " y=" ; printSetInt y ;
  print_str "r=" ; MPairsIntSet.printSet r ; 
  print_str "\n"
let debugBarWithSATEntryPoint (p,destStX,formX,y,r) = 
  print_str ("I've entered barSAT p=" ^ string_of_int p) ;
  print_str (" destStX=" ^ string_of_int destStX) ;
  print_str " y=" ; printSetInt y ;
  print_str "r=" ; MPairsIntSet.printSet r ; print_string "\n"
let debugEquivFooEntryPoint (x,transX,y,transY,r) =
  print_str "I've entered foo x="; printSetInt x ; 
  print_str " transX=" ; printTransitionQuads transX ;
  print_str " y=" ; printSetInt y ; 
  print_str " transY=" ; printTransitionQuads transY ;
  print_str " r=" ; MPairsIntSet.printSet r; 
  print_str "\n"
let debugEquivBarEntryPoint (p,destStX,beX,(atBeX,t'),y,r) = 
  print_str ("I've entered bar p=" ^ string_of_int p) ;
  print_str (" destStX=" ^ string_of_int destStX) ;
  print_str " beX=" ; MKE.printBexp_nnl beX ;
  print_str " atBeX=" ; MKE.debugAt atBeX ; 
  print_str " t'=" ; MKE.debugT' t' ;
  print_str " y=" ; printSetInt y ;
  print_str "r=" ; MPairsIntSet.printSet r ; 
  print_str "\n"
let debugEquivBarWithSATEntryPoint (p,destStX,formX,y,r) = 
  print_str ("I've entered barSAT p=" ^ string_of_int p) ;
  print_str (" destStX=" ^ string_of_int destStX) ;
  print_str " y=" ; printSetInt y ;
  print_str "r=" ; MPairsIntSet.printSet r ; print_string "\n"

(* -------------- Auxiliary functions for the decision tests ------------ *)

(* when testing (X,Y), checks if some (X',Y'), with Y' a subset of Y and X a subset of X', exists in R *)
let memR1stIncl2ndCont (x,y) r = 
  SetPairsIntSet.exists (fun (a,b) -> SetInt.subset x a && SetInt.subset b y) r

let memR1stCont2ndIncl (x,y) r = 
  SetPairsIntSet.exists (fun (a,b) -> SetInt.subset a x && SetInt.subset y b) r  (* .equal -> .subset *)

let fromStList2StSet (ls:state list) : SetInt.t = 
  if ls=[] then bottomStSet else fromIntList2IntSet ls

let oCardinal setSt outputs = 
  buildDisj (List.map (
    fun x -> (
      try List.assoc x outputs 
      with Not_found -> raise (ERROR ("The element " ^ (string_of_int x) ^ " wasn't found in the outputs")))
  ) ((SetInt.elements setSt)))

(* given a set of states and a list of transitions, the function returns 
the disjunction of all the bes in the transitions by p from states in that set *)
let getAllBesTrans statesSet trans p = 
  let ls = [? List: be | ((depSt,q),(_,be)) <- List: trans ; SetInt.mem depSt statesSet ; p=q ?] in
  if ls=[] then MKE.Zero else buildDisj ls

(* same as List.assoc but returns the contatenation of all List.assoc k1 ls,
   List.assoc k2 ls, ..., List.assoc kn ls, where {k1,...,kn}=keySet *)
let listAssoc (keySet:intSet) (ls:(int*(int*((be*(int list)) list)) list)list) = 
  List.fold_left (fun a b -> a@(try List.assoc b ls with Not_found -> [])) [] (SetInt.elements keySet)

let rec listAssoc2 (p:int) (transFromZ:(int*((be*(int list)) list)) list) = 
  let fst = try List.assoc p transFromZ with Not_found -> [] in
  if fst = [] then
    fst
  else 
    fst@(listAssoc2 p (List.remove_assoc p transFromZ))
    
(* Decision Test (Lang. Inclusion) *)

(* x and y - the sets of states to be checked for lang. equality
   h - keeps track of the pairs which didn't fail the o() test
   r - the simulation (it only isn't equal to h if it's an up-to simulation *)
let rec incl (x:SetInt.t) (y:SetInt.t) (h:rel) (r:rel) : (bool*rel*rel) = 
  let () = if !debugging then (print_str "\nI'm in incl with x="; printSetInt x; print_str " y="; 
			       printSetInt y; print_str "r="; MPairsIntSet.printSet r) in
  if x=y then     
    (true,h,r)
  else if memR (x,y) h then 
    (true,h,r)
  else if !optRSubset && memR1stIncl2ndCont (x,y) h then
    let () = if !debugging then print "I was going to test a pair (X,Y) but Y contains Y' and I know (X,Y') gave True already" in
    (true,addToR (x,y) h,r)
  else
    let () = if !debugging then debugCurrentStates (x,y) in
    let answer = implies (oCardinal x !outputsX)
      (if !doTest then (oCardinal y !outputsY) else (oCardinal x !outputsX)) in
    let () = if !debugging then
	debugOutputsImplies (x,y,answer) in
    if not(answer) then
      let () = if !debugging then print ("The answer to the sets of states " ^ setInt2String x ^ 
					    " and " ^ setInt2String y ^ " is " ^ string_of_bool answer)in
      (false,h,r)
    else 
      let (answer,h,r) = browse x y (addToR (x,y) h) (addToR (x,y) r) in
      let () = if !debugging then print ("The answer to the sets of states " ^ setInt2String x ^ 
					    " and " ^ setInt2String y ^ " is " ^ string_of_bool answer)in
      (answer,h,r)
and browse (x:intSet) (y:intSet) (h:rel) (r:rel) = 
  let transFromX =      (* All trans. departing from x *)
    lsRemoveRepet [? List: (p,(x',be)) | ((depSt,p),(x',be)) <- List: !transX; SetInt.mem depSt x ?] in
  let transFromY =      (* All trans. departing from y *)
    lsRemoveRepet [? List: (p,(y',be)) | ((depSt,p),(y',be)) <- List: !transY; SetInt.mem depSt y ?] in
  let psInX = lsRemoveRepet (fst (List.split transFromX)) in   (* a list with the p's occuring in transFromX *)
  checkAllPs psInX transFromX transFromY h r
and checkAllPs (ps:int list) (transFromX:(int*(int*be)) list) (transFromY:(int*(int*be)) list) (h:rel) (r:rel) =
  match ps with
      []     ->  (true,h,r)
    | p::tl  ->
      let transFromXByP =     (* filters in transFromX only the trans. by p *)
	lsRemoveRepet [? List: (destSt,be) | (q,(destSt,be)) <- List: transFromX; q=p ?] in
      let transFromYByP =     (* filters in transFromY only the trans. by p *)
	lsRemoveRepet [? List: (destSt,be) | (q,(destSt,be)) <- List: transFromY; q=p ?] in
      let (at,t') = getAtAndTpFromBexpList (snd (List.split (transFromXByP@transFromYByP))) in
      let (answer',h',r') = checkAllAtoms at t' transFromXByP transFromYByP h r in
      if answer' then 
	checkAllPs tl transFromX transFromY h' r'  
      else (false,h,r)
and checkAllAtoms (at:atom list) (t':int list) (transFromXByP:(int*be) list) (transFromYByP:(int*be) list) 
                  (h:rel) (r:rel) =
  match at with
      []        ->  (true,h,r)
    | atom::tl  ->  
      let destStX = [? List: x' | (x',be) <- List: transFromXByP; evaluate atom t' be ?] in
      let destStY = [? List: y' | (y',be) <- List: transFromYByP; evaluate atom t' be ?] in
      let x' = fromStList2StSet destStX in
      let y' = fromStList2StSet destStY in 
      let (answer',h',r') = incl x' y' h r in
      if answer' then
	checkAllAtoms tl t' transFromXByP transFromYByP h' r'
      else
	(false,h,r)

let callIncl (x,outputs1,trans1) (y,outputs2,trans2) = 
  outputsX := outputs1 ; transX := trans1 ;
  outputsY := outputs2 ; transY := trans2 ;
  let (answer,_,r) = incl (SetInt.singleton x) (SetInt.singleton y) emptyR emptyR in
  (answer,r)

let main (automaton1:automaton) (automaton2:automaton) = 
  callIncl automaton1 automaton2
