
open Scanf
open Batteries_uni   (* Non-threaded version *)
open Marshal

open Module_KatExpressions
open Module_KatExpressionsPos
open Module_KatExpressionsPos2
open Module_PairsKatExps
open Module_NFAGS
open Module_RewriteRules
open Module_Utils
open Module_SetBuilder
open Module_String2FloatMap
open Module_CallNFAGSEquivChecker
open RegstabBridge

exception ERROR of string

(* Global Variables Declaration *)
let verbose = ref false
let optBSKE = ref false      (* To use the Berry-Sethi construction applied to KAT expressions. *)
let optDontSL = ref false
let optDisj = ref false
let optSelf = ref false
let inclV1 = ref false
let inclV2 = ref true

(* Keywords Declaration *)
let keyw_naiv = "-naive"
let keyw_noup = "-noUpTo"
let keyw_lain = "-linclusion"
let keyw_laeq = "-lequality"
let keyw_nosat = "-noSATS"
let keyw_prdr = "-parDer"
let keyw_audi = "-NFAGSDirect"
let keyw_asmp1 = "-useAssumptions1"
let keyw_asmp2 = "-useAssumptions2"

(* Type Aliases *)
type ke = MKE.ke
type kePos = MKEPos.ke
type timeMap = MString2FloatMap.map

(* Function Aliases *)
let sumTimeMaps = MString2FloatMap.sum
let buildPairNFAGSUsingBS = NFAGS.buildAutomataUsingBS
let buildPairNFAGSUsingBSKE = NFAGS.buildAutomataUsingBSKE
let buildSigfromKe = MKE.buildSigmaFromKe
let buildTfromKe = MKE.buildTfromKe
let countTestsInKe = MKE.countTestsInKe
let countTestsInKePos = MKEPos.countTestsInKe
let divideTimeMapByN = MString2FloatMap.divideByN
let incrementTimeMap = MString2FloatMap.increment
let lsDiff2 = MUtils.lsDiff2
let newTimeMap () = MString2FloatMap.empty
let print str = print_endline str ; flush stdout
let print_str str = print_string str ; flush stdout
let printKE = MKE.printKE
let printKE_el = MKE.printKE_el
let printKEPos = MKEPos.printKE
let printKEPos_el = MKEPos.printKE_el
let printTimeMap = MString2FloatMap.printMap_nl
let shuffleList = MUtils.shuffleList
let subArray = MUtils.subArray
let tryListHd = MUtils.tryListHd
  
let intro_text = 
  "\n \t\t  Welcome to Kat Decider
\nThe available symbols for building Kat expressions are:
\t   'p'i      names of primitive actions, where i is the index and can be any integer
\t   't'i      names of primitive tests, where i is the index and can be any integer
\t    '+'      alternation
\t'not'bexp    negation, where bexp can be any boolean expression
\t    '*'      Kleene star operator
\t   '(' ')'   parentheses 
\t   'One'         
\t   'Zero'        \n"

let assumptions_text = "The assumptions mode allows you to test the equivalence of two KAT expressions in the presence of a set of assumptions. These assumptions may represent annotated Hoare triples, commutativity conditios or any other assumption of the form b'p'i = b'p'ic or b'p'ic = b'p'i, where b and c are two boolean expressions and i is an integer. Both b and c can be omitted in the case they represent One (Truth).

The input must be provided in three different lines, separated by Enters. The first line must be the sequence of assumptions separated by &. The two following lines must contain the two expressions to be tested.\n
The following is an example with the Composition rule of the Hoare Logic:
       \t t1p1=t1p1t2 & t2p2=t2p2t3
       \t t1p1t2p2
       \t t1p1t2p2t3
Answer: True "

let assumptions_text_2 = "This mode allows you to test the equivalence of two KAT expressions in the presence of a set of assumptions, using two distinct methods: derivatives on the assumptions (identical to the option \"assumptions\") and including the assumptions in the expressions. The cardinality of H will be printed in both cases for comparison purposes. 

The assumptions you provide may represent annotated Hoare triples, commutativity conditios or any other assumption of the form b'p'i = b'p'ic or b'p'ic = b'p'i, where b and c are two boolean expressions and i is an integer. Both b and c can be omitted in the case they represent One (Truth).

The input must be provided in three different lines, separated by Enters. The first line must be the sequence of assumptions separated by &. The two following lines must contain the two expressions to be tested.\n
The following is an example with the Composition rule of the Hoare Logic:
       \t t1p1=t1p1t2 & t2p2=t2p2t3
       \t t1p1t2p2
       \t t1p1t2p2t3
Answer: True "

let rules_text = "This mode accepts an expression exp from input and applies the decision procedure to exp and exp', where exp' is the result of applying some rewrite rules on exp.\n"
  
type description = string
type nfags = NFAGS.automaton
type exp =
    Exp of MKE.ke
  | ExpPos of MKEPos.ke
  | ExpPos2 of MKEPos2.ke
  | Str of string  (* an exp as a string *)

type model = 
    Exp of MKE.ke
  | ExpPos of MKEPos.ke
  | ExpPos2 of MKEPos2.ke
  | Str of string
  | NFAGS of nfags

type resultsAccum =           (* accumulates all pairs of exps that gave True or the number of such cases *)
    List of ((model*model) list)
  | Number of int

type setH = 
    KeSet of SetPairsKE.t     (* set of pairs of Kat exps (used by the partial deriv algorithm) *)
  | StSet of SetPairsIntSet.t   (* set of pairs of sets of states of the automata *)
  | StSetPair of SetPairsIntSet.t*SetPairsIntSet.t

let keyw_time_NFAGSConstr = "NFAGS Construction"
let keyw_time_DecisionProc = "Decision Procedure"

(*------------------- Auxiliary functions -------------------*)
let resetAllOptsToDefault () = 
  let () = 
    verbose:=true; optDisj:=false; optSelf:=false; Module_KatExpressions.optSAT:=true;
    inclV1:=false; inclV2:=true; optBSKE:=true in

  (* Where are the opts? *)
  let opts = 
    EquivalenceCheckerAGS.opts @ EquivalenceCheckerPD.opts @ Module_KatExpressions.opts 
    @ InclusionCheckerAGS1.opts @ InclusionCheckerAGS2.opts @ CallNFAGSEquivChecker.opts in

  let opts_local = ref opts in
  while not(!opts_local=[]) do
    let (opt,default,_) = List.hd !opts_local in
    let () = opt:=default in
    let () = opts_local := List.tl !opts_local in
    ()
  done

let printOptsEnabled () = 
  let () = print_str "The following options (if any) are being used: " in
  let () = if !Module_KatExpressions.optSAT=false then print_str "No SAT Solver; " in
  let () = if !inclV1=true then 
      let () = if !InclusionCheckerAGS1.optRSubset=false then print_str "No Up-To Technique; " in 
      print_str "Naive approach"; in
  let () = if !inclV2=true then 
      if !InclusionCheckerAGS2.optRSubset=false then print_str "No Up-To Technique; " in
  let () = 
    if !optBSKE=false then (if !EquivalenceCheckerPD.opt2a=false then print_str "No Up-To Technique; ";
			    print_str "Partial Derivatives Method; ")
    else if !CallNFAGSEquivChecker.optIncl1=false && !CallNFAGSEquivChecker.optIncl2=false then
      print_str "Lang. equality directly on the NFAGS"
  in
  print_str "\n"

let printExp exp = 
  match exp with 
    Exp e    ->  printKE e
  | ExpPos e ->  printKEPos e
  | Str e    ->  printKE (MUtils.parsePolish e)
  | _        ->  raise (ERROR "Unexpected pattern in main.printExp")

let printExp_el exp = 
  match exp with 
    Exp e    ->  printKE_el e
  | ExpPos e ->  printKEPos_el e
  | Str e    ->  printKE_el (MUtils.parsePolish e)
  | _        ->  raise (ERROR "Unexpected pattern in main.printExp_el")

let enableDebugging () = 
  EquivalenceCheckerAGS.debugging := true ;
  EquivalenceCheckerPD.debugging := true ;
  Module_KatExpressions.debugging := true ; RegstabBridge.debugging := true

let setHCardinal h =
  match h with
    KeSet h'          ->  SetPairsKE.cardinal h'
  | StSet h'          ->  SetPairsIntSet.cardinal h'
  | StSetPair (h1,h2) -> (SetPairsIntSet.cardinal h1) + (SetPairsIntSet.cardinal h2)

let calculateAtSize exp1 exp2 = 
  match (exp1,exp2) with 
      (Exp e1, Exp e2)       ->  MKE.sizeAt e1 e2
    | (ExpPos e1, ExpPos e2) ->  MKEPos.sizeAt e1 e2
    | _                      ->  raise (ERROR "main.calculateAtSize received an unrecognized pattern")

(* parses the list of rules used X number of uses *)
let parse_lst lst =
  let rec aux lst = 
    match lst with
	[(name_rule,numbUses)]    ->  name_rule ^ " x " ^ (string_of_int numbUses)
      | (name_rule,numbUses)::ls  ->  name_rule ^ " x " ^ (string_of_int numbUses) ^ " " ^ (aux ls)
      | []                        ->  ""
  in "Rules used: " ^ aux lst
  
let rec count_ruleUses lst = 
  match lst with
      [(name_rule,numbUses)]    ->  numbUses
    | (name_rule,numbUses)::ls  ->  numbUses + count_ruleUses ls 
    | []                        ->  0

let rec average_lst lst n_float = 
  match lst with
      [(name_rule,numbUses)]    ->  name_rule ^ " x " ^ (string_of_float ((float numbUses)/.n_float))
    | (name_rule,numbUses)::ls  ->  name_rule ^ " x " ^ (string_of_float ((float numbUses)/.n_float)) ^ " " ^ (average_lst ls n_float)
    | []                        ->  ""

let rec build_total_lst old_lst lst = if old_lst=[] then lst else
  [? List : (name,old_n+n) | (name,n) <- List : lst ; (name',old_n) <- List : old_lst ; name=name' ?]

let printH h =
  let () = 
    match h with
	KeSet h  ->  
	  print ("The History set has cardinality " ^ string_of_int (SetPairsKE.cardinal h) ^ ". Would you like me to print it? (y/n) ")
      | StSet h  ->  
	print ("The History set has cardinality " ^ string_of_int (SetPairsIntSet.cardinal h) ^ ". Would you like me to print it? (y/n) ")
      | StSetPair (h1,h2) ->
	print ("The History sets have cardinality " ^ string_of_int (SetPairsIntSet.cardinal h1) ^ " (<=) and " ^
		  string_of_int (SetPairsIntSet.cardinal h2) ^ " (>=). Would you like me to print them? (y/n)") in
  let choice = read_line () in
  let f x = if (x="y") then
      match h with 
	KeSet h           -> (print_str (" History: "); MPairsKE.printSetPairsKE_nl h)
      | StSet h           -> (print_str (" History: "); MPairsIntSet.printSet_nl h) 
      | StSetPair (h1,h2) -> (print_str (" History1 (<=): "); MPairsIntSet.printSet_nl h1;
			      print_str (" History2 (>=): "); MPairsIntSet.printSet_nl h2;)
  in
  f choice

let unmarshal i_marshal_c = 
  let exp1 = Marshal.from_channel i_marshal_c in
  exp1
(* If you know the type of the unmarshaled data, you may uncomment this and make the necessary adjustments.
   if !optBSKE then
   let exp1 : MKEPos.ke = Marshal.from_channel i_marshal_c in
   ExpPos exp1
   else
   let exp1 : MKE.ke = Marshal.from_channel i_marshal_c in
   Exp exp1*)

let print_list_pairsKE lst =
  let browse (e1,e2) = 
    match (e1,e2) with
	(Exp eA,Exp eB)       ->  print ((MKE.katExp2String eA) ^ " & " ^ (MKE.katExp2String eB) ^ "\n")
      | (ExpPos eA,ExpPos eB) ->  print ((MKEPos.katExp2String_normalize eA)^" & "^(MKEPos.katExp2String_normalize eB) ^ "\n") 
      | (Str sA,Str sB)       ->  print (sA ^ " & " ^ sB)
      | _                     ->  raise (ERROR "Unexpected pattern received in main.print_list_pairsKE")
  in List.iter browse lst

let rec checkSameResults resultsLs =
  let (lsOfTriples:(resultsAccum*int*timeMap) list) = fst (List.split resultsLs) in
  match lsOfTriples with
      []                 ->  (true,Number 0)
    | (List ls,_,_)::tl  ->  
      let (answer,lst) = foo tl ls in
      if answer then (true,Number (List.length ls))
      else 
	(false,List lst)
    | (Number n,_,_)::tl ->
      let answer = bar tl n in
      if answer then
	(true,Number n)
      else 
	(false, Number (-1))
and foo (list:(resultsAccum*int*timeMap) list) (ls1:(model*model) list) : (bool*((model*model) list)) = 
  match list with
      []                 ->  (true,[])
    | (List ls2,_,_)::tl ->  let (answer,ls) = foo tl ls2 in
			     let lsD = lsDiff2 ls1 ls2 in
			     ((lsD=[])&&answer,lsD@ls) 
    | _                  ->  raise (Error "unexpected pattern") 
and bar (list:(resultsAccum*int*timeMap) list) (n:int) = 
  match list with
      (Number m,_,_)::tl   ->  (n=m) && (bar tl n)
    | []                   ->  true
    | _                    ->  raise (Error "unexpected pattern")
     
(* The following function calls the appropriate equiv function, given the parameters it receives *) 
let rec equiv (param1:model) (param2:model) : (bool*setH) =
  match (param1,param2) with
      (NFAGS a1, NFAGS a2)  -> 
	( match (CallNFAGSEquivChecker.call a1 a2) with 
	    (answer, Some setR, None)        -> (answer,StSet setR)
	  | (answer, Some setR1, Some setR2) -> (answer,StSetPair (setR1,setR2))
	  | (answer, None, Some setR)        -> (answer,StSet setR)
	  | (answer, None, None)             -> raise (ERROR "Wrong pattern received by main.equiv") )
    | (Exp e1,Exp e2)       ->
      ( match (EquivalenceCheckerPD.(main e1 e2)) with (answer, setR) -> (answer,KeSet setR) )
    | (ExpPos e1, ExpPos e2)->
      let (a1,a2) = buildPairNFAGSUsingBSKE e1 e2 in
      equiv (NFAGS a1) (NFAGS a2) 
    | _                     ->  raise (ERROR "Unexpected pattern in function main.equiv")

let rec incl (param1:model) (param2:model) : (bool*setH) = 
  match (param1,param2) with
      (NFAGS (NFAGS.TransProd a1), NFAGS (NFAGS.TransProd a2)) ->
	let main = 
	  if !inclV1 then InclusionCheckerAGS1.main else 
	    if !inclV2 then InclusionCheckerAGS2.main else 
	      raise (ERROR "No version given in function incl.") in
	( match (main a1 a2) with (answer,r) -> (answer,StSet r) )
    | (ExpPos e1, ExpPos e2) ->
      let (a1,a2) = buildPairNFAGSUsingBSKE e1 e2 in
      incl (NFAGS a1) (NFAGS a2)
    | _                     ->  raise (ERROR "Unexpected pattern in function main.incl")

(*----------- Parsing functions -------------*)

let parseInfix stdinbuf = 
  let lexer_line = Lexer.line stdinbuf in     (* a line from the stdinbuf (a string) *)
  let linebuf = Lexing.from_string lexer_line in
  let exp = List.hd (Parser.data Lexer.token linebuf) in
  Exp exp

let parsePolish stdinbuf = 
  let lexer_line = LexerPolish.line stdinbuf in
  let linebuf = Lexing.from_string lexer_line in
  let exp = tryListHd (ParserPolish.data LexerPolish.token linebuf) "main.parsePolish" in
  Exp exp

let parseInfixWithPositions stdinbuf = 
  let lexer_line = LexerPosition.line stdinbuf in
  let linebuf = Lexing.from_string lexer_line in
  let exp = List.hd (ParserPosition.data LexerPosition.token linebuf) in
  ExpPos exp

let parsePolishWithPositions stdinbuf = 
  let lexer_line = LexerPolishPosition.line stdinbuf in
  let linebuf = Lexing.from_string lexer_line in
  let exp = 
    tryListHd (ParserPolishPosition.data LexerPolishPosition.token linebuf) "main.parsePolishWithPositions" in
  ExpPos exp

let parseWithPositions stdinbuf =
  if !verbose then
    parseInfixWithPositions stdinbuf
  else
    parsePolishWithPositions stdinbuf

let parseWithoutPositions stdinbuf = 
  if !verbose then
    parseInfix stdinbuf
  else
    parsePolish stdinbuf

let parse_stdinbuf stdinbuf =
  if !optBSKE then
    parseWithPositions stdinbuf
  else 
    parseWithoutPositions stdinbuf

let parse_string line = 
  if (!optBSKE) then
    ExpPos (MUtils.parsePolishPosition line)
  else
    if !verbose then
      Exp (MUtils.parse line)
    else
      Exp (MUtils.parsePolish line)

(* ------------ Input/Output routines ------------- *)

let verbose_pair () = 
  printOptsEnabled () ;
  print_str intro_text;
  
  while true do
    print "\nPlease provide two KAT expressions to check for language equality. Two NFAGS (Non-Deterministic Automata on Guarded Strings) will be generated using the Berry-Sethi construction for KAT expressions and the decision algorithm will be applied on these automata. Type the expressions separated by an ENTER:" ;
     
    let stdinbuf = Lexing.from_channel stdin in
    let exp1 = parse_stdinbuf stdinbuf in
    let exp2 = parse_stdinbuf stdinbuf in
    let (answer, h) = equiv exp1 exp2 in
    let () = print ("Answer: " ^ string_of_bool answer) in
    printH h ; if answer then 
	match h with
	    StSetPair _ -> 
	      print "History1 and History2 are simulations (and the union of the two is a bisimulation)."
	  | _           -> 
	    print "History is a bisimulation."
  done

let verbose_incl_pair () = 
  printOptsEnabled () ;
  print_str intro_text;
  
  while true do
    print "\nPlease provide two KAT expressions to check for language inclusion. Two NFAGS (Non-Deterministic Automata on Guarded Strings) will be generated using the Berry-Sethi construction for KAT expressions and the decision algorithm will be applied on these automata. Type the expressions separated by an ENTER:" ;  
     
    let stdinbuf = Lexing.from_channel stdin in
    let exp1 = parse_stdinbuf stdinbuf in
    let exp2 = parse_stdinbuf stdinbuf in
    let (answer, h) = incl exp1 exp2 in
    let () = print ("Answer: " ^ string_of_bool answer) in
    printH h ; if answer then print "History is a simulation" 
  done

(* With assumptions/hypotheses *)
let verbose_assumptions () =
  printOptsEnabled () ;
  print intro_text;
  print assumptions_text;

  while true do    
    let stdinbuf = Lexing.from_channel stdin in
    let () = print_str "\n> Assumptions: " ; flush stdout in
    let linebuf0 = Lexing.from_string (LexerAssumptions.line stdinbuf) in
    let () = print_str "> 1st Expression: " ; flush stdout in
    let linebuf1 = Lexing.from_string (Lexer.line stdinbuf) in
    let () = print_str "> 2nd Expression: " ; flush stdout in
    let linebuf2 = Lexing.from_string (Lexer.line stdinbuf) in
    let assumpPair = ParserAssumptions.data LexerAssumptions.token linebuf0 in
    let exp1 = List.hd (Parser.data Lexer.token linebuf1) in
    let exp2 = List.hd (Parser.data Lexer.token linebuf2) in
    let (answer, h) = EquivalenceCheckerPDWithAssumptions.(main assumpPair exp1 exp2) in
    let () =   
      print_str ("Answer: " ^ string_of_bool answer ^ "\n");
      print_str ("The History set has cardinality " ^ string_of_int (SetPairsKE.cardinal h) ^ ". Would you like me to print it? (y/n) "); in
    let () = flush stdout in
    let choice = read_line () in
    let f x = if (x="y") then (print_str (" History: " ); MPairsKE.printSetPairsKE h
			      ; if answer then print "History is a bisimulation" ) in
    f choice
  done 

let verbose_assumptions_compare () = 
  printOptsEnabled () ;
  print intro_text;
  print assumptions_text_2;
  
  while true do    
    let stdinbuf = Lexing.from_channel stdin in
    let () = print_str "\n> Assumptions: " ; flush stdout in
    let linebuf0 = Lexing.from_string (LexerAssumptions.line stdinbuf) in
    let () = print_str "> 1st Expression: " ; flush stdout in
    let linebuf1 = Lexing.from_string (Lexer.line stdinbuf) in
    let () = print_str "> 2nd Expression: " ; flush stdout in
    let linebuf2 = Lexing.from_string (Lexer.line stdinbuf) in
    let assumpPair = ParserAssumptions.data LexerAssumptions.token linebuf0 in
    let exp1 = List.hd (Parser.data Lexer.token linebuf1) in
    let exp2 = List.hd (Parser.data Lexer.token linebuf2) in
    let (answer, h) = EquivalenceCheckerPDWithAssumptions.(main assumpPair exp1 exp2) in
    let () =   
      print_str ("Answer1: " ^ string_of_bool answer ^ "\n");
      print_str ("The History set has cardinality " ^ string_of_int (SetPairsKE.cardinal h) ^ ". Would you like me to print it? (y/n) "); in
    let () = flush stdout in
    let choice = read_line () in
    let f x = if (x="y") then (print_str (" History: " ); MPairsKE.printSetPairsKE h 
			      ; if answer then print "History is a bisimulation" ) in
    let () = f choice in
    let ((answer,h),exp1',exp2') = EquivalenceCheckerPDWithAssumptionsUru.(main assumpPair exp1 exp2) in
    let () = print "\nAfter including the assumptions in the expressions they become: " in
    let () = print (MKE.katExp2String exp1') in
    let () = print (MKE.katExp2String exp2') in 
    let () = flush stdout in
    let () =   
      print_str ("Answer2: " ^ string_of_bool answer ^ "\n");
      print_str ("The History set has cardinality " ^ string_of_int (SetPairsKE.cardinal h) ^ ". Would you like me to print it? (y/n) "); in
    let () = flush stdout in
    let choice = read_line () in
    let f x = if (x="y") then (print_str (" History: " ); MPairsKE.printSetPairsKE h ; 
			       if answer then print "History is a bisimulation") in
    f choice
  done ;;


(*---- The following are interface-less routines for dealing not with 1 pair of exps but with a large sequence of exps.  -----*)
(* m=total number of expressions; *)
let rec automatic_self m =
  try 
    let stdinbuf = Lexing.from_channel stdin in
    let marshal_filename = "marshaled_exps_self_" ^ (string_of_float (Unix.time())) in
    let o_marshal_c = open_out marshal_filename in
    let (numbTests_accum,ats_accum) = preprocess stdinbuf o_marshal_c m 0 0 0. in
    let i_marshal_c = open_in marshal_filename in
    let () = loop m stdinbuf i_marshal_c 0 0 (newTimeMap ()) (ref true) in
    let () = print ("The average number of test occurrences per expression was: " ^ (string_of_float ((float numbTests_accum) /. (float m))) ) in
    let () = print ("The average cardinality of At was: " ^ (string_of_float (ats_accum /. (float m)))) in
    Sys.remove marshal_filename
  with LexerPolish.Eof -> exit 0
(* m=total no. of exps; n=no. of exps tested so far; lst=list of pairs of
   the random exps that were equivalent *)
and loop m stdinbuf i_marshal_c n h_accum t_accum success = 
  if n=m then 
    (if !success==true then
	print ("\nSummary: I was successful in testing the equivalence of every expression with itself.")
     else 
	print ("Failure: I obtained False when testing some expressions with themselves...") ; 
     print ("The average cardinality of H was: " ^ string_of_float ((float h_accum)/.(float n))) ;
     print ("The average time taken per test was: "); printTimeMap (divideTimeMapByN m t_accum) ;
     close_in i_marshal_c )
  else
    (* The test *)
    (*let exp : MKE.ke = Marshal.from_channel i_marshal_c in*)
    let exp = unmarshal i_marshal_c in
    
    let (answer,h,new_t_accum) = 
      if !optBSKE then
	let t_i = Sys.time () in
	let (a1,a2) = buildPairNFAGSUsingBSKE exp exp in
	let ta = Sys.time() -. t_i in
	let t_i = Sys.time () in
	let (answer,h) = equiv (NFAGS a1) (NFAGS a2) in
	let td = Sys.time() -. t_i in
	let new_t_accum = incrementTimeMap keyw_time_NFAGSConstr ta t_accum  in
	let new_t_accum = incrementTimeMap keyw_time_DecisionProc td new_t_accum in
	(answer,h,new_t_accum)
      else 
	let t_i = Sys.time() in  
	let (answer,h) = equiv (Exp exp) exp in
	let td = Sys.time() -. t_i in
	let new_t_accum = incrementTimeMap keyw_time_DecisionProc td t_accum in
	(answer,h,new_t_accum)    in
    let new_h_accum = h_accum+(setHCardinal h) in
    let () =
      if answer!=true then 
	(print ("Test: " ^ (string_of_int (n+1)) ^ " #H: " ^ (string_of_int (setHCardinal h)) ^ " false");
	 success := false)
      else 
	(print ("Test: " ^ (string_of_int (n+1)) ^ " #H: " ^ (string_of_int (setHCardinal h)) ^ " " ^ (string_of_bool answer)) ) in
    loop m stdinbuf i_marshal_c (n+1) new_h_accum new_t_accum success
and print_total_vals h_accum = 
  print ("Total values so far: H=" ^ (string_of_int h_accum)) ; 
  flush stdout
and preprocess stdinbuf oc m n numbTests_accum ats_accum =
  if n<m then
    let exp = parse_stdinbuf stdinbuf in
    let numbTests = 
      match exp with 
	  Exp e    ->  countTestsInKe e 
	| ExpPos e ->  countTestsInKePos e
	| _        ->  raise (ERROR "preprocess in automatic_self received unexpected input pattern")   in
    let ats = calculateAtSize exp exp in
    let () = Marshal.to_channel oc exp [] in
    preprocess stdinbuf oc m (n+1) (numbTests_accum+numbTests) (ats_accum +. ats)
  else 
    let () = close_out oc in
    (numbTests_accum,ats_accum)

(* automatic_pairs starts here *)
let automatic_pairs_preprocess stdinbuf oc m n =
  let rec aux exp1 stdinbuf oc m n ats_accum =
    if n<m then
      let exp = parse_stdinbuf stdinbuf in
      let ats = calculateAtSize exp exp1 in
      let () = Marshal.to_channel oc exp [] in
      aux exp stdinbuf oc m (n+1) (ats_accum+.ats)
    else 
      let () = close_out oc in
      ats_accum   in
  let exp = parse_stdinbuf stdinbuf in
  let () = Marshal.to_channel oc exp [] in
  aux exp stdinbuf oc (m-1) n 0.
(* The data comes from a marshaled file where the expressions are already parsed;
   m=total no. of exps; n=no. of exps tested so far; lst=list of pairs of
   the random exps that were equivalent *)
let rec automatic_pairs_loop_parsed exp1 m i_marshal_c n h_accum t_accum resultsAccum = 
  if n=m then 
    (* End *)
    (close_in i_marshal_c ; (resultsAccum,h_accum,t_accum))
  else
    (* The test *)    
    let exp2 = unmarshal i_marshal_c in
    
    let (answer,h,new_t_accum) = 
      match (exp1,exp2) with
	  (ExpPos e1,ExpPos e2) -> 
	    let t_i = Sys.time () in
	    let (a1,a2) = buildPairNFAGSUsingBSKE e1 e2 in
	    let ta = Sys.time() -. t_i in
	    let t_i = Sys.time() in
	    let (answer,h) = equiv (NFAGS a1) (NFAGS a2) in
	    let td = Sys.time() -. t_i in
	    let new_t_accum = incrementTimeMap keyw_time_NFAGSConstr ta t_accum in
	    let new_t_accum = incrementTimeMap keyw_time_DecisionProc td new_t_accum in
	    (answer,h,new_t_accum)
	| (Exp e1,Exp e2)       ->
	  let t_i = Sys.time() in
	  let (answer,h) = equiv (Exp e1) (Exp e2) in
	  let td = Sys.time() -. t_i in
	  let new_t_accum = incrementTimeMap keyw_time_DecisionProc td t_accum in
	  (answer,h,new_t_accum) 
	| _                     ->
	  raise (ERROR "Unexpected pattern found")   in
    let new_h_accum = h_accum+(setHCardinal h) in
    let () = 
      print ("Test: " ^ string_of_int (n+1) ^ " #H: " ^ (string_of_int (setHCardinal h)) 
	     ^ " answer: " ^ (string_of_bool answer)) in
    let newResultsAccum = 
      if answer then 
	match resultsAccum with 
	    Number n  ->  Number (n+1)
	  | List ls   ->  List ((exp1,exp2)::ls)
      else resultsAccum 
    in
    automatic_pairs_loop_parsed exp2 m i_marshal_c (n+1) new_h_accum new_t_accum newResultsAccum
let automatic_pairs_finish (resultsAccum,h_accum,t_accum) m flagPrintEqExps =
  (match resultsAccum with 
      List lst  ->  
	if not(lst=[]) then
	  (print_str("Summary: I found "^string_of_int (List.length lst)^" pairs of expressions to be equivalent") 
	  ; if flagPrintEqExps then (print ":"; print_list_pairsKE lst ) else print "")
	else
	  print "Summary: All tests returned False" 
    | Number n  ->  
      print ("I found "^string_of_int n^" pairs of expressions to be equivalent") )
  ; print ("The average cardinality of H was: " ^ string_of_float ((float h_accum)/.(float m))) 
  ; print ("The average time taken per test was: "); printTimeMap (divideTimeMapByN m t_accum)

(*  m=total number of expressions; *)
let automatic_pairs_main m flagPrintEqExps =
  try 
    let marshal_filename = "marshaled_exps_pair_" ^ (string_of_float (Unix.time())) in 
    let i_marshal_c = open_in marshal_filename in
    let exp1 = unmarshal i_marshal_c in
    let (resultsAccum,h_accum,t_accum) = 
      let resultsAccum = if !optDontSL then Number 0 else List [] in
      automatic_pairs_loop_parsed exp1 (m-1) i_marshal_c 0 0 (newTimeMap ()) resultsAccum in
    let () = automatic_pairs_finish (resultsAccum,h_accum,t_accum) (m-1) flagPrintEqExps in
    Sys.remove marshal_filename
  with LexerPolish.Eof -> exit 0

let rec preprocess_marshal_without_parsing output_channel m =
  if m>0 then
    let line = input_line stdin in
    let () = Marshal.to_channel output_channel line [] in
    preprocess_marshal_without_parsing output_channel (m-1)
  else
    close_out output_channel

let rec automatic_pairs_loop_unparsed n line1 m i_marshal_c h_accum t_accum resultsAccum print_each_test flagDontSL = 
  if m=0 then
    (* End *)
    (close_in i_marshal_c ; (resultsAccum,h_accum,t_accum))
  else
    (* The Test *)
    let exp1 = parse_string line1 in
    let line2 = unmarshal i_marshal_c in
    let exp2 = parse_string line2 in
    let (answer,h,new_t_accum) = 
      match (exp1,exp2) with 
	  (ExpPos e1,ExpPos e2)  ->  
	    let t_i = Sys.time () in
	    let (a1,a2) = buildPairNFAGSUsingBSKE e1 e2 in
	    let ta = Sys.time() -. t_i in
	    let t_i = Sys.time () in
	    let (answer,h) = equiv (NFAGS a1) (NFAGS a2) in
	    let td = Sys.time() -. t_i in
	    let new_t_accum = incrementTimeMap keyw_time_NFAGSConstr ta t_accum  in
	    let new_t_accum = incrementTimeMap keyw_time_DecisionProc td new_t_accum in
	    (answer,h,new_t_accum)
	| (Exp e1,Exp e2)        ->
	  let t_i = Sys.time() in  
	  let (answer,h) = equiv exp1 exp2 in
	  let td = Sys.time() -. t_i in
	  let new_t_accum = incrementTimeMap keyw_time_DecisionProc td t_accum in
	  (answer,h,new_t_accum)    
	| _                      -> 
	  raise (ERROR "Unexpected pattern found")   in
    let () = 
      if print_each_test then 
	print ("Test: "^string_of_int (n+1)^" #H: "^(string_of_int (setHCardinal h))^" answer: "^(string_of_bool answer)) in
    let new_h_accum = h_accum + (setHCardinal h) in
    let newResultsAccum = 
      match resultsAccum with
	  List ls  ->  if answer=true then List ((Str line1,Str line2)::ls) else List ls 
	| Number n ->  if answer=true then Number (n+1) else Number n
    in
    automatic_pairs_loop_unparsed (n+1) line2 (m-1) i_marshal_c new_h_accum new_t_accum newResultsAccum print_each_test flagDontSL
      
(* Same as prev function but for disjoint pairs of exps *)
let rec automatic_pairs_loop_unparsed_disj n m i_marshal_c h_accum t_accum resultsAccum print_each_test flagDontSL = 
  if m<=0 then
    (* End *)
    (close_in i_marshal_c ; (resultsAccum,h_accum,t_accum))
  else
    (* The Test *)
    let line1 = unmarshal i_marshal_c in
    let exp1 = parse_string line1 in
    let line2 = unmarshal i_marshal_c in
    let exp2 = parse_string line2 in
    let (answer,h,new_t_accum) = 
      match (exp1,exp2) with 
	(ExpPos e1,ExpPos e2)  ->  
	  let t_i = Sys.time () in
	  let (a1,a2) = buildPairNFAGSUsingBSKE e1 e2 in
	  let ta = Sys.time() -. t_i in
	  let t_i = Sys.time () in
	  let (answer,h) = equiv (NFAGS a1) (NFAGS a2) in
	  let td = Sys.time() -. t_i in
	  let new_t_accum = incrementTimeMap keyw_time_NFAGSConstr ta t_accum  in
	  let new_t_accum = incrementTimeMap keyw_time_DecisionProc td new_t_accum in
	  (answer,h,new_t_accum)
      | (Exp e1,Exp e2)        ->
	let t_i = Sys.time() in  
	let (answer,h) = equiv exp1 exp2 in
	let td = Sys.time() -. t_i in
	let new_t_accum = incrementTimeMap keyw_time_DecisionProc td t_accum in
	(answer,h,new_t_accum)    
      | _                      -> 
	raise (ERROR "Unexpected pattern found")   in
    let () = 
      if print_each_test then 
	print ("Test: "^string_of_int (n+1)^" #H: "^(string_of_int (setHCardinal h))^" answer: "^(string_of_bool answer)) in
    let new_h_accum = h_accum + (setHCardinal h) in
    let newResultsAccum = 
      match resultsAccum with
	  List ls  ->  if answer=true then List ((Str line1,Str line2)::ls) else List ls 
	| Number n ->  if answer=true then Number (n+1) else Number n
    in
    automatic_pairs_loop_unparsed_disj (n+1) (m-2) i_marshal_c new_h_accum new_t_accum newResultsAccum print_each_test flagDontSL

let rec automatic_self_loop_unparsed n m i_marshal_c h_accum t_accum resultsAccum print_each_test flagDontSL = 
  if m=0 then
    (* End *)
    (close_in i_marshal_c ; (resultsAccum,h_accum,t_accum))
  else
    (* The Test *)
    let line = unmarshal i_marshal_c in
    let exp1 = parse_string line in
    let exp2 = parse_string line in
    let (answer,h,new_t_accum) = 
      match (exp1,exp2) with 
	(ExpPos e1,ExpPos e2)  ->  
	  let t_i = Sys.time () in
	  let (a1,a2) = buildPairNFAGSUsingBSKE e1 e2 in
	  let ta = Sys.time() -. t_i in
	  let t_i = Sys.time () in
	  let (answer,h) = equiv (NFAGS a1) (NFAGS a2) in
	  let td = Sys.time() -. t_i in
	  let new_t_accum = incrementTimeMap keyw_time_NFAGSConstr ta t_accum  in
	  let new_t_accum = incrementTimeMap keyw_time_DecisionProc td new_t_accum in
	  (answer,h,new_t_accum)
      | (Exp e1,Exp e2)        ->
	let t_i = Sys.time() in  
	let (answer,h) = equiv exp1 exp2 in
	let td = Sys.time() -. t_i in
	let new_t_accum = incrementTimeMap keyw_time_DecisionProc td t_accum in
	(answer,h,new_t_accum)    
      | _                      -> 
	raise (ERROR "Unexpected pattern found")   in
    let () = 
      if print_each_test then 
	print ("Test: "^string_of_int (n+1)^" #H: "^(string_of_int (setHCardinal h))
	       ^" answer: "^(string_of_bool answer)) in
    let new_h_accum = h_accum + (setHCardinal h) in
    let newResultsAccum = 
      match resultsAccum with
	  List ls  ->  if answer=true then List ((Str line,Str line)::ls) else List ls 
	| Number n ->  if answer=true then Number (n+1) else Number n
    in
    automatic_self_loop_unparsed (n+1) (m-1) i_marshal_c new_h_accum new_t_accum newResultsAccum print_each_test flagDontSL

let rec automatic_pairs_incl_loop_unparsed n line1 m i_marshal_c h_accum t_accum resultsAccum print_each_test flagDontSL = 
  if m=0 then
    (* End *)
    (close_in i_marshal_c ; (resultsAccum,h_accum,t_accum))
  else
    (* The Test *)
    let e1 =  (MUtils.parsePolishPosition line1) in
    let line2 = unmarshal i_marshal_c in
    let e2 =  (MUtils.parsePolishPosition line2) in
    let (answer,h,new_t_accum) = 
      let t_i = Sys.time () in
      let (a1,a2) = buildPairNFAGSUsingBSKE e1 e2 in
      let ta = Sys.time() -. t_i in
      let t_i = Sys.time () in
      let (answer,h) = incl (NFAGS a1) (NFAGS a2) in
      let td = Sys.time() -. t_i in
      let new_t_accum = incrementTimeMap keyw_time_NFAGSConstr ta t_accum  in
      let new_t_accum = incrementTimeMap keyw_time_DecisionProc td new_t_accum in
      (answer,h,new_t_accum) in
    let () = 
      if print_each_test then 
	print ("Test: "^string_of_int (n+1)^" #H: "^(string_of_int (setHCardinal h))
	       ^" answer: "^(string_of_bool answer)) in
    let new_h_accum = h_accum + (setHCardinal h) in
    let newResultsAccum = 
      match resultsAccum with
	  List ls  -> if answer=true then List ((Str line1,Str line2)::ls) else List ls 
	| Number n -> if answer=true then Number (n+1) else Number n
    in
    automatic_pairs_incl_loop_unparsed (n+1) line2 (m-1) i_marshal_c new_h_accum new_t_accum newResultsAccum print_each_test flagDontSL
      
(* Same as prev function but for disjoint pairs of exps *)
let rec automatic_pairs_incl_loop_unparsed_disj n m i_marshal_c h_accum t_accum resultsAccum print_each_test flagDontSL = 
  if m<=0 then
    (* End *)
    (close_in i_marshal_c ; (resultsAccum,h_accum,t_accum))
  else
    (* The Test *)
    let line1 = unmarshal i_marshal_c in
    let e1 = MUtils.parsePolishPosition line1 in
    let line2 = unmarshal i_marshal_c in
    let e2 = MUtils.parsePolishPosition line2 in
    let (answer,h,new_t_accum) = 
      let t_i = Sys.time () in
      let (a1,a2) = buildPairNFAGSUsingBSKE e1 e2 in
      let ta = Sys.time() -. t_i in
      let t_i = Sys.time () in
      let (answer,h) = incl (NFAGS a1) (NFAGS a2) in
      let td = Sys.time() -. t_i in
      let new_t_accum = incrementTimeMap keyw_time_NFAGSConstr ta t_accum  in
      let new_t_accum = incrementTimeMap keyw_time_DecisionProc td new_t_accum in
      (answer,h,new_t_accum) in
    let () = 
      if print_each_test then 
	print ("Test: "^string_of_int (n+1)^" #H: "^(string_of_int (setHCardinal h))^" answer: "
	       ^(string_of_bool answer)) in
    let new_h_accum = h_accum + (setHCardinal h) in
    let newResultsAccum = 
      match resultsAccum with
	  List ls  -> if answer=true then List ((Str line1,Str line2)::ls) else List ls 
	| Number n -> if answer=true then Number (n+1) else Number n
    in
    automatic_pairs_incl_loop_unparsed_disj (n+1) (m-2) i_marshal_c new_h_accum new_t_accum newResultsAccum print_each_test flagDontSL

let rec automatic_self_incl_loop_unparsed n m i_marshal_c h_accum t_accum resultsAccum print_each_test flagDontSL = 
  if m=0 then
    (* End *)
    (close_in i_marshal_c ; (resultsAccum,h_accum,t_accum))
  else
    (* The Test *)
    let line = unmarshal i_marshal_c in
    let e1 = MUtils.parsePolishPosition line in
    let e2 = MUtils.parsePolishPosition line in
    let (answer,h,new_t_accum) = 
      let t_i = Sys.time () in
      let (a1,a2) = buildPairNFAGSUsingBSKE e1 e2 in
      let ta = Sys.time() -. t_i in
      let t_i = Sys.time () in
      let (answer,h) = incl (NFAGS a1) (NFAGS a2) in
      let td = Sys.time() -. t_i in
      let new_t_accum = incrementTimeMap keyw_time_NFAGSConstr ta t_accum  in
      let new_t_accum = incrementTimeMap keyw_time_DecisionProc td new_t_accum in
      (answer,h,new_t_accum) in
    let () = 
      if print_each_test then 
	print ("Test: "^string_of_int (n+1)^" #H: "^(string_of_int (setHCardinal h))^" answer: "
	       ^(string_of_bool answer)) in
    let new_h_accum = h_accum + (setHCardinal h) in
    let newResultsAccum = 
      match resultsAccum with
	  List ls  -> if answer=true then List ((Str line,Str line)::ls) else List ls 
	| Number n -> if answer=true then Number (n+1) else Number n
    in
    automatic_self_incl_loop_unparsed (n+1) (m-1) i_marshal_c new_h_accum new_t_accum newResultsAccum print_each_test flagDontSL

(* Runs 2 methods: Partial Derivatives (PD) and Berry-Sethi for KAT Expressions (BSKE).
   At the end indicates if the set of equivalent pairs found was the same regardless of the method;
   It also prints some statistic measures *)
let automatic_pairs_comparePDWithBS m =
    (* marshaling of the input lines *)
    let marshal_filename = "marshaled_exps_pair_compare_" ^ (string_of_float (Unix.time())) in 
    let o_marshal_c = open_out marshal_filename in
    let () = preprocess_marshal_without_parsing o_marshal_c m in

    let () = print "\nBeggining of the Berry-Sethi (BSKE) automata construction method for KAT Expressions" in
    let () = optBSKE:=true in
    let i_marshal_BSKE_c = open_in marshal_filename in
    let (answers_BSKE,h_accum_BSKE,t_accum_BSKE) = 
      let resultsAccum = if !optDontSL then Number 0 else List [] in
      if !optDisj then
	automatic_pairs_loop_unparsed_disj 0 m i_marshal_BSKE_c 0 (newTimeMap ()) resultsAccum true !optDontSL
      else
	let line1 = unmarshal i_marshal_BSKE_c in
	automatic_pairs_loop_unparsed 0 line1 (m-1) i_marshal_BSKE_c 0 (newTimeMap()) resultsAccum true !optDontSL in
    let () = optBSKE:=false in
    let () = print "End of the Berry-Sethi (BSKE) automata construction method for KAT Expressions\n" in
    
    let () = print "Beginning of the Partial Derivatives (PD) method\n" in
    let () = optBSKE:=false in
    let i_marshal_PD_c = open_in marshal_filename in 
    let (answers_PD,h_accum_PD,t_accum_PD) = 
      let resultsAccum = if !optDontSL then Number 0 else List [] in
      if !optDisj then
	automatic_pairs_loop_unparsed_disj 0 m i_marshal_PD_c 0 (newTimeMap ()) resultsAccum true !optDontSL
      else
	let line1 = unmarshal i_marshal_PD_c in
	automatic_pairs_loop_unparsed 0 line1 (m-1) i_marshal_PD_c 0 (newTimeMap ()) resultsAccum true !optDontSL in
    let () = print "End of the Partial Derivatives (PD) method\n" in

    (* The result of the two methods and a comparison *)    
    let () = print "\nStats of the Berry-Sethi (BSKE) method:" in
    let () = automatic_pairs_finish (answers_BSKE,h_accum_BSKE,t_accum_BSKE) (if !optDisj then (m/2) else (m-1)) false in
    
    let () = print "Stats of the Partial Derivatives (PD) method:" in
    let () = automatic_pairs_finish (answers_PD,h_accum_PD,t_accum_PD) (if !optDisj then (m/2) else (m-1)) false in
    
    let () = 
      match (answers_PD, answers_BSKE) with
	  (List lsPD, List lsBSKE)  ->  
	    if lsPD=lsBSKE then 
	      print ("\nConclusion: the two methods gave equal answers in every test!") 
	    else 
	      (print "\nConclusion: the two methods were not successful in the same tests... Here are some (or all) of the divergent pairs of expressions:" ; 
	       print_list_pairsKE (lsDiff2 lsPD lsBSKE) )
	| (Number n1, Number n3) ->
	  if n1=n3 then 
	    print ("\nConclusion: the two methods found "^string_of_int n1^" pairs of exps to be equivalent")
	  else
	    (print_str ("\nConclusion: the two methods were not successful in the same tests...") ; 
	     print_str (" PD found "^string_of_int n1^" pairs of exps to be equivalent");
	     print     (" and BSKE found " ^ (string_of_int n3)))
      | _                                 ->  raise (Error "unexpected pattern found")
    in 
    Sys.remove marshal_filename
      
let automatic_pairs_compareOpts ?(marshal_filename="") m =
  let just_this_execution = marshal_filename="" in
  let marshal_filename = 
    if just_this_execution then
      (* marshaling of the input lines *)
      let marshal_filename = "marshaled_exps_pair_compareOpts_" ^ (string_of_float (Unix.time())) in 
      let o_marshal_c = open_out marshal_filename in
      let () = preprocess_marshal_without_parsing o_marshal_c m in
      marshal_filename
    else 
      marshal_filename in
  let results = ref [] in
  let opts = ref (if !optBSKE then (EquivalenceCheckerAGS.opts@CallNFAGSEquivChecker.opts) else EquivalenceCheckerPD.opts) in
  let () = opts:=shuffleList !opts () in    

  while not(!opts=[]) do
    let (opt,_,desc) = List.hd !opts in
    let () =
      ( let () = opt:=true in

	let () = print ("Beginning of the Execution using the Option " ^ desc) in
	let i_marshal_c = open_in marshal_filename in 
	let (lst,h_accum,t_accum_1) = 
	  let resultsAccum = if !optDontSL then Number 0 else List [] in
	  if !optDisj then
	    automatic_pairs_loop_unparsed_disj 0 m i_marshal_c 0 (newTimeMap ()) resultsAccum just_this_execution !optDontSL
	  else
	    let line1 = unmarshal i_marshal_c in
	     automatic_pairs_loop_unparsed 0 line1 (m-1) i_marshal_c 0 (newTimeMap ()) resultsAccum just_this_execution !optDontSL in
	
	let new_t_accum = t_accum_1 in
	let triple = (lst,h_accum, new_t_accum) in
	let () = results := (!results)@[(triple,desc)] in
	opt := false) in
    let () = opts := List.tl !opts in
    print ("End of the Execution using the Option " ^ desc)
  done 
  ;  
  let () = print ("Beginning of the Execution without any options enabled") in
  let i_marshal_c = open_in marshal_filename in 
  let (lst,h_accum,t_accum_1) = 
    let resultsAccum = if !optDontSL then Number 0 else List [] in
    if !optDisj then 
      automatic_pairs_loop_unparsed_disj 0 m i_marshal_c 0 (newTimeMap ()) resultsAccum just_this_execution !optDontSL
    else
      let line1 = unmarshal i_marshal_c in
      automatic_pairs_loop_unparsed 0 line1 (m-1) i_marshal_c 0 (newTimeMap ()) resultsAccum just_this_execution !optDontSL in

  let new_t_accum = t_accum_1 in
  let triple = (lst,h_accum,new_t_accum) in
  let () = results := (!results)@[(triple,"Using no options")] in
  let () = print ("End of the Execution without any options") in
  
  (* Checking if all options found the same equivalence pairs *)
  let (sameAnsr,y) = checkSameResults !results in
  (* The results of using each option *)
  let () = results := (List.sort ~cmp:compare !results) in
  while not(!results=[]) do
    let ((lst,h_accum,t_accum),desc) = List.hd !results in
    let () = print ("\nStats of the Option " ^ desc ^ "\n") in
    let () = automatic_pairs_finish (lst,h_accum,t_accum) (if !optDisj then (m/2) else (m-1)) false in
    results := List.tl !results
  done 
  ; 
  (match (sameAnsr,y) with
      (true,Number n)   ->  
	if !optDontSL then 
	  print ("\nAll different options found success in "^string_of_int n^" pairs")
	else
	  print ("\nAll options found the same pairs ("^string_of_int n^")!")
    | (false,List ls) ->  
      if not(!optDontSL) then 
	(print "\nThe options didn't find the same pairs. Some or all of the divergent pairs are:" ;
	 List.iter (fun (e1,e2) -> (printExp e1); print " and "; (printExp_el e2)) ls )
    | (false,_)       ->  
      if !optDontSL then
	print ("\nThe options didn't find success in the same number of pairs")
    | (true,List _)   ->  raise (Error "unexpected pattern"))
  ; 
  if just_this_execution then Sys.remove marshal_filename

let automatic_self_compareOpts ?(marshal_filename="") m =
  let just_this_execution = marshal_filename="" in
  let marshal_filename = 
    if just_this_execution then
      (* Marshaling of the input lines. *)
      let marshal_filename = "marshaled_exps_self_compareOpts_" ^ (string_of_float (Unix.time())) in 
      let o_marshal_c = open_out marshal_filename in
      let () = preprocess_marshal_without_parsing o_marshal_c m in
      marshal_filename
    else 
      marshal_filename in
  let results = ref [] in
  let opts = ref (
    if !optBSKE then (EquivalenceCheckerAGS.opts@CallNFAGSEquivChecker.opts) else EquivalenceCheckerPD.opts) in
  let () = opts:=shuffleList !opts () in    

  while not(!opts=[]) do
    let (opt,_,desc) = List.hd !opts in
    let () =
      ( let () = opt:=true in

	let () = print ("Beginning of the Execution using the Option " ^ desc) in
	let i_marshal_c = open_in marshal_filename in 
	let (lst,h_accum,t_accum_1) = 
	  let resultsAccum = if !optDontSL then Number 0 else List [] in
	  automatic_self_loop_unparsed 0 m i_marshal_c 0 (newTimeMap ()) resultsAccum just_this_execution !optDontSL in
	
	let new_t_accum = t_accum_1 in
	let triple = (lst,h_accum,new_t_accum) in
	let () = results := (!results)@[(triple,desc)] in
	opt := false) in
    let () = opts := List.tl !opts in
    print ("End of the Execution using the Option " ^ desc)
  done 
  ;  
  let () = print ("Beginning of the Execution without any options enabled") in
  let i_marshal_c = open_in marshal_filename in 
  let (lst,h_accum,t_accum_1) = 
    let resultsAccum = if !optDontSL then Number 0 else List [] in
    automatic_self_loop_unparsed 0 m i_marshal_c 0 (newTimeMap ()) resultsAccum just_this_execution !optDontSL in

  let new_t_accum = t_accum_1 in
  let triple = (lst,h_accum,new_t_accum) in
  let () = results := (!results)@[(triple,"Using no options")] in
  let () = print ("End of the Execution without any options") in
  
  (* Checking if all options found the same equivalence pairs *)
  let (sameAnsr,y) = checkSameResults !results in
  (* The results of using each option *)
  let () = results := (List.sort ~cmp:compare !results) in
  while not(!results=[]) do
    let ((lst,h_accum,t_accum),desc) = List.hd !results in
    let () = print ("\nStats of the Option " ^ desc ^ "\n") in
    let () = automatic_pairs_finish (lst,h_accum,t_accum) (m-1) false in
    results := List.tl !results
  done 
  ; 
  (match (sameAnsr,y) with
      (true,Number n)   ->  
	if !optDontSL then 
	  print ("\nAll different options found success in "^string_of_int n^" pairs")
	else
	  print ("\nAll options found the same pairs ("^string_of_int n^")!")
    | (false,List ls) ->  
      if not(!optDontSL) then 
	(print "\nThe options didn't find the same pairs. Some or all of the divergent pairs are:" ;
	 List.iter (fun (e1,e2) -> (printExp e1); print " and "; (printExp_el e2)) ls )
    | (false,_)       ->  
      if !optDontSL then
	print ("\nThe options didn't find success in the same number of pairs")
    | (true,List _)   ->  raise (Error "unexpected pattern"))
  ; 
  if just_this_execution then Sys.remove marshal_filename

let automatic_self_incl_compareOpts ?(marshal_filename="") (m:int)  =
  let version = if !inclV1 then 1 else if !inclV2 then 2 else raise (ERROR "No inclusion version defined.") in
  let just_this_execution = marshal_filename="" in
  let marshal_filename = 
    if just_this_execution then
      (* Marshaling of the input lines *)
      let marshal_filename = 
	("marshaled_exps_self_incl_"^(string_of_int version)^"_compareOpts_"^(string_of_float (Unix.time()))) in 
      let o_marshal_c = open_out marshal_filename in
      let () = preprocess_marshal_without_parsing o_marshal_c m in
      marshal_filename
    else 
      marshal_filename in
  let results = ref [] in
  let opts = ref (if version=1 then InclusionCheckerAGS1.opts else InclusionCheckerAGS2.opts) in 
  let () = opts:=(shuffleList !opts ()) in
  while not(!opts=[]) do
    let (opt,_,desc) = List.hd !opts in
    let () =
      ( let () = opt:=true in
	
	let () = print ("Beginning of the Execution using the Option " ^ desc) in
	let i_marshal_c = open_in marshal_filename in 
	let (lst,h_accum,t_accum_1) = 
	  let resultsAccum = if !optDontSL then Number 0 else List [] in
	    automatic_self_incl_loop_unparsed 0 m i_marshal_c 0 (newTimeMap ()) resultsAccum just_this_execution !optDontSL in
	
	let triple= (lst,h_accum, t_accum_1) in
	let () = results := (!results)@[(triple,desc)] in
	opt := false) in
    let () = opts := List.tl !opts in
    print ("End of the Execution using the Option " ^ desc)
  done 
  ;
  let () = print ("Beginning of the Execution without any options enabled") in
  let i_marshal_c = open_in marshal_filename in 
  let (lst,h_accum,t_accum_1) = 
    let resultsAccum = if !optDontSL then Number 0 else List [] in
     automatic_self_incl_loop_unparsed 0 m i_marshal_c 0 (newTimeMap ()) resultsAccum just_this_execution !optDontSL in
  let () = print ("End of the Execution without any options") in
  
  let triple = (lst,h_accum,t_accum_1) in
  let () = results := (!results)@[(triple,"Using no options")] in
  
  (* Checking if all options found the same equivalence pairs *)
  let (sameAnsr,y) = checkSameResults !results in
  (* The results of using each option *)
  let () = results := (List.sort ~cmp:compare !results) in
  while not(!results=[]) do
    let ((lst,h_accum,t_accum),desc) = List.hd !results in
    let () = print ("\nStats of the Option " ^ desc ^ "\n") in
    let () = automatic_pairs_finish (lst,h_accum,t_accum) (if !optDisj then (m/2) else (m-1)) false in
    results := List.tl !results
  done 
  ; 
  (match (sameAnsr,y) with
      (true,Number n)   ->  
	if !optDontSL then 
	  print ("\nAll different options found success in "^string_of_int n^" pairs")
	else
	  print ("\nAll options found the same pairs ("^string_of_int n^")!")
    | (false,List ls) ->  
      if not(!optDontSL) then 
	(print "\nThe options didn't find the same pairs. Some or all of the divergent pairs are:" ;
	 List.iter (fun (e1,e2) -> (printExp e1); print " and "; (printExp_el e2)) ls )
    | (false,_)       ->  
      if !optDontSL then
	print ("\nThe options didn't find success in the same number of pairs")
    | (true,List _)   ->  raise (Error "unexpected pattern"))
  ; 
  if just_this_execution then Sys.remove marshal_filename

let automatic_pairs_incl_compareOpts ?(marshal_filename="") (m:int)  =
  let version = if !inclV1 then 1 else if !inclV2 then 2 else raise (ERROR "No inclusion version defined.") in
  let just_this_execution = marshal_filename="" in
  let marshal_filename = 
    if just_this_execution then
      (* marshaling of the input lines *)
      let marshal_filename = 
	("marshaled_exps_pairs_incl_"^(string_of_int version)^"_compareOpts_"^(string_of_float (Unix.time()))) in 
      let o_marshal_c = open_out marshal_filename in
      let () = preprocess_marshal_without_parsing o_marshal_c m in
      marshal_filename
    else 
      marshal_filename in
  let results = ref [] in
  let opts = ref (if version=1 then InclusionCheckerAGS1.opts else InclusionCheckerAGS2.opts) in 
  let () = opts:=(shuffleList !opts ()) in
  while not(!opts=[]) do
    let (opt,_,desc) = List.hd !opts in
    let () =
      ( let () = opt:=true in
	
	let () = print ("Beginning of the Execution using the Option " ^ desc) in
	let i_marshal_c = open_in marshal_filename in 
	let (lst,h_accum,t_accum_1) = 
	  let resultsAccum = if !optDontSL then Number 0 else List [] in
	  if !optDisj then
	    automatic_pairs_incl_loop_unparsed_disj 0 m i_marshal_c 0 (newTimeMap ()) resultsAccum just_this_execution !optDontSL
	  else
	    let line1 = unmarshal i_marshal_c in
	    automatic_pairs_incl_loop_unparsed 0 line1 (m-1) i_marshal_c 0 (newTimeMap ()) resultsAccum just_this_execution !optDontSL in
	
	let triple= (lst,h_accum, t_accum_1) in
	let () = results := (!results)@[(triple,desc)] in
	opt := false) in
    let () = opts := List.tl !opts in
    print ("End of the Execution using the Option " ^ desc)
  done 
  ;
  let () = print ("Beginning of the Execution without any options enabled") in
  let i_marshal_c = open_in marshal_filename in 
  let (lst,h_accum,t_accum_1) = 
    let resultsAccum = if !optDontSL then Number 0 else List [] in
    if !optDisj then 
      automatic_pairs_incl_loop_unparsed_disj 0 m i_marshal_c 0 (newTimeMap ()) resultsAccum just_this_execution !optDontSL
    else
      let line1 = unmarshal i_marshal_c in
      automatic_pairs_incl_loop_unparsed 0 line1 (m-1) i_marshal_c 0 (newTimeMap ()) resultsAccum just_this_execution !optDontSL in
  let () = print ("End of the Execution without any options") in
  
  let triple = (lst,h_accum, t_accum_1) in
  let () = results := (!results)@[(triple,"Using no options")] in
  
  (* Checking if all options found the same equivalence pairs *)
  let (sameAnsr,y) = checkSameResults !results in
  (* The results of using each option *)
  let () = results := (List.sort ~cmp:compare !results) in
  while not(!results=[]) do
    let ((lst,h_accum,t_accum),desc) = List.hd !results in
    let () = print ("\nStats of the Option " ^ desc ^ "\n") in
    let () = automatic_pairs_finish (lst,h_accum,t_accum) (if !optDisj then (m/2) else (m-1)) false in
    results := List.tl !results
  done 
  ; 
  (match (sameAnsr,y) with
      (true,Number n)   ->  
	if !optDontSL then 
	  print ("\nAll different options found success in "^string_of_int n^" pairs")
	else
	  print ("\nAll options found the same pairs ("^string_of_int n^")!")
    | (false,List ls) ->  
      if not(!optDontSL) then 
	(print "\nThe options didn't find the same pairs. Some or all of the divergent pairs are:" ;
	 List.iter (fun (e1,e2) -> (printExp e1); print " and "; (printExp_el e2)) ls )
    | (false,_)       ->  
      if !optDontSL then
	print ("\nThe options didn't find success in the same number of pairs")
    | (true,List _)   ->  raise (Error "unexpected pattern"))
  ; 
  if just_this_execution then Sys.remove marshal_filename

let automatic_pairs_comparePDWithBS_withAllOpts m =
  (* Marshaling of the input lines *)
  let marshal_filename="marshaled_exps_pair_comparePDWithBS_withAllOpts_"^(string_of_float (Unix.time())) in 
  let o_marshal_c = open_out marshal_filename in
  preprocess_marshal_without_parsing o_marshal_c m ;

  optBSKE:=true ; print "** Using the Berry-Sethi method for KAT Expressions **" ; 
  automatic_pairs_compareOpts ~marshal_filename:marshal_filename m ; optBSKE:=false ; 
  
  print "** Using Partial Derivatives **" ;
  automatic_pairs_compareOpts ~marshal_filename:marshal_filename m ; 

  Sys.remove marshal_filename

let automatic_self_comparePDWithBS_withAllOpts m =
  (* Marshaling of the input lines *)
  let marshal_filename="marshaled_exps_self_comparePDWithBS_withAllOpts_"^(string_of_float (Unix.time())) in 
  let o_marshal_c = open_out marshal_filename in
  preprocess_marshal_without_parsing o_marshal_c m ;

  optBSKE:=true ; print "** Using the Berry-Sethi method for KAT Expressions **" ; 
  automatic_self_compareOpts ~marshal_filename:marshal_filename m ; optBSKE:=false ; 
  
  print "** Using Partial Derivatives **" ;
  automatic_self_compareOpts ~marshal_filename:marshal_filename m ; 

  Sys.remove marshal_filename
    
(* m=total number of expressions; *)
let rec automatic_pairs_disj m flagPrintEqExps =
  try 
    let stdinbuf = Lexing.from_channel stdin in
    let marshal_filename = "marshaled_exps_pairs_disj_" ^ (string_of_float (Unix.time())) in 
    let o_marshal_c = open_out marshal_filename in
    let ats_accum = preprocess stdinbuf o_marshal_c m 0 0. in
    let i_marshal_c = open_in marshal_filename in
    let () = loop (m/2) i_marshal_c 0 0 (newTimeMap ()) [] flagPrintEqExps in
    let () = print ("The average cardinality of At was: "^ (string_of_float (ats_accum /. (float (m/2))))) in
    Sys.remove marshal_filename
  with LexerPolish.Eof -> exit 0
(* m=total no. of tests; n=no. of tests performed so far; lst=list of pairs of
   the random exps that were equivalent so far *)
and loop m i_marshal_c n h_accum t_accum lst flagPrintEqExps = 
  if n=m then 
    (* End *)
    (if not (lst=[]) then
	(print ("Summary: I found " ^ string_of_int (List.length lst) ^ " pairs of expressions to be equivalent") ;
	 if flagPrintEqExps then (print ":"; print_list_pairsKE lst) else print "")
     else
	print "Summary: All tests returned False"  ;
     print ("The average cardinality of H was: " ^ string_of_float ((float h_accum)/.(float m))) ;
     print ("The average time taken per test was: "); printTimeMap (divideTimeMapByN m t_accum) ;
     close_in i_marshal_c )
  else
    (* The test *)
(*    let exp1 : MKE.ke = Marshal.from_channel i_marshal_c in
    let exp2 : MKE.ke = Marshal.from_channel i_marshal_c in*)
    let exp1 = unmarshal i_marshal_c in
    let exp2 = unmarshal i_marshal_c in

    let (answer,h,new_t_accum) = 
      match (exp1,exp2) with
	  (ExpPos e1,ExpPos e2) -> 
	    let t_i = Sys.time () in
	    let (a1,a2) = buildPairNFAGSUsingBSKE e1 e2 in
	    let ta = Sys.time() -. t_i in
	    let t_i = Sys.time() in
	    let (answer,h) = equiv (NFAGS a1) (NFAGS a2) in
	    let td = Sys.time() -. t_i in
	    let new_t_accum = incrementTimeMap keyw_time_NFAGSConstr ta t_accum in
	    let new_t_accum = incrementTimeMap keyw_time_DecisionProc td new_t_accum in
	    (answer,h,new_t_accum)
	| (Exp e1,Exp e2)       ->
	  let t_i = Sys.time() in
	  let (answer,h) = equiv (Exp e1) (Exp e2) in
	  let td = Sys.time() -. t_i in
	  let new_t_accum = incrementTimeMap keyw_time_DecisionProc td t_accum in
	  (answer,h,new_t_accum) 
	| _                     ->
	  raise (ERROR "Unexpected pattern found")   in

    let new_h_accum = h_accum+(setHCardinal h) in
    let () = 
      print ("Test: " ^ string_of_int (n+1) ^ " #H: " ^ (string_of_int (setHCardinal h)) ^ " answer: " ^ (string_of_bool answer)) in
    let newLst = if answer=true then ((exp1,exp2)::lst) else lst in
    loop m i_marshal_c (n+1) new_h_accum new_t_accum newLst flagPrintEqExps
and preprocess stdinbuf oc m n ats_accum =
  let rec aux exp1 stdinbuf oc m n ats_accum =
    if n<m then
      let exp = parse_stdinbuf stdinbuf in
      let ats = calculateAtSize exp exp1 in 
      let () = Marshal.to_channel oc exp [] in
      aux exp stdinbuf oc m (n+1) (ats_accum+.ats)
    else 
      let () = close_out oc in
      ats_accum
  in 
  let exp = parse_stdinbuf stdinbuf in
  let () = Marshal.to_channel oc exp [] in
  aux exp stdinbuf oc (m-1) n ats_accum
(*-- End of automatic/no-instruction-printing functions --*)
 
(*------------------- Utility routines -------------------*)

let rec generateEquivalentExps numbExps minLength maxLength minNumbAxiomUses =
  let stdinbuf = Lexing.from_channel stdin in
  let datetime = string_of_float (Unix.time()) in
  let filename_polish = ("equivalent_expressions_polish_" ^ datetime ^ "txt") in
  let filename_infix = ("equivalent_expressions_infix_" ^ datetime ^ "txt") in
  let file_polish = Unix.openfile filename_polish [Unix.O_WRONLY;Unix.O_CREAT] 0o644 in
  let file_infix = Unix.openfile filename_infix [Unix.O_WRONLY;Unix.O_CREAT] 0o644 in
  loop numbExps minLength maxLength minNumbAxiomUses stdinbuf file_polish file_infix
and loop numbExps minLength maxLength minNumbAxiomUses stdinbuf file_polish file_infix = 
  if (numbExps>0) then
    let lexer_line = LexerPolish.line stdinbuf in
    let linebuf = Lexing.from_string lexer_line in
    let exp1 = List.hd (ParserPolish.data LexerPolish.token linebuf) in
    let (exp2,lst) = MRewriteRules.applyRewriteRules exp1 minLength maxLength in
    if (count_ruleUses lst < minNumbAxiomUses) then 
      loop numbExps minLength maxLength minNumbAxiomUses stdinbuf file_polish file_infix
    else 
      let exp2_strPol = MKE.katExp2String_polish exp2 ^ "\n" in
      let _ = Unix.write file_polish lexer_line 0 (String.length lexer_line) in
      let _ = Unix.write file_polish exp2_strPol 0 (String.length exp2_strPol) in
      (* during or after parsing there often is some reduction of the exps, so the following exp will probably not be syntactically equal to the one received from the input *)
      let exp1_strIfx = MKE.katExp2String exp1 ^ "\n" in 
      let exp2_strIfx = MKE.katExp2String exp2 ^ "\n" in
      let _ = Unix.write file_infix exp1_strIfx 0 (String.length exp1_strIfx) in
      let _ = Unix.write file_infix exp2_strIfx 0 (String.length exp2_strIfx) in
      loop (numbExps-1) minLength maxLength minNumbAxiomUses stdinbuf file_polish file_infix

(* Converts a given exp in infix notation to polish notation. *)
let convertFromInf2PolNot () = 
  let rec aux stdinbuf = 
    let lexer_line = Lexer.line stdinbuf in
    if not(lexer_line = "QUIT\n") then  
      let linebuf = Lexing.from_string lexer_line in
      let parsed_infix = List.hd (Parser.data Lexer.token linebuf) in
      let () = MKE.printKE_polish parsed_infix in
      aux stdinbuf
  in let stdinbuf = Lexing.from_channel stdin in
     aux stdinbuf

let convertFromPol2InfNot () = 
  let rec aux stdinbuf = 
    let lexer_line = LexerPolish.line stdinbuf in
    if not(lexer_line = "QUIT\n") then  
      let linebuf = Lexing.from_string lexer_line in
      let parsed_infix = List.hd (ParserPolish.data LexerPolish.token linebuf) in
      let () = MKE.printKE_el parsed_infix in
      aux stdinbuf
  in let stdinbuf = Lexing.from_channel stdin in
     aux stdinbuf

let indexPosition () =
  let rec aux stdinbuf = 
    let lexer_line = LexerPolishPosition.line stdinbuf in
    if not(lexer_line = "QUIT\n") then
      let linebuf = Lexing.from_string lexer_line in
      let parsed_exp =
	List.hd (ParserPolishPosition.data LexerPolishPosition.token linebuf) in
      let () = print_str lexer_line in
      let () = print_str " -> " in
      let () = MKEPos.printKE_polish parsed_exp in
      let () = print_str "\n" in
      aux stdinbuf in
  let stdinbuf = Lexing.from_channel stdin in
  aux stdinbuf

(* n is the total number of exps to be analyzed *)
let countDiffTestsAndActions n notation = 
  let stdinbuf = Lexing.from_channel stdin in
  let parseFun =  
    match notation with 
	"polish" ->  parsePolish
      | "infix"  ->  parseInfix
      | _        ->  raise (Error "Please provide a correct notation option") in
  let rec loop n (accumSig,accumT) = 
    if n=0 then
      (accumSig,accumT)
    else
      let exp = parseFun stdinbuf in
      let (cardSig,cardT) = 
	match exp with 
	    Exp ke  ->  (SetInt.cardinal (buildSigfromKe ke), SetInt.cardinal (buildTfromKe ke))
	  | _       ->  raise (Error "Unexpected pattern")  in
      loop (n-1) ((accumSig+cardSig), (accumT+cardT)) 
  in let (totalSig,totalT) = loop n (0,0) in
     let avgSig = (float totalSig) /. (float n) in
     let avgT = (float totalT) /. (float n) in
     print ("The average number of different actions (i.e., of the cardinality of Σ) per expression is "^string_of_float avgSig) ;
     print ("The average number of different tests (i.e., of the cardinality of T) per expression is "^string_of_float avgT)


(* Helpful routines for digesting the command-line arguments. *)      
let runVerbose str =
  verbose:=true ; printOptsEnabled () ;
  match str with
      "assumptions"      ->  verbose_assumptions ()
    | "pair"             ->  verbose_pair ()
    | "assumptions2"     ->  verbose_assumptions_compare ()
    | _                  ->  raise (ERROR "you must provide a correct option to -v")
let runAutoSelf array =
  printOptsEnabled () ;
  let rec loop array n = 
    if array.(n)="-self" then 
      automatic_self (int_of_string array.(n+1))
    else if array.(n)="-selfIncl1CompareOpts" then
      automatic_self_incl_compareOpts (int_of_string array.(n+1))
    else if array.(n)="-selfIncl2CompareOpts" then
      automatic_self_incl_compareOpts (int_of_string array.(n+1))
    else loop array (n+1) in
  loop array 1
let runAutoPairs array = 
  printOptsEnabled () ;
  let rec loop array n = 
    if array.(n)="-pairs" then 
      automatic_pairs_main (int_of_string array.(n+1)) (bool_of_string array.(n+2))
    else if array.(n)="-pairsCompareOpts" then
      automatic_pairs_compareOpts (int_of_string array.(n+1))
    else if array.(n)="-pairsIncl1CompareOpts" then
      automatic_pairs_incl_compareOpts (int_of_string array.(n+1))
    else if array.(n)="-pairsIncl2CompareOpts" then
      automatic_pairs_incl_compareOpts (int_of_string array.(n+1))
    else loop array (n+1) in
  loop array 1
let runAutoPairsDisj array =
  printOptsEnabled () ;
  let rec loop array n = 
    if array.(n)="-pairsDisj" then 
      automatic_pairs_disj (int_of_string array.(n+1)) (bool_of_string array.(n+2))
    else loop array (n+1) in
  loop array 1

let digestUtils argv = 
  let rec loop argv n =
    if argv.(n)="convertInf2Pol" then convertFromInf2PolNot ()
    else if argv.(n)="convertPol2Inf" then convertFromPol2InfNot ()
    else if argv.(n)="rewrite" then generateEquivalentExps (int_of_string argv.(n+1)) (int_of_string argv.(n+2)) (int_of_string argv.(n+3)) (int_of_string argv.(n+4))
    else if argv.(n)="indexPosition" then indexPosition ()
    else if argv.(n)="comparePDWithBS" then automatic_pairs_comparePDWithBS (int_of_string argv.(n+1))
    else if argv.(n)="comparePDWithBS_withAllOpts" then automatic_pairs_comparePDWithBS_withAllOpts (int_of_string argv.(n+1))
    else if argv.(n)="selfComparePDWithBS_withAllOpts" then automatic_self_comparePDWithBS_withAllOpts (int_of_string argv.(n+1))
    else if argv.(n)="countDiffTestsAndActions" then countDiffTestsAndActions (int_of_string argv.(n+1)) argv.(n+2)
    else if n=Array.length argv then raise (ERROR "wrong option provided in the utils mode")
    else loop argv (n+1) in
  loop argv 0
let digestLanInclusion argv = 
  let () = optBSKE:=true in
  let rec loop argv n = 
    if n=Array.length argv then
      verbose_incl_pair ()
    else 
      if argv.(n)=keyw_naiv then 
	let () = inclV1:=true; inclV2:=false in loop argv (n+1)
      else if argv.(n)=keyw_noup then 
	let () = InclusionCheckerAGS1.optRSubset:=false; InclusionCheckerAGS2.optRSubset:=false in loop argv (n+1)
  in
  loop argv 1
let digestParDer argv = 
  let rec loop argv n = 
    if n=Array.length argv then
      verbose_pair ()
    else 
      if argv.(n)=keyw_noup then 
	let () = EquivalenceCheckerPD.opt2a:=false in loop argv (n+1)
      else if argv.(n)=keyw_asmp1 then 
	verbose_assumptions ()
      else if argv.(n)=keyw_asmp2 then
	verbose_assumptions_compare ()
  in
  loop argv 1
let digestLanEquality argv = 
  let rec loop argv n = 
    if n=Array.length argv then
      verbose_pair ()
    else 
      if argv.(n)=keyw_noup then 
	let () = EquivalenceCheckerAGS.optCongr:=false; InclusionCheckerAGS1.optRSubset:=false; 
	  InclusionCheckerAGS2.optRSubset:=false in loop argv (n+1)
      else if argv.(n)=keyw_prdr then
	let () = optBSKE:=false in digestParDer (subArray argv n)
      else if argv.(n)=keyw_audi then
	let () = CallNFAGSEquivChecker.optIncl2 := false in loop argv (n+1)
  in
  loop argv 1
let digestDecide argv = 
  let rec loop argv n =
    if argv.(n)=keyw_lain then digestLanInclusion (subArray argv n)
    else if argv.(n)=keyw_laeq then digestLanEquality (subArray argv n) 
    else if argv.(n)=keyw_nosat then let () = Module_KatExpressions.optSAT:=false in loop argv (n+1)
    else if n=Array.length argv then raise (ERROR "Decide command provided with a wrong argument.")
    else loop argv (n+1) in
  loop argv 0

let () =
  resetAllOptsToDefault () ;
  Arg.parse [
    ("-decide", Arg.Unit (fun () -> digestDecide (subArray Sys.argv 2)), " 
      \t [-noSATS] -linclusion [-naive] [-noUpTo]     This mode tests the language inclusion represented by two given expressions, by first constructing the NFAGS from the expressions and then applying the language inclusion algorithm to the automata. Three optional parameters may be used: -noSATS disables the SAT solver and all the logical implication problems are resolved via explicit computation of the set of all valorizations (set At); -naive uses the naive approach of the algorithm instead; and -noUpTo disables any up-to technique (see README.txt for more details);
      \t [-noSATS] -lequality [-parDer,-NFAGSDirect] [-noUpTo] [-useAssumptions1,-useAssumptions2]     This mode tests the language equality represented by two given expressions. If neither -parDer nor -NFAGSDirect are used, the language equality is tested by first constructing the NFAGS from the expressions and then applying the language inclusion algorithm to the automata in both ways, since this is the most efficient strategy. Several optional parameters are available: -noSATS disables the SAT solver and all the logical equivalence problems are resolved via explicit computation of the set of all valorizations (set At); -parDer resorts the problem to the computation of the partial derivatives of the expressions (this method is by far the least efficient); -NFAGSDirect tests the language equality directly on the NFAGS instead of testing language inclusion twice; -noUpTo disables any up-to technique; -useAssumptions1 allows you to test the language equality represented by two given expressions in the presence of a set of assumptions (only works when -parDer is being used); and -useAssumptions2 compares two different methods of deciding language equality of expressions in the presence of assumptions (idem);");
    ("-utils",  Arg.Unit (fun () -> digestUtils (subArray Sys.argv 2)), " Use one of the utility tools provided:
      \t 'convertInf2Pol'     Type an expression in infix notation and it will be printed in polish notation. Type QUIT to quit this mode;
      \t 'convertPol2Inf'     Type an expression in polish notation and it will be printed in infix notation. Type QUIT to quit this mode;
      \t 'rewrite' n minLength maxLength minNumbRuleUses        Type expressions (separated by Enter) and for each of them a new expression will be generated, which will be the  result of applying some rewrite rules to the expression. maxLength is the maximum length that should be accepted for an expression after application of the rewrite rules, so expressions that become longer than maxLength after being rewritten will not be considered. minNumbRuleUses is the minimum number of rule applications per expression, so expressions on which substitutions could only be instantiated less than minNumbRuleUses times are not considered. The execution stops when n expressions have been managed to be rewritten. Expressions must be typed in Polish notation. The output will be printed in the files \"equivalent_expressions_polish.txt\" and \"equivalent_expressions_infix.txt\";
      \t 'indexPosition'          Prints each given expressions but with the action symbols marked with the respective position in the expression. Expressions must be typed in polish notation. Terminate with QUIT;
      \t 'countDiffTestsAndActions' n notation  Counts the number of different tests and the number of different actions in the given expressions and returns the average values; n must be the total number of expressions that will be given and notation must be either \"polish\" or \"infix\"");
  ] (fun _ -> ()) (Printf.sprintf "Usage: %s <options>" Sys.argv.(0))
    

(* To following command-line options are commented
    ("-self", Arg.Unit (fun () -> runAutoSelf Sys.argv), "n \t If you want to provide n expressions (separated by Enter) and test the equivalence of each expression with itself. For each test, the result will be printed to the output. After the n expressions have been tested, some statistical results will be printed. Expressions must be typed in Polish notation;");
    ("-pairs", Arg.Unit (fun () -> runAutoPairs Sys.argv), "n flagPrint \t Provide n expressions and the test will be performed as follows: the 1st expression will be tested with the 2nd, the 2nd with the 3rd, the 3rd with the 4th, ..., the (n-1)th with the nth, making a total of (n-1) tests. For each test, the result will be printed. After all tests have been performed, some statistical results will be printed. flagPrint must be \"true\" in order for all pairs of expressions found to be equivalent to be printed at the end of the execution. Expressions must be typed in Polish notation and separated by ENTER;");
    ("-pairsCompareOpts", Arg.Unit (fun () -> runAutoPairs Sys.argv), "The same as -pairs but runs the tests using each of the optimizations separatedly;");
    ("-pairsIncl1CompareOpts", 
     Arg.Unit (fun () -> let () = inclV1:=true; inclV2:=false in runAutoPairs Sys.argv), 
     "Test language inclusion using the naive approach;");
    ("-pairsIncl2CompareOpts", 
     Arg.Unit (fun () -> let () = inclV2:=true; inclV1:=false in runAutoPairs Sys.argv), 
     "Test language inclusion using an approach which is more efficient in most cases;");
    ("-selfIncl1CompareOpts", 
     Arg.Unit (fun () -> let () = inclV1:=true; inclV2:=false in runAutoSelf Sys.argv), 
     "Test language inclusion between each expression and itself using the naive approach, but runs once separatedly for each optimization available;");
    ("-selfIncl2CompareOpts", 
     Arg.Unit (fun () -> let () = inclV2:=true; inclV1:=false in runAutoSelf Sys.argv), 
     "Test language inclusion between each expression and itlsef, but runs once separatedly for each optimization available");
    ("-pairsDisj", Arg.Unit (fun () -> runAutoPairsDisj Sys.argv), "n flagPrint \t Provide n expressions and the test will be performed as follows: the 1st will be tested with the 2nd, the 3rd with the 4th, etc. n must be even (and the number of tests performed is thus n/2) otherwise the nth expression given will not be used. For each test, the result will be printed. After all tests have been performed, some statistical results will be printed as well. flagPrint must be \"true\" in order for all pairs of expressions found to be equivalent to be printed at the end of the execution. Expressions must be typed in Polish notation (for more info see README.txt) and separated by ENTER;");
    ("-inclV1", Arg.Unit (fun () -> inclV1:=true; inclV2:=false), "");
    ("-inclV2", Arg.Unit (fun () -> inclV2:=true; inclV1:=false), "");
    ("-optDisj", Arg.Unit (fun () -> optDisj:=true), "");
    ("-optRSubset", Arg.Unit (fun () -> EquivalenceCheckerAGS.optRSubset:=true ), "");
*)
