
open RegstabBridge
open Module_SetBuilder

exception ERROR of string
exception Found_Zero

let debugging = ref false
let optSAT = ref true
let opts = 
  [
    (optSAT, true, "SAT - Uses a SAT Solver for satisfiability problems")
  ]
let print str = print_endline str ; flush stdout
let print_str str = print_string str ; flush stdout

(* a module for integers *)
module rec MInt : sig
  (* Warning: never compare two sets by doing x=y; only Set.equal x y actually 
     compares the content of the sets! *)
  type t = int
  val setIntFromList : int list -> SetInt.t
  val unions : SetInt.t list -> SetInt.t
  val setInt2String : SetInt.t -> string
  val printListPairs : (SetInt.t*SetInt.t) list -> unit
  val printSetInt : SetInt.t -> unit
end
  = struct
      
    type t = int

    (* 'unions lst' recursively makes union of all the sets in list 'lst' *)
    let rec unions lst =
      match lst with
	  []		->  SetInt.empty
	| l::ls		->  SetInt.union l (unions ls) ;;
    
    let rec setIntFromList lst =
      match lst with
	  []       ->  SetInt.empty
	| l::ls    ->  SetInt.add l (setIntFromList ls)

    let setInt2String set = 
      let rec aux lst = match lst with
	  []       ->  ""
	| [i]      ->  (string_of_int i)
	| (i::is)  ->  (string_of_int i)^", " ^ (aux is)
      in (aux (SetInt.elements set))

    let printListPairs ls = List.iter (fun (x,y) -> MInt.printSetInt x; MInt.printSetInt y) ls

    let printSetInt set = print_string (setInt2String set)

  end

(* a module for Kat expressions *)
module rec MKE : sig 
  type index = int
  type be = Zero               (* be are boolean exps *)
	  | One
	  | Test of index
	  | NegBexp of be
	  | Conj of SetBexps.t         (* conjunction *)
	  | Disj of SetBexps.t         (* disjunction *)
  type ke = Prog of index
	  | Bexp of be
	  | Star of ke
	  | Conc of ke list
	  | Or of SetKE.t 
  type atom = Lst of bool list | EmptyAtom
  type nf = Final of SetKE.t                     (* type to be used when computing *)
	    | NonFinal of SetKE.t * (ke list)    (* the normal form of an exp *) 
  (* aliases *)
  type formula = RegstabBridge.formula
  type polarity = RegstabBridge.polarity
  type witness = RegstabBridge.witness
  
  val areEquiv : be -> be -> bool
  val atIsEmpty : atom list -> bool
  val atom2String : atom -> string
  val debugAtom : atom -> unit
  val debugAt : atom list -> unit
  val debugAtomNl : atom -> unit
  val debugT' : int list -> unit
  val e : atom -> int list -> ke -> bool
  val eBexp : atom -> int list -> be -> bool
  val eAll : ?at:(atom list) -> ?t':(int list) -> ke -> ke -> bool
  val eAllBexps : ?at:(atom list) -> ?t':(int list) -> be -> be -> bool
  val evaluate : atom -> int list -> be -> bool
  val implies : be -> be -> bool
  val isAlwaysFalse : be -> bool
  val isATautology : be -> bool
  val setKEfromList : ke list -> SetKE.t
  val setBexpsfromList : be list -> SetBexps.t
  val buildConcat : ke list -> ke
  val buildOr : ke list -> ke
  val buildDisj : be list -> be
  val buildConj : be list -> be
  val buildPlus : ke list -> ke
  val buildTimes : ke list -> ke
  val buildNeg : be -> be
  val simplify : ke -> ke
  val simplifyBexp : be -> be
  val convertKe2Be : ke -> be
  val katExp2RegstabFormula : ke -> RegstabBridge.formula
  val katExp2String : ke -> string
  val katExp2String_polish : ke -> string
  val bexp2RegstabFormula : be -> RegstabBridge.formula
  val bexp2String : be -> string
  val bexp2String_polish : be -> string
  val katExp2StringPar : ke -> string
  val bexp2StringPar : be -> string
  val printKE : ke -> unit
  val printKE_el : ke -> unit
  val printKEpar_el : ke -> unit
  val printKE_polish : ke -> unit
  val printBexp : be -> unit
  val printBexp_nnl : be -> unit
  val printSetKE : SetKE.t -> unit
  val setKe2String : SetKE.t -> string
  val t'2String : int list -> string
  val flattenConcats : ke -> ke
  val concat : ke -> ke -> ke
  val buildTfromKe : ke -> SetInt.t
  val buildTfromBe : be -> SetInt.t  
  val buildTfromAssumptions : (be * be) list -> SetInt.t
  val buildT : ((int * (be * be)) list * ((be * be) list)) -> ke -> ke -> int list
  val buildSigma : ke -> ke -> SetKE.t
  val buildSigmaFromKe : ke -> SetInt.t
  val buildAt : int -> atom list 
  val expandT : int list -> int list
  val getAtAndTp : ke -> ke -> (atom list * int list)
  val getAtAndTpFromBexp : be -> (atom list * int list)
  val getAtAndTpFromBexpList : be list -> (atom list * int list)
  val getAtAndTpFrom2Bexps : be -> be -> (atom list * int list)
  val getTAndTpFromBexp : be -> (int list * int list)
  val setBexpsMap : (be -> be) -> SetBexps.t -> SetBexps.t 
  val sizeKE : ke -> int
  val sizeBexp : be -> int
  val sizeAt : ke -> ke -> float
  val unions : SetKE.t list -> SetKE.t
  val keEq : ke -> ke -> bool
  val bexpEq : be -> be -> bool
  val countTestsInKe : ke -> int
  val countTestsInBexp : be -> int
  val map : (ke -> ke) -> SetKE.t -> SetKE.t

end 
  = struct
    type index = int
    type be = Zero               (* be are boolean exps *)
	    | One
	    | Test of index
	    | NegBexp of be
	    | Conj of SetBexps.t         (* conjunction *)
	    | Disj of SetBexps.t         (* disjunction *)
    type ke = Prog of index
	    | Bexp of be
	    | Star of ke
	    | Conc of ke list
	    | Or of SetKE.t 
    type atom = Lst of bool list | EmptyAtom
    type nf = Final of SetKE.t                   (* type to be used when computing *)
	    | NonFinal of SetKE.t * (ke list)    (* the normal form of an exp *) 

    (* aliases *)
    type formula = RegstabBridge.formula
    type witness = RegstabBridge.witness
    type polarity = RegstabBridge.polarity

    let fromIntList2IntSet = MInt.setIntFromList

    let simplify_conj = true
    let simplify_disj = true
      
    let atom2String atom = 
      match atom with
	  EmptyAtom  ->  ""
	| Lst ls     ->  "[" ^ (List.fold_left (fun x y -> x ^ ((string_of_bool y)^",")) "" ls) ^ "]"

    let at2String at = "[" ^ (List.fold_left (fun x y -> x ^ (atom2String y)^",") "" at) ^ "]"

    let debugAtom atom = print_string (atom2String atom)

    let debugAtomNl atom = print_endline (atom2String atom)

    let debugAt at = print_string (at2String at)

    let t'2String t' = "[" ^ (List.fold_left (fun x y -> x ^ ((string_of_int y)^",")) "" t') ^ "]"
      
    let debugT' t' = print_string (t'2String t')

    let setBexpsMap foo set = SetBexps.fold (fun x -> SetBexps.add (foo x)) set SetBexps.empty
		
    let rec setKEfromList lst =
      match lst with
	  []       -> SetKE.empty
	| l::ls    -> SetKE.add l (setKEfromList ls)
	  
    let rec setBexpsfromList lst =
      match lst with
	  []       -> SetBexps.empty
	| l::ls    -> SetBexps.add l (setBexpsfromList ls)

    let rec bexp2String a = 
      match a with
	  Zero               ->  "Zero"
	| One                ->  "One"
	| Test i             ->  "t"^(string_of_int i)
	| NegBexp (Test i)   ->  "¬t"^(string_of_int i)
	| NegBexp Zero       ->  "¬Zero"
	| NegBexp One        ->  "¬One"
	| NegBexp b          ->  "¬" ^ (bexp2String b)
	| Conj set   ->  let rec aux = function
	                    [l]    ->  (bexp2String l)
	                  | l::ls  ->  (bexp2String l) ^ (aux ls)
			  | []     ->  raise (ERROR "function bexp2String received Conj with empty set")
			 in "(" ^ (aux (SetBexps.elements set)) ^ ")"
	| Disj set   ->  let rec aux = function
	                    [l]    ->  (bexp2String l)
	                  | l::ls  ->  (bexp2String l) ^ "+" ^ (aux ls)
			  | []     ->  raise (ERROR "function bexp2String received Disj with empty set")
			 in "(" ^ (aux (SetBexps.elements set)) ^ ")"

    let rec bexp2StringPar a = 
      match a with
	  Zero              ->  "(Zero)"
	| One               ->  "(One)"
	| Test i            ->  "t" ^ (string_of_int i)
	| NegBexp (Test i)  ->  "(¬(t" ^ (string_of_int i) ^ ")"
	| NegBexp b         ->  "(¬(" ^ (bexp2StringPar b) ^ "))"
	| Conj set          ->  let rec aux = function
	                          [l]    ->  (bexp2StringPar l)
	                        | l::ls  ->  (bexp2StringPar l) ^ (aux ls)
				| []     ->  raise (ERROR "function bexp2String received Conj with empty set")
				in "(" ^ (aux (SetBexps.elements set)) ^ ")"
	| Disj set          ->  let rec aux = function
	                          [l]    ->  (bexp2StringPar l)
	                        | l::ls  ->  (bexp2StringPar l) ^ "+" ^ (aux ls)
				| []     ->  raise (ERROR "function bexp2String received Disj with empty set")
			 in "(" ^ (aux (SetBexps.elements set)) ^ ")"
	  
    let rec katExp2String exp =
      match exp with
	  Prog i     ->  "p"^(string_of_int i)
	| Bexp b     ->  bexp2String b
	| Star exp   ->  "(" ^ (katExp2String exp) ^ ")*"
	| Conc lst   ->  let rec aux = function  
	                    [l]    ->  katExp2String l
	                  | l::ls  ->  (katExp2String l) ^ (aux ls)
			  | []     ->  raise (ERROR "aux in katExp2String.Conc received []")
			 in "(" ^ (aux lst) ^ ")"
	| Or set     ->  let rec printAux = function
	                    []     -> ""
	                  | l::ls  -> "+" ^ (katExp2String l) ^ (printAux ls)
			  in let set_elements = SetKE.elements set
			     in if (set_elements == []) then (raise (ERROR "function printAux in MKEPos.katExp2String received an empty set")) else "(" ^ (katExp2String (List.hd set_elements)) ^ (printAux (List.tl set_elements)) ^ ")" ;;

    let rec katExp2StringPar exp =
      match exp with
	  Prog i     ->  "(p"^(string_of_int i)^")"
	| Bexp b     ->  bexp2StringPar b
	| Star exp   ->  "(" ^ (katExp2StringPar exp) ^ ")*"
	| Conc lst   ->  let rec aux = function  
	                    [l]    ->  katExp2StringPar l
	                  | l::ls  ->  (katExp2StringPar l) ^ (aux ls)
			  | []     ->  raise (ERROR "aux in katExp2StringPar received []")
			 in "(" ^ (aux lst) ^ ")"
	| Or set     ->  let rec printAux = function
	                    []     -> ""
	                  | l::ls  -> "+" ^ (katExp2StringPar l) ^ (printAux ls)
			 in let set_elements = SetKE.elements set
			    in if (set_elements == []) then (raise (ERROR "function printAux in katExp2String received an empty set")) else "(" ^ (katExp2StringPar (List.hd set_elements)) ^ (printAux (List.tl set_elements)) ^ ")" ;;

    let printKE exp = print_str (katExp2String exp)
    let printKE_el exp = print (katExp2String exp)

    let printKEpar_el exp = print_endline (katExp2StringPar exp)
    
    let rec bexp2String_polish a = 
      match a with
	  Zero               ->  "Zero"
	| One                ->  "One"
	| Test i             ->  "t" ^ (string_of_int i)
	| NegBexp (Test i)   ->  "¬t" ^ (string_of_int i)
	| NegBexp b          ->  "¬" ^ (bexp2String_polish b)
	| Conj set   ->  let rec aux = function
	                    [l]    ->  (bexp2String_polish l)
	                  | l::ls  ->  "." ^ (bexp2String_polish l) ^ (aux ls)
			  | []     ->  raise (ERROR "function bexp2String_polish received Conj with empty set")
			 in let set_elements = SetBexps.elements set in
			     if (List.length set_elements == 1) then (bexp2String_polish (List.hd set_elements)) else (aux set_elements)
	| Disj set   ->  let rec aux = function
	                    [l]    ->  bexp2String_polish l
	                  | l::ls  ->  "+" ^ (bexp2String_polish l) ^ (aux ls)
			  | []     ->  raise (ERROR "function bexp2String_polish received Disj with empty set")
			 in let set_elements = SetBexps.elements set in
			     if (List.length set_elements == 1) then (bexp2String_polish (List.hd set_elements)) else (aux set_elements)
  
    let rec katExp2String_polish exp =
      match exp with
	  Prog i     ->  "p"^(string_of_int i)
	| Bexp b     ->  bexp2String_polish b
	| Star exp   ->  "*" ^ (katExp2String_polish exp)
	| Conc lst   ->  let rec aux = function  
	                    [l]    ->  katExp2String_polish l
	                  | l::ls  ->  "." ^ (katExp2String_polish l) ^ (aux ls)
			  | []     ->  raise (ERROR "aux in katExp2String_polish.Conc received []")
			 in if (List.length lst == 1) then (katExp2String_polish (List.hd lst)) else aux lst
	| Or set     ->  let rec printAux = function
	                    []     -> raise (ERROR "printAux in katExp2String_polish.Or received []")
			  | [l]    ->  katExp2String_polish l
	                  | l::ls  -> "+" ^ (katExp2String_polish l) ^ (printAux ls)
			  in let set_elements = SetKE.elements set in
			     if (List.length set_elements == 1) then (katExp2String_polish (List.hd set_elements)) else (printAux set_elements)
        			 
    let printKE_polish exp = print_endline (katExp2String_polish exp)

    let printBexp bexp = print_endline (bexp2String bexp)
    let printBexp_nnl bexp = print_string ((bexp2String bexp) ^ " ")

    let setKe2String set =
      let rec loop set = 
	if (SetKE.is_empty set) then "<this set is empty>"
	else if (SetKE.cardinal set)=1 then
	  let x = SetKE.min_elt set in
	  (katExp2String x)
	else
	  let x = SetKE.min_elt set in
	  (katExp2String x) ^ ", " ^ (loop (SetKE.remove x set)) in
      "{" ^ (loop set) ^ "}"

    let printSetKE set = print_endline (setKe2String set)

    let rec buildTfromBe be = 
      match be with
	  Test i                 ->  SetInt.singleton(i)
	| NegBexp b              ->  buildTfromBe b
	| (Conj set | Disj set)  ->  MInt.unions (List.map buildTfromBe (SetBexps.elements set))
	| (Zero|One)             ->  SetInt.empty

    let rec buildTfromAssumptions lst =
      match lst with
	  [(be0,be1)]      ->   SetInt.union (buildTfromBe be0) (buildTfromBe be1)
	| (be0,be1)::ls    ->   SetInt.union (SetInt.union (buildTfromBe be0) (buildTfromBe be1)) (buildTfromAssumptions ls)
       	| []               ->   raise (ERROR "buildTfromAssumptions received []")
	  
    let rec buildTfromKe exp =
    match exp with
	  Bexp b    ->  buildTfromBe b
	| Prog _    ->  SetInt.empty
	| Star e    ->  buildTfromKe e
	| Conc ls   ->  MInt.unions (List.map buildTfromKe ls)
	| Or set    ->  MInt.unions (List.map buildTfromKe (SetKE.elements set))

    (* returns  the alphabet T *)
    let buildT (assumpsEq,assumpsLeq) exp1 exp2 =
      let (_,lsb) = List.split assumpsEq in
      SetInt.elements (SetInt.union (buildTfromAssumptions (lsb@assumpsLeq)) (SetInt.union (buildTfromKe exp1) (buildTfromKe exp2)))
	
    let rec buildAt n =
      if n==0 then []
      else [? List : (Lst (l::ls)) | l <- List:[true;false]; ls <- List:(aux (buildAt (n-1))) ?]
    and aux at = 
      match at with
	  [Lst l]        ->   [l]
	| ((Lst l)::ls)  ->   l::(aux ls)
	| otherwise      ->   [[]]

    (* transforms t, an int list which contains the numbers of the tests 
       in an exp, to t', a different int list in 
       which every index i contains the position in atom that is associated 
       to Test i (it will be -1 if non applicable);
       eg: if t=[2;4;5], expandT t = [-1;-1;0;-1;1;2] *)
    let expandT (lst: int list) =
      let rec aux lst n m = 
	match lst with
	    (l::ls)   ->  if l=m then n::(aux ls (n+1) (m+1)) else (-1)::(aux lst (n) (m+1))
	  | []        ->  []
      in aux lst 0 0

    let getAtAndTp exp1 exp2 =
      let t = SetInt.elements (SetInt.union (buildTfromKe exp1) (buildTfromKe exp2)) in
      let t' = expandT t in
      let at = if t=[] then [MKE.EmptyAtom] else buildAt (List.length t) 
      in (at,t')

    let getAtAndTpFromKexp exp =
      let t = SetInt.elements (buildTfromKe exp) in
      let t' = expandT t in
      let at = if t=[] then [MKE.EmptyAtom] else buildAt (List.length t) 
      in (at,t')

    let getTAndTpFromBexp bexp = 
      let t = SetInt.elements (buildTfromBe bexp) in
      (t,expandT t)

    let getAtAndTpFromBexp bexp =
      let (t,t') = getTAndTpFromBexp bexp in
      let at = if t=[] then [MKE.EmptyAtom] else buildAt (List.length t) in
      (at,t')

    let getAtAndTpFromBexpList bs = 
      let tSet = MInt.unions (List.map buildTfromBe bs) in
      let t = SetInt.elements tSet in
      let t' = expandT t in
      let at = if t=[] then [MKE.EmptyAtom] else buildAt (List.length t) in
      let () = if !debugging then 
	  print_endline ("getAtAndTpFromBexpList built at="^at2String at^" and t'="^t'2String t') in
      (at,t')

    let getAtAndTpFrom2Bexps bexp1 bexp2 =
      let t = SetInt.elements (SetInt.union (buildTfromBe bexp1) (buildTfromBe bexp2)) in
      let t' = expandT t in
      let at = if t=[] then [MKE.EmptyAtom] else buildAt (List.length t) in
      let () = if !debugging then 
	  print_endline ("getAtAndTpFrom2Bexps built at="^at2String at^" and t'="^t'2String t'
			 ^" based on the bexp1="^bexp2String bexp1^" and bexp2="^bexp2String bexp2) in
      (at,t')

    let atIsEmpty at = 
      at = [EmptyAtom]
	  
    let rec bexp2RegstabFormula bexp = 
      let bottom = RegstabBridge.bottom in
      let top = RegstabBridge.top in
      let buildLiteral = RegstabBridge.buildLiteral in
      let buildOpposite = RegstabBridge.buildOpposite in
      let buildAnd = RegstabBridge.buildAnd in
      let buildOr = RegstabBridge.buildOr in 
      let loop = bexp2RegstabFormula in
      match bexp with
	Zero             ->  bottom
      | One              ->  top
      | Test i           ->  buildLiteral true "p" i  (* I always use "p" but it is only a matter of choice *)
      | NegBexp Zero     ->  top
      | NegBexp One      ->  bottom
      | NegBexp (Test i) ->  buildLiteral false "p" i
      | NegBexp b        ->  buildOpposite (bexp2RegstabFormula b)
      | Conj set when (SetBexps.cardinal set)=1  ->  loop (SetBexps.min_elt set)
      | Conj set         ->  let min_elt = SetBexps.min_elt set in 
			     SetBexps.fold (fun x y -> buildAnd y (loop x)) (SetBexps.remove min_elt set) (loop min_elt)
      | Disj set when (SetBexps.cardinal set)=1  ->  loop (SetBexps.min_elt set)
      | Disj set         ->  let min_elt = SetBexps.min_elt set in
			     SetBexps.fold (fun x y -> buildOr y (loop x)) (SetBexps.remove min_elt set) (loop min_elt)
			       
    let rec katExp2RegstabFormula exp = 
      let bottom = RegstabBridge.bottom in
      let top = RegstabBridge.top in
      let buildAnds = RegstabBridge.buildAnds in
      let buildOr = RegstabBridge.buildOr in 
      let loop = katExp2RegstabFormula in
      match exp with
	Prog _           ->  bottom
      | Bexp bexp        ->  bexp2RegstabFormula bexp
      | Star _           ->  top
      | Conc [h]         ->  loop h
      | Conc ls          ->  
	if List.exists (fun a -> match a with Prog _ -> true | _ -> false) ls then bottom
	else let ls' = List.map loop ls in
	     buildAnds ls'
      | Or set when (SetKE.cardinal set)=1  ->  loop (SetKE.min_elt set)
      | Or set           ->  let min_elt = SetKE.min_elt set in
			     SetKE.fold (fun x y -> buildOr y (loop x)) (SetKE.remove min_elt set) (loop min_elt)

    let rec evaluate atom t' b =
      match b with
	  MKE.Zero       ->  false
	| MKE.One        ->  true
	| MKE.Test i     ->  
	  (match atom with   
	      Lst ls    -> (
		try (List.nth ls (
		  try (List.nth t' i) 
		  with Failure("nth") -> raise (ERROR ("evaluate (error1) couldn't find the position " ^ string_of_int i ^ " in t'=" ^ t'2String t' ^ " when evaluating the expression " ^ MKE.bexp2String b)) )) 
		with Failure("nth") -> raise (ERROR ("evaluate (error2) couldn't find the " ^ string_of_int (List.nth t' i) ^"th position in " ^ atom2String atom ^ "; that position corresponds to the position of " ^ string_of_int i ^ " (from Test " ^ string_of_int i ^ ") in t'=" ^ t'2String t' ^ " when evaluating the expression " ^ MKE.bexp2String b)))
	    | EmptyAtom  ->  false)
	| MKE.NegBexp b  ->  not(evaluate atom t' b)
	| MKE.Conj set   ->  SetBexps.for_all (evaluate atom t') set
	| MKE.Disj set   ->  SetBexps.exists (evaluate atom t') set  
    
    (* returns true if be1 -> be2 and false otherwise *)
    let implies be1 be2 =
      let () = if !debugging then print ("I'm going to solve an implication: "^(bexp2String be1)
					 ^" and "^(bexp2String be2)) in
      if be1=MKE.Zero then
	true
      else if be2=MKE.One then
	true
      else if !optSAT then
	let () = if !debugging then print ("I'm going to solve an implication using SAT solver ") in
	let answer = RegstabBridge.implies (bexp2RegstabFormula be1) (bexp2RegstabFormula be2) in
	answer
      else
	let () = if !debugging then print ("I'm going to solve an implication using At set ") in
	let (at,t') = getAtAndTpFrom2Bexps be1 be2 in
	List.for_all (fun atom -> if evaluate atom t' be1 then evaluate atom t' be2 else true) at

    let isAlwaysFalse be =  
      if !optSAT then
	RegstabBridge.isUnsat (bexp2RegstabFormula be)
      else
	let (at,t') = getAtAndTpFromBexp be in
	let eval t' b atom = evaluate atom t' b in
	not(List.exists (eval t' be) at)

    let isATautology be =  
      if !optSAT then
	RegstabBridge.isATautology (bexp2RegstabFormula be)
      else
	let (at,t') = getAtAndTpFromBexp be in
	let eval t' b atom = evaluate atom t' b in
	List.for_all (eval t' be) at
	
    (* 'e atom exp' returns true if atom <= exp and false otherwise *)
    let rec e atom t' exp =
      match exp with
	  MKE.Prog _     ->  false
	| MKE.Bexp b     ->  evaluate atom t' b
	| MKE.Star _     ->  true
	| MKE.Or set     ->  SetKE.exists (e atom t') set
	| MKE.Conc lst   ->  List.for_all (e atom t') lst

    let eBexp atom t' be = 
      let () = if !debugging then print_endline "i'm in eBexp" in
      evaluate atom t' be

    let rec eAll ?(at=[(MKE.Lst [])]) ?(t'=[]) exp1 exp2 = 
      match (exp1,exp2) with
	(MKE.Bexp One, MKE.Bexp One)   ->  true
      | (MKE.Bexp Zero, MKE.Bexp Zero) ->  true 
      | (MKE.Bexp Zero, MKE.Bexp One)  ->  false
      | (MKE.Bexp One, MKE.Bexp Zero)  ->  false
      | otherwise                      ->  
	if !optSAT then
	  let () = 
	    if !debugging then print "I'm going to solve a satisfiability problem using the SAT Solver" in
	  RegstabBridge.areEquiv (katExp2RegstabFormula exp1) (katExp2RegstabFormula exp2)
	else
	  let () = 
	    if !debugging then print "I'm going to solve a satisfiability problem using the At set" in
	  let (at,t') = if at=[(MKE.Lst [])] && t'=[] then MKE.getAtAndTp exp1 exp2 else (at,t') in
	  loop at t' exp1 exp2
    and loop at t' exp1 exp2 =
      match at with
	atom::ls   ->  if (e atom t' exp1)=(e atom t' exp2) then loop ls t' exp1 exp2 else false
      | otherwise  ->  true

    let rec eAllBexps ?(at=[(MKE.Lst [])]) ?(t'=[]) bexp1 bexp2 = 
      match (bexp1,bexp2) with
(*	(x, y) when x=y  ->  true     *)
      | (Zero, One)      ->  false
      | (One, Zero)      ->  false
      | otherwise        ->  
	if !optSAT then
	  let () = if !debugging then print "I'm going to solve an equivalence problem using the SAT Solver" in
	  RegstabBridge.areEquiv (bexp2RegstabFormula bexp1) (bexp2RegstabFormula bexp2)
	else
	  let () = if !debugging then print "I'm going to solve an equivalence problem using the At set" in
	  let (at,t') = if at=[(MKE.Lst [])] && t'=[] then MKE.getAtAndTpFrom2Bexps bexp1 bexp2 else (at,t') in
	  loop at t' bexp1 bexp2
    and loop at t' bexp1 bexp2 =
      match at with
	  atom::ls   ->  (evaluate atom t' bexp1)=(evaluate atom t' bexp2) 
	                 && loop ls t' bexp1 bexp2 
	| otherwise  ->  true 

    let areEquiv be1 be2 = 
      let () = if !debugging then print ("I'm going to solve an equivalence: "^(bexp2String be1)
					 ^" and "^(bexp2String be2)) in
      eAllBexps be1 be2
	
    (* 'concat exp1 exp2' concats the exp 'exp1' to the end of the exp 'exp2' *)
    let concat exp1 exp2 =
      match (exp1, exp2) with
	  (exp1, Bexp One)	->  exp1
	| (Bexp One, exp2)	->  exp2
	| (_, Bexp Zero)	->  Bexp Zero
	| (Bexp Zero, _)	->  Bexp Zero
	| (Conc ls1, Conc ls2)  ->  (Conc (ls2 @ ls1))
	| (exp1, Conc ls2)	->  (Conc (ls2 @ [exp1]))
	| (Conc ls1, exp2)	->  (Conc ([exp2] @ ls1))
	| (exp1, exp2)		->  (Conc ([exp2; exp1])) ;;

    let rec flattenConcats exp =
      let rec flattenConcats' lst = 
	match lst with
	  | (Conc lst')::ls   ->  (flattenConcats' lst')@(flattenConcats' ls)
	  | (Or set)::ls      ->  let lst' = SetKE.elements set in
				  Or (setKEfromList (List.map flattenConcats lst'))::(flattenConcats' ls)
	  | []                ->  []
	  | (Bexp Zero)::ls   ->  (Bexp Zero)::(flattenConcats' ls)
	  | (l::ls)           ->  l::(flattenConcats' ls)
      in match exp with
	  Conc lst    ->  (Conc (flattenConcats' lst))
	| Star exp'   ->  (Star (flattenConcats exp'))
	| Or set      ->  let lst = SetKE.elements set
			  in Or (setKEfromList (List.map flattenConcats lst))
	| otherwise   ->  exp
	
    let convertKe2Be exp =
      match exp with
	Bexp bexp  ->  bexp
      | otherwise  ->  raise (ERROR "convertKe2Be received a non Bexp expression")

    (* returns a list containing all symbols of Sigma in exp1 or exp2 *)
    let buildSigma exp1 exp2 =
      let rec buildAux exp =
	match exp with
	    Bexp b  ->  SetKE.empty
	  | Prog i  ->  SetKE.singleton (MKE.Prog i)
	  | Star e  ->  buildAux e
	  | Conc ls ->  MKE.unions (List.map buildAux ls)
	  | Or set  ->  MKE.unions (List.map buildAux (SetKE.elements set))
      in SetKE.union (buildAux exp1) (buildAux exp2)

    (* returns a list of ints corresponding to the indexes of the actions in exp *)
    let rec buildSigmaFromKe exp = 
      match exp with
	  Bexp _  ->  SetInt.empty
	| Prog i  ->  SetInt.singleton i
	| Star e  ->  buildSigmaFromKe e
	| Conc ls ->  MInt.unions (List.map buildSigmaFromKe ls)
	| Or set  ->  MInt.unions (List.map buildSigmaFromKe (SetKE.elements set))

    let buildConcat ls =
      let simplify = true in   (* turns to 0 all concats of the form x..0..y *)
      match ls with
	  [exp]      ->  exp
	| otherwise  ->  
	  if simplify then
	    let rec loop ls =
	      match ls with
		| (Bexp Zero)::tl       ->  raise Found_Zero 
		| [Bexp One]            ->  ls
		| (Bexp One)::tl        ->  loop tl
		| (Bexp r)::tl          ->  
		  (*      if isAlwaysFalse r then      Commented, since it's probably too heavy to compute..
			  raise Found_Zero
		  else  *)
		  (Bexp r)::(loop tl)
		| exp::tl               ->  exp::(loop tl)
		| []                    ->  []
	    in match (try loop ls with Found_Zero -> [Bexp Zero]) with
		[h]    ->  h
	      |	h::tl  ->  MKE.Conc (h::tl)
	      | []     ->  raise (ERROR "buildConcat received []")
	  else (MKE.Conc ls)
	    
    let opposite be = 
      match be with
	  NegBexp be' ->  be'
	| _           ->  NegBexp be

    let rec buildConj lst =
      match lst with
	[]         ->  raise (ERROR "buildConj received [] as argument ")
      | otherwise  ->  let foo lst =
			 let rec aux lst' =
			   if simplify_conj then
			     match lst' with
				 (Conj set)::ls ->
				   if List.mem (NegBexp (Conj set)) ls then 
				     raise Found_Zero 
				   else
				     let set' = aux ls in
				     SetBexps.union set set'
			       | (Zero)::ls       ->  raise Found_Zero 
			       | [One]            ->  SetBexps.singleton One
			       | (One)::ls        ->  aux ls
			       | r::ls            ->  
			(*	 if List.mem (opposite r) ls then   Commented, since it's probably too heavy...
				   raise Found_Zero
				 else
				 if isAlwaysFalse r then    
				     raise Found_Zero 
				   else *)
				     let set' = aux ls in
				     SetBexps.add r set'
			       | []               ->  SetBexps.empty
			   else 
			     match lst' with
			       | r::ls   ->  let set' = aux ls in
					     SetBexps.add r set'
			       | []      ->  SetBexps.empty
			 in try (aux lst) with
			     Found_Zero  ->  SetBexps.singleton Zero
			   | _           ->  aux lst
		       in buildConj_set (foo lst)
    and buildConj_set set = 
      match (SetBexps.cardinal set) with
	  1  ->  SetBexps.min_elt set
	| 0  ->  raise (ERROR "buildConj_set received an empty set")
	| n  ->  Conj set
	  
    let rec buildOr lst =
      match lst with
	  []         ->  raise (ERROR "buildOr received [] as argument ")
	| [l]        ->  l
	| otherwise  ->  let rec aux lst' =
			   match lst' with
			       (Or set)::ls    ->  let set' = aux ls in
						   SetKE.union set set'
			     | [Bexp Zero]     ->  SetKE.singleton (Bexp Zero)
			     | (Bexp Zero)::ls ->  aux ls
			     | r::ls           ->  let set' = aux ls in
						   (SetKE.union (SetKE.singleton r) set')
			     | []              ->  SetKE.empty
			 in let newSet = aux lst in
			    buildOr_set newSet
    and buildOr_set set = 
      match (SetKE.cardinal set) with
	  1  ->  SetKE.min_elt set
	| 0  ->  raise (ERROR "buildOr_set received an empty set")
	| n  ->  Or set
	  
    let buildDisj lst =
      match lst with
	  []         ->  raise (ERROR "buildDisj received [] as argument ")
	| otherwise  ->  let rec aux lst' =
			   if simplify_disj then
			     match lst' with
				 (Disj set)::ls  ->  let set' = aux ls 
						     in (SetBexps.union set set')
			       | [Zero]          ->  SetBexps.singleton Zero
			       | Zero::ls        ->  aux ls
			       | b::ls           ->  let set' = aux ls in
						     SetBexps.add b set'
			       | []              ->  SetBexps.empty
			   else
			     match lst' with
			       | b::ls           ->  let set' = aux ls in
						     SetBexps.add b set'
			       | []              ->  SetBexps.empty in 
			 let set = aux lst in
			 match (SetBexps.cardinal set) with
			     1   ->   SetBexps.min_elt set
			   | 0   ->   raise (ERROR "buildDisj empty set")
			   | n   ->   Disj set
			  
    let buildNeg e =
      match e with
	  NegBexp f  ->  f
	| otherwise  ->  (NegBexp e)

    let rec simplifyBexp bexp = 
      let rec loop bexp = 
	match bexp with
	    (One | Zero | Test _)->   bexp
	  | NegBexp b            ->   buildNeg (loop b)
	  | Conj set             ->   buildConj (List.map loop (SetBexps.elements set))
	  | Disj set             ->   buildDisj (List.map loop (SetBexps.elements set))
      in let simplifiedOnce = loop bexp in
	 let simplifiedTwice = loop simplifiedOnce in
	 if (MKE.bexpEq simplifiedOnce simplifiedTwice) then simplifiedTwice else 
	   (simplifyBexp simplifiedTwice) 
	     
    let rec simplify exp =
      let rec loop exp = 
	match exp with
	    Prog i         ->  exp
	  | Star e         ->  Star (loop e)
	  | Conc ls        ->  buildConcat (List.map loop ls)
	  | Or set         ->  buildOr (SetKE.elements (MKE.map loop set))
	  | Bexp b         ->  Bexp (simplifyBexp b)
      in loop exp

    let rec sizeBexp exp =
      match exp with
	  (Zero | One | Test _)   ->   1
	| NegBexp exp2            ->   1 + (sizeBexp exp2)
	| (Conj set | Disj set)   ->   runListBexp (SetBexps.elements set)
    and runListBexp lst =
      match lst with
	  [l]      ->  sizeBexp l
	| (l::ls)  ->  (sizeBexp l) + 1 + (runListBexp ls)  
	| []       ->  raise (ERROR "runListBexp received []")

    (* sizeKE exp returns the syntatic size of exp *)
    let rec sizeKE exp =
      match exp with
	  Prog _     ->  1
	| Bexp bexp  ->  sizeBexp bexp
	| Star sexp  ->  1 + sizeKE sexp
	| Conc lst   ->  runList lst
	| Or set     ->  runList (SetKE.elements set)
    and runList lst =
      match lst with
	  [l]        ->  sizeKE l
	| (l::ls)    ->  (sizeKE l) + 1 + (runList ls)  
	| []         ->  raise (ERROR "runListBexp received []")

    let sizeAt exp1 exp2 = 
      let n = float (SetInt.cardinal (SetInt.union (MKE.buildTfromKe exp1) (MKE.buildTfromKe exp2))) in
      if n=0. then 0. else (2. ** n)
	  
    (* Parser helper functions *)

    (* 'buildPlus lst' builds either a Or or a Bexp Disj depending 
       on the types of the elements in 'lst'. Elements of type be are always 
       grouped in a Bexp Disj. So the function builds a Bexp Disj iff 
       all the expressions and subexpressions in lst are of type be; otherwise it 
       builds a Or. (Note that [Bexp (Test 1), Or(Bexp (Test 2), 
       Bexp (Test 3))] is not transformed to [Bexp (Test 1), Bexp (Test 2),
       Bexp (Test 3)], it remains unchanged *)
    let buildPlus lst =
      let rec aux lst (lK,lB) = 
	match lst with
	    (Bexp b)::ls  ->  aux ls (lK,b::lB)
	  | (k)::ls       ->  aux ls (k::lK,lB)
	  | []            ->  (lK,lB)
      in let (keExps,beExps) = aux lst ([],[])
	 in if beExps=[] then (buildOr keExps)
	   else if keExps=[] then (MKE.Bexp (buildDisj beExps))
	   else match (buildOr keExps) with
	       Or set     ->  Or (SetKE.add (MKE.Bexp (buildDisj beExps)) set)
	     | element    ->  Or (SetKE.add (MKE.Bexp (buildDisj beExps)) (SetKE.singleton element))
	       
    let buildTimes lst =
      let rec aux accum lst = match lst with
	  (Bexp (Conj set))::ls  ->  aux (accum@(SetBexps.elements set)) ls
	| (Bexp be)::ls  ->  aux (accum@[be]) ls
	| e::ls ->  
	  if (List.length accum)=0 then 
	    e::(aux [] ls) 
	  else if (List.length accum)=1 then
	    (Bexp (List.hd accum))::e::(aux [] ls)
	  else 
	    (Bexp (Conj (setBexpsfromList accum)))::e::(aux [] ls)
	| []             ->  
	  if accum=[] then [] else (if (List.length accum)=1 then [MKE.Bexp (List.hd accum)] else [Bexp (Conj (setBexpsfromList accum))])
      in let lst_result = aux [] lst
	 in if (List.length lst_result > 1) then buildConcat lst_result else List.hd lst_result
	     
    let rec unions lst =
      match lst with
	  []	 ->  SetKE.empty
	| l::ls	 ->  SetKE.union l (unions ls) 

    let rec bexpEq exp1 exp2 =
      match (exp1,exp2) with
	  (MKE.Test i, MKE.Test j)          ->   i=j
	| (MKE.NegBexp b1, MKE.NegBexp b2)  ->   bexpEq b1 b2
	| (MKE.Conj s1, MKE.Conj s2)        ->   SetBexps.equal s1 s2 
	| (MKE.Disj s1, MKE.Disj s2)        ->   SetBexps.equal s1 s2
	| (MKE.One, MKE.One)                ->   true
	| (MKE.Zero, MKE.Zero)              ->   true
	| otherwise                         ->   false

    let rec keEq exp1 exp2 =
      match (exp1,exp2) with
	  (MKE.Prog i, MKE.Prog j)          ->   i=j
	| (MKE.Bexp e1, MKE.Bexp e2)        ->   bexpEq e1 e2
	| (MKE.Conc ls1, MKE.Conc ls2)      ->   auxConc ls1 ls2
	| (MKE.Or s1, MKE.Or s2)            ->   SetKE.equal s1 s2
	| otherwise                         ->   false
    and auxConc ls1 ls2 =
      match (ls1,ls2) with ([],[]) -> true | (x::xs,y::ys) -> (keEq x y) && (auxConc xs ys) | otherwise -> false
	
    let countTestsInBexp be =
      let rec aux be base = 
	match be with
	    Test _                 ->  1 + base
	  | NegBexp b              ->  aux b base
	  | (Conj set | Disj set)  ->  SetBexps.fold aux set base
	  | (Zero|One)             ->  base
      in aux be 0

    (* Counts the test ocurrences in exp *)
    let countTestsInKe exp =
      let rec aux exp base = 
	match exp with
	    Bexp b         ->  countTestsInBexp b + base
	  | Prog _         ->  base
	  | Star e         ->  aux e base
	  | Conc ls        ->  List.fold_right aux ls base
	  | Or set         ->  SetKE.fold aux set base
      in aux exp 0

    let map f set = SetKE.fold (fun x y -> SetKE.add (f x) y) set SetKE.empty
	    
  end
and OrderKE : Set.OrderedType   (* an order for expressions inside an 'Or' *)
  = struct
    type t = MKE.ke      
    let compare = Pervasives.compare
  end
and SetKE : Set.S with type elt = MKE.ke   (* a set for expressions inside an 'Or' *)
  = Set.Make( OrderKE )
		  and OrderBexps : Set.OrderedType
  = struct
    type t = MKE.be
    let compare = Pervasives.compare
  end
and SetBexps : Set.S with type elt = MKE.be   (* a set for boolean exps *)
      = Set.Make( OrderBexps )
