
exception ERROR of string
exception Found_Zero

open Module_KatExpressions
open ParserPolish
open Module_SetBuilder

let print str = print_endline str ; flush stdout
let print_el str = print_string str ; flush stdout

(* a module for pairs with two sets of integers *)
module rec MPairsIntSet : sig
  (* Warning: avoid using Set.mem x X as it is not the same as checking if Set.equal x x' for some x' in X;
     it appears to have the same result as doing x = x', which is erroneous. *)
  type elt = SetInt.t*SetInt.t
  type t = SetPairsIntSet.t
  val mem : elt -> t -> bool
  val set2String : SetPairsIntSet.t -> string
  val setFromList : (SetInt.t*SetInt.t) list -> SetPairsIntSet.t
  val printSet : SetPairsIntSet.t -> unit
  val printSet_nl : SetPairsIntSet.t -> unit
end
  = struct
      
    type elt = SetInt.t*SetInt.t
    type t = SetPairsIntSet.t

    let mem (x,y) set = 
      SetPairsIntSet.exists (fun (a,b) -> SetInt.equal x a && SetInt.equal y b) set

    let setFromList ls = List.fold_left (fun a b -> SetPairsIntSet.add b a) SetPairsIntSet.empty ls

    let set2String set = 
      let rec aux lst = match lst with
	  []           ->  "{empty_set}"
	| [(a,b)]      ->  "(" ^ (MInt.setInt2String a) ^ "," ^ (MInt.setInt2String b) ^ ")"
	| ((a,b)::xs)  ->  "(" ^ (MInt.setInt2String a) ^ "," ^ (MInt.setInt2String b) ^ "), " ^ (aux xs)
      in "{" ^ (aux (SetPairsIntSet.elements set)) ^ "}"

    let printSet set = print_string (set2String set)
    let printSet_nl set = print_endline (set2String set)

  end

(* a module for Kat expressions with actions distinguished
   based on their positions *)
module rec MKEPos : sig 
  type ke = Prog of int*int            (* second index is the position *)
	    | Bexp of MKE.be
	    | Star of ke
	    | Conc of ke list
	    | Or of SetKEPos.t 
  val buildConcat : ke list -> ke
  val buildPlus : ke list -> ke
  val buildTimes : ke list -> ke
  val simplify : ke -> ke
  val katExp2String : ke -> string
  val katExp2String_normalize : ke -> string
  val katExp2String_polish : ke -> string
  val printKE : ke -> unit
  val printKE_el : ke -> unit
  val printKE_polish : ke -> unit
  val buildTfromKe : ke -> SetInt.t
  val sizeKE : ke -> int
  val sizeAt : ke -> ke -> float 
  val unions : SetKEPos.t list -> SetKEPos.t
  val countTestsInKe : ke -> int
  val map : (ke -> ke) -> SetKEPos.t -> SetKEPos.t
  val removeIndexesPosition : MKEPos.ke -> MKE.ke ;;
end 
  = struct
      
    type ke = Prog of int*int
	      | Bexp of MKE.be
	      | Star of ke
	      | Conc of ke list
	      | Or of SetKEPos.t 

    let rec removeIndexesPosition exp = 
      match exp with
	Prog (p,i)  ->  MKE.Prog p
      | Bexp bexp   ->  MKE.Bexp bexp
      | Star exp'   ->  MKE.Star (removeIndexesPosition exp')
      | Conc lst    ->  MKE.Conc (List.map removeIndexesPosition lst)
      | Or set      ->  MKE.Or (SetKEPos.fold (fun x y -> SetKE.add (removeIndexesPosition x) y) set SetKE.empty)

    let rec katExp2String_polish exp =
      match exp with
	  Prog (p,i)     ->  "p"^(string_of_int p) (*^(string_of_int i)*)
	| Bexp b     ->  MKE.bexp2String_polish b
	| Star exp   ->  "*" ^ (katExp2String_polish exp)
	| Conc lst   ->  let rec aux = function  
	                    [l]    ->  katExp2String_polish l
	                  | l::ls  ->  "." ^ (katExp2String_polish l) ^ (aux ls)
			  | []     ->  raise (ERROR "aux in katExp2String_polish.Conc received []")
			 in if (List.length lst == 1) then (katExp2String_polish (List.hd lst)) else aux lst
	| Or set     ->  let rec printAux = function
                            []     ->  raise (ERROR "printAux in katExp2String_polish.Or received []")
			  | [l]    ->  katExp2String_polish l
	                  | l::ls  ->  "+" ^ (katExp2String_polish l) ^ (printAux ls)
			 in let set_elements = SetKEPos.elements set in
			    if (List.length set_elements == 1) then (katExp2String_polish (List.hd set_elements)) else (printAux set_elements)
		 
    let printKE_polish exp = print_endline (katExp2String_polish exp)
      
    let rec katExp2String exp =
      match exp with
	  Prog (p,i) ->  "p"^(string_of_int p) 
	| Bexp b     ->  MKE.bexp2String b
	| Star exp   ->  "(" ^ (katExp2String exp) ^ ")*"
	| Conc lst   ->  let rec aux = function  
	                    [l]    ->  katExp2String l
	                  | l::ls  ->  (katExp2String l) ^ (aux ls)
			  | []     ->  raise (ERROR "aux in katExp2String.Conc received []")
			 in "(" ^ (aux lst) ^ ")"
	| Or set     ->  let rec printAux = function
                            []     -> ""
	                  | l::ls  -> "+" ^ (katExp2String l) ^ (printAux ls)
			 in let set_elements = SetKEPos.elements set
			    in if (set_elements == []) then (raise (ERROR "function printAux in MKE.katExp2String received an empty set")) else "(" ^ (katExp2String (List.hd set_elements)) ^ (printAux (List.tl set_elements)) ^ ")"

    let printKE exp = print_str (katExp2String exp)
    let printKE_el exp = print (katExp2String exp)

    let katExp2String_normalize exp = 
      let parse str = 
	let linebuf = Lexing.from_string (str ^ "\nEOF") in
	List.hd (ParserPolish.data LexerPolish.token linebuf)
      in MKE.katExp2String (parse (MKE.katExp2String_polish (removeIndexesPosition exp)))

    let rec buildTfromKe exp =
    match exp with
	  Bexp b    ->  MKE.buildTfromBe b
	| Prog _    ->  SetInt.empty
	| Star e    ->  buildTfromKe e
	| Conc ls   ->  MInt.unions (List.map buildTfromKe ls)
	| Or set    ->  MInt.unions (List.map buildTfromKe (SetKEPos.elements set))

    let buildConcat ls =
      let simplify = true in   (* turns to 0 all concats of the form x..0..y *)
      match ls with
	  [exp]      ->  exp
	| otherwise  ->  
	  if simplify then
	    let rec loop ls =
	      match ls with
		| (Bexp MKE.Zero)::tl   ->  raise Found_Zero
		| [Bexp MKE.One]        ->  ls
		| (Bexp MKE.One)::tl    ->  loop tl
		| (Bexp r)::tl          ->  
		  if MKE.isAlwaysFalse r then
		    raise Found_Zero
		  else
		    (Bexp r)::(loop tl)
		| exp::tl               ->  exp::(loop tl)
		| []                    ->  []
	    in match (try loop ls with Found_Zero -> [Bexp MKE.Zero]) with
		[h]    ->  h
	      |	h::tl  ->  MKEPos.Conc (h::tl)
	      | []     ->  raise (ERROR "buildConcat received []")
	  else (MKEPos.Conc ls)

    let rec buildOr lst =
      match lst with
	  []         ->  raise (ERROR "buildOr received [] as argument ")
	| [l]        ->  l
	| otherwise  ->  let rec aux lst' =
			   match lst' with
			       (Or set)::ls    ->  let set' = aux ls 
						   in SetKEPos.union set set'
			     | [Bexp MKE.Zero]     ->  SetKEPos.singleton (Bexp MKE.Zero)
			     | (Bexp MKE.Zero)::ls ->  aux ls
			     | r::ls           ->  let set' = aux ls in
						   (SetKEPos.union (SetKEPos.singleton r) set')
			     | []              ->  SetKEPos.empty
			 in let newSet = aux lst in
			    buildOr_set newSet
    and buildOr_set set = 
      match (SetKEPos.cardinal set) with
	  1  ->  SetKEPos.min_elt set
	| 0  ->  raise (ERROR "buildOr_set received an empty set")
	| n  ->  Or set

    let rec simplify exp =
      let rec loop exp = 
	match exp with
	    Prog _         ->  exp
	  | Star e         ->  Star (loop e)
	  | Conc ls        ->  buildConcat (List.map loop ls)
	  | Or set         ->  buildOr (SetKEPos.elements (MKEPos.map loop set))
	  | Bexp b         ->  Bexp (MKE.simplifyBexp b)
      in loop exp 

    (* sizeKE exp returns the syntatic size of exp *)
    let rec sizeKE exp =
      match exp with
	  Prog _     ->  1
	| Bexp bexp  ->  MKE.sizeBexp bexp
	| Star sexp  ->  1 + sizeKE sexp
	| Conc lst   ->  runList lst
	| Or set     ->  runList (SetKEPos.elements set)
    and runList lst =
      match lst with
	  [l]        ->  sizeKE l
	| (l::ls)    ->  (sizeKE l) + 1 + (runList ls)  
	| []         ->  raise (ERROR "runList received []")

    let sizeAt exp1 exp2 = 
      let n = float (SetInt.cardinal (SetInt.union (buildTfromKe exp1) (buildTfromKe exp2))) in
      if n=0. then 0. else (2. ** n)
	  
    (* Parser helper functions *)

    (* 'buildPlus lst' builds either a ke.Or or a ke.(Bexp Disj) depending 
       on the types of the elements in 'lst'. Elements of type be are always 
       grouped in a Bexp Disj. So the function builds a ke.(Bexp Disj) iff 
       all the expressions and subexpressions in lst are of type be; otherwise it 
       builds a ke.Or. (Note that [Bexp (Test 1), Or(Bexp (Test 2), 
       Bexp (Test 3))] is not transformed to [Bexp (Test 1), Bexp (Test 2),
       Bexp (Test 3)], it remains unchanged *)
    let buildPlus lst =
      let rec aux lst (lK,lB) = 
	match lst with
	    (Bexp b)::ls   ->  aux ls (lK,b::lB)
	  | (k)::ls        ->  aux ls (k::lK,lB)
	  | []             ->  (lK,lB)
      in let (keExps,beExps) = aux lst ([],[])
	 in if beExps=[] then (buildOr keExps)
	   else if keExps=[] then (Bexp (MKE.buildDisj beExps))
	   else match (buildOr keExps) with
	       Or set      ->  Or (SetKEPos.add (Bexp (MKE.buildDisj beExps)) set)
	     | element     ->  Or (SetKEPos.add (Bexp (MKE.buildDisj beExps)) (SetKEPos.singleton element))
	       
    let buildTimes lst =
      let rec aux accum lst = match lst with
	  (Bexp be)::ls  ->  aux (accum@[be]) ls
	| e::ls          ->  
	  if accum=[] then e::(aux [] ls) else (Bexp (MKE.Conj (MKE.setBexpsfromList accum)))::e::(aux [] ls)
	| []             ->  
	  if accum=[] then [] 
	  else (if (List.length accum)=1 then [Bexp (List.hd accum)] else [Bexp (MKE.Conj (MKE.setBexpsfromList accum))])
      in let lst_result = aux [] lst
	 in if (List.length lst_result > 1) then buildConcat lst_result else List.hd lst_result

    let rec unions lst =
      match lst with
	  []	 ->  SetKEPos.empty
	| l::ls	 ->  SetKEPos.union l (unions ls) 

    let countTestsInKe exp =
      let rec aux exp base = 
	match exp with
	    Bexp b         ->  MKE.countTestsInBexp b + base
	  | Prog _         ->  base
	  | Star e         ->  aux e base
	  | Conc ls        ->  List.fold_right aux ls base
	  | Or set         ->  SetKEPos.fold aux set base
      in aux exp 0

    let map f set = SetKEPos.fold (fun x y -> SetKEPos.add (f x) y) set SetKEPos.empty
	    
  end
and OrderKEPos : Set.OrderedType   (* an order for expressions inside an 'Or' *)
  = struct
    type t = MKEPos.ke      
    let compare = Pervasives.compare
  end
and SetKEPos : Set.S with type elt = MKEPos.ke   (* a set for expressions inside an 'Or' *)
= Set.Make( OrderKEPos )

