
(* This module provides two methods for the construction of Non-Deterministic Finite Automata
   on Guarded Strings (NFAGS) from KAT expressions:
   BS - The Berry-Sethi construction method, as defined for regular expressions. This method
        produces automata of type 
           (Q, <o#,t#>), with 
           o#: P(Q) ---> 2
           t#: P(Q) ---> (P(Q))^{B+Σ},
        where B is the free Boolean algebra.
   BSKE - The BS variant method for KAT expressions. This method produces automata of type
           (Q, <o#,t#>), with 
           o#: P(Q) ---> B
           t#: P(Q) ---> (P(Q))^{BxΣ},
        where B is the free Boolean algebra.  
   In this project only the BSKE method is used. *)

open Module_KatExpressions
open Module_KatExpressionsPos
open Module_KatExpressionsPos2
open Module_PairsActionBexp
open Module_Utils
open Module_SetBuilder

exception ERROR of string

module rec NFAGS : sig
  type actBePairSet = SetPairsActBe.t
  type actionIndex = int
  type be = MKE.be
  type bePos = MKEPos2.be
  type kePos = MKEPos.ke
  type kePos2 = MKEPos2.ke
  type actionPos = kePos
  type intSet = SetInt.t
  type state = int
  type stateSet = intSet
  type o = int  (* either 1 or 0 *)
  type outputsBS = (state*o) list
  type outputsBSKE = (state*be) list
  type symbol = Prog of int | Bexp of be
  type transBSKE =       (* Each transition is a quad of the form 
			 ((departingState,action),(destinationState,test)) *)
      ((state*actionIndex)*(state*be)) list
  type transBS = (state*(symbol*state)) list 
  type automatonBS   = state*outputsBS*transBS 
  type automatonBSKE = state*outputsBSKE*transBSKE
  type automaton = 
      TransSum of automatonBS     (* when the transitions are by sums of actions and tests *)
    | TransProd of automatonBSKE  (* when the transitions are by products of actions and tests *)
  type description = string
  type index = int
  type position = int
  val buildAutUsingBS: kePos -> int -> automaton
  val buildAutUsingBSKE: kePos -> int -> automaton
  val buildAutomataUsingBS: kePos -> kePos -> automaton*automaton
  val buildAutomataUsingBSKE: kePos -> kePos -> automaton*automaton
  val buildTransBSKE: ((int*int)*be) list -> ((int*int)*actBePairSet) list -> state -> transBSKE
  val buildOutputsBSKE: ((int*int)*be) list -> ((int*int)*actBePairSet) list -> state -> outputsBSKE
  val bottomSt: state
  val bottomStSet: stateSet
  val e: kePos -> be
  val finSymbIndx: index
  val finSymbPos: position
  val finSymbol: actionPos 
  val firstBSKE: kePos -> actBePairSet 
  val initialState1: int
  val initialState2: int
end
  = struct
      
    type actBePairSet = SetPairsActBe.t
    type actionIndex = int
    type be = MKE.be
    type bePos = MKEPos2.be
    type ke = MKE.ke
    type kePos = MKEPos.ke
    type kePos2 = MKEPos2.ke
    type actionPos = kePos
    type intSet = SetInt.t
    type state = int
    type stateSet = intSet
    type o = int       (* either 1 or 0 *)
    type outputsBS = (state*o) list
    type outputsBSKE = (state*be) list
    type symbol = Prog of int | Bexp of be
    type transBSKE = ((state*actionIndex)*(state*be)) list
    type transBS = (state*(symbol*state)) list
    type automatonBS   = state*outputsBS*transBS 
    type automatonBSKE = state*outputsBSKE*transBSKE
    type automaton = 
	TransSum of automatonBS     (* when the transitions are by sums of actions and tests *)
      | TransProd of automatonBSKE  (* when the transitions are by products of actions and tests *)
    type description = string
    type index = int
    type position = int
	
    (* Function Aliases *)
    let bexp2String = MKE.bexp2String
    let buildConj = MKE.buildConj
    let buildConcPos = MKEPos.buildConcat
    let buildConcPos2 = MKEPos2.buildConcat
    let lsRemoveRepet = MUtils.lsRemoveRepetitions
    let markPositionsInSymbols = MKEPos2.markPositionsInSymbols
    let printBexp_nnl = MKE.printBexp_nnl
    let tryListHd = MUtils.tryListHd

    module T1: sig type t = symbol * position end = struct type t = symbol * position end
    module SymbPosPair = SM (T1)
    module SymbPosPairSet = SymbPosPair.S
    type symbPosPairSet = SymbPosPairSet.t

    module T2: sig type t = symbol end = struct	type t = symbol end
    module Symbol = SM (T2)
    module SymbolSet = Symbol.S
    type symbolSet = SymbolSet.t

    module T3: sig type t = (symbol * position) * symbPosPairSet end 
      = struct type t = (symbol * position) * symbPosPairSet end
    module PairsFBS = SM (T3)
    module SetPairsFBS = PairsFBS.S
    type pairsFBSSet = SetPairsFBS.t

    module T4: sig type t = (actionIndex * position) * actBePairSet end = 
    struct type t=(actionIndex * position) * actBePairSet end
    module PairsFBSKE = SM (T4)
    module SetPairsFBSKE = PairsFBSKE.S
    type pairsFBSKESet = SetPairsFBSKE.t

    (* Helpful Debugging Printers *)
    let printOutputsBSKE ls = 
      List.iter (fun (state,be) -> print_str (string_of_int state) ; MKE.printBexp_nnl be) ls ; flush stdout
    let printTransitionQuadsBSKE trans = 
      List.iter (fun ((x,y),(z,w)) -> print_str ("from " ^ string_of_int x ^ " by "); 
	print_str (string_of_int y ^ ","); print_str (" to " ^ string_of_int z ^ " if "); 
	MKE.printBexp_nnl w) trans; flush stdout
    let printSymbol symbol = 
      match symbol with 
	Prog p  ->  print_str ("p"^(string_of_int p))
      | Bexp b  ->  printBexp_nnl b
    let printFirst lsFirst = 
      print_str "{"; 
      List.iter (
	fun (symb,pos) -> print_str "(";printSymbol symb;print_str ",";
	  print_str (string_of_int pos);print_str "),") 
	lsFirst; print_str "}"
    let printActBePairSet_nnl (set:actBePairSet) = 
      print_str "{";
      List.iter (fun ((ind,pos),be) -> print_str ("((p"^(string_of_int ind)^","^string_of_int pos^"),"^bexp2String be^"),")) (SetPairsActBe.elements set);
      print_str "}"
    let printSymbPosPairSet_nnl (set:symbPosPairSet) = 
      print_str "{";
      List.iter (fun (symb,pos) -> 
	print_str "("; printSymbol symb; print_str ","; print_str (string_of_int pos); print_str "),")
	(SymbPosPairSet.elements set) ;
      print_str "}"
    let printFollow (lsF:((symbol*position)*symbPosPairSet) list) = 
      print_str "{";
      List.iter (fun ((symb,pos),set) -> print_str "<("; printSymbol symb; print_str ",";
	print_str (string_of_int pos); print_str "),"; printSymbPosPairSet_nnl set ;print_str ">, ") lsF;
      print_str "}"
    let debugAutomataBSKE (outputs,trans,i) = 
      (print ("\nAutomaton" ^ string_of_int i ^ ": \noutputs:") ; (printOutputsBSKE outputs) ; 
       print ("\ntransitions:") ; printTransitionQuadsBSKE trans)
	
    (* States of the Automaton1:
        - initial state = 1
        - bottom state (where non-existent transitions "go") = 0 
        - any other state > 1
       States of the Automaton2:
        - initial state = -1
        - bottom state (where non-existent transitions "go") = 0 
        - any other state < -1   *)
    let initialState1 = 1
    let initialState2 = -1
    let bottomSt = 0  
    let bottomStSet = SetInt.singleton bottomSt

    (* Before creating an automata out of an exp, the finalSymbol is concatenated to the
       end of exp; it corresponds to the ! symbol in the literature *)
    let finSymbIndx = 99
    let finSymbPos = 99
    let finSymbol = MKEPos.Prog (finSymbIndx,finSymbPos)
    let finSymbol2 = MKEPos2.Prog (finSymbIndx,finSymbPos)

    (* NFAGS Construction using the original Berry-Sethi Construction, as defined for Regular Expressions *)
    let rec oBexp b = 
      match b with
	  MKE.One       ->  1
	| MKE.Zero      ->  0
	| MKE.Test _    ->  0
	| MKE.NegBexp n ->  if (oBexp n)=1 then 0 else 1
	| MKE.Disj set  ->  
	  if SetBexps.is_empty set then raise (ERROR "empty set in oBexp")
	  else if SetBexps.exists (fun e -> (oBexp e)=1) set then 1 else 0
	| MKE.Conj set  ->  
	  if SetBexps.is_empty set then raise (ERROR "empty set in oBexp")
	  else if SetBexps.for_all (fun e -> (oBexp e)=1) set then 1 else 0

    let rec o (exp:kePos2) =
      match exp with
	  MKEPos2.Bexp (b,i)  ->  oBexp b
	| MKEPos2.Star _      ->  1
	| MKEPos2.Prog _      ->  0
	| MKEPos2.Or set      ->  
	  if SetKEPos2.is_empty set then raise (ERROR "empty set in o")
	  else if SetKEPos2.exists (fun e -> (o e)=1) set then 1 else 0
	| MKEPos2.Conc ls     ->  if List.for_all (fun e -> (o e)=1) ls then 1 else 0

    (* isBoolean checks if exp is boolean, which can happen if exp is directly of type Bexp or
       if it's of the form of a Star recursively applied to a Bexp. If the answer is positive, the position
       of the outermost expression is returned *)
    let rec isBoolean exp = 
      match exp with 
	  MKEPos2.Bexp (_,pos)  ->  Some pos
	| MKEPos2.Star exp'     ->  isBoolean exp'
	| _                     ->  None
	       
    let rec firstBS (exp:kePos2) : SymbPosPairSet.t =
      match exp with
	MKEPos2.Prog (p,i)       ->  SymbPosPairSet.singleton (Prog p,i)
      | MKEPos2.Bexp (b,i)       ->  SymbPosPairSet.singleton (Bexp b,i)
      | MKEPos2.Star (MKEPos2.Bexp (_,i)) -> SymbPosPairSet.singleton (Bexp MKE.One,i) 
      | MKEPos2.Star exp'        ->  firstBS exp'
      | MKEPos2.Conc (exp1::ls)  ->  let first1 = firstBS exp1 in
				     if (o exp1)=1 then
				       let exp2 = buildConcPos2 ls in
				       SymbPosPairSet.union first1 (firstBS exp2)
				     else 
				       first1
      | MKEPos2.Conc []          ->  raise (ERROR "function firstBS received an unexpected pattern: Conc []")
      | MKEPos2.Or set           ->  let rec loop set =
				       if SetKEPos2.is_empty set then SymbPosPairSet.empty
				       else 
					 let min_elt = SetKEPos2.min_elt set in
					 SymbPosPairSet.union (firstBS min_elt) 
					   (loop (SetKEPos2.remove min_elt set))
				     in loop set
				     
    (* fBS exp set build the follow of exp. *)
    let rec fBS (exp:kePos2) (set:symbPosPairSet) : pairsFBSSet =
      match exp with
	MKEPos2.Prog (p,i)       ->  SetPairsFBS.singleton ((Prog p,i),set)
      | MKEPos2.Bexp (b,i)       ->  SetPairsFBS.singleton ((Bexp b,i),set)
      | MKEPos2.Star exp'        ->  
	(match isBoolean exp' with
	    Some pos  ->  SetPairsFBS.singleton ((Bexp MKE.One,pos),set)
	  | None      ->  fBS exp' (SymbPosPairSet.union (firstBS exp') set))
      | MKEPos2.Conc (exp1::ls)  ->  let exp2 = buildConcPos2 ls in
				     let set1 = 
				       if (o exp2)=1 then
					 SymbPosPairSet.union (firstBS exp2) set
				       else
					 firstBS exp2
				     in  SetPairsFBS.union (fBS exp1 set1) (fBS exp2 set)
      | MKEPos2.Conc []          ->  raise (ERROR "function fBS received an unexpected pattern: Conc []")
      | MKEPos2.Or set'          ->  let rec loop set'' =
				       if SetKEPos2.is_empty set'' then SetPairsFBS.empty
				       else 
					 let min_elt = SetKEPos2.min_elt set'' in
					 SetPairsFBS.union (fBS min_elt set) 
					   (loop (SetKEPos2.remove min_elt set''))
				     in loop set'
					 
    let buildOutputsBS (lsFirst:(symbol*index) list) (lsF:((symbol*index)*symbPosPairSet) list) initialState =
      let ls1 =    (* the outputs of the initial state *)
	[? List: (initialState,1) | (z,_) <- List: lsFirst ; z=Prog finSymbIndx ?] in
      (* ls2 = the outputs of the other states; we make not(p'=finSymbIndx) to avoid infinite loops 
	 from the state -1 to itself *)
      let ls2 = [? List: (i,1) | ((x,i),set) <- List: lsF ; (y,_) <- List: SymbPosPairSet.elements set ; 
		 y=Prog finSymbIndx ; not(x=Prog finSymbIndx); not(i=initialState) ?] in  
      let ls3 = if ls1=[] then ([(initialState,0)]@ls2) else (ls1@ls2) in
      (* all other states for which not output was found, have output=0 *)
      let ls4 = [? List: (i,0) | ((symb,i),_) <- List: lsF ; (List.mem_assoc i ls3)=false ; 
		 not(symb = Prog finSymbIndx) ?] in
      (ls3@(ls4@[(bottomSt,0)]))        (* by convention, the bottom state has output 0 *)
	
    (* Builds a list of the transitions *)
    let buildTransBS lsFirst lsF initialState =
      (* firstTrans = the transitions from the initial state *)
      let firstTrans = 
	[? List: (initialState,(symb,j)) | (symb,j) <- List: lsFirst ; not(symb = Prog finSymbIndx) ?] in
      let otherTrans = 
	[? List: (i,(symb,j)) | ((symb,i),set) <- List: lsF ; 
	 (symb,j) <- List: SymbPosPairSet.elements set ; not(symb=Prog finSymbIndx) ?]
      in lsRemoveRepet (firstTrans@otherTrans)
      
    let buildAutUsingBS exp i =
      let exp = MKEPos2.Conc [markPositionsInSymbols exp i; finSymbol2] in
      let lsFirst = SymbPosPairSet.elements (firstBS exp) in                    (* firstBS *)
      let setAux = SymbPosPairSet.singleton (Prog finSymbIndx,99) in
      let lsF = SetPairsFBS.elements (fBS exp setAux) in                        (* follow *)
      let initialState = if i=1 then initialState1 else initialState2 in
      let outputs = buildOutputsBS lsFirst lsF initialState in
      let () = if !debugging then 
	  (print ("\nExpression "^(MKEPos2.katExp2String exp)^" has first: "); printFirst lsFirst) in
      let () = if !debugging then 
	  (print (" and follow: "); printFollow lsF) in
      let trans = buildTransBS lsFirst lsF initialState in 
      TransSum (initialState,outputs,trans)
	
    (* NFAGS Construction using the Berry-Sethi Construction for KAT Expressions *)
    let rec e exp =
      match exp with
	  MKEPos.Bexp b   ->  b
	| MKEPos.Star _   ->  MKE.One
	| MKEPos.Prog _   ->  MKE.Zero
	| MKEPos.Or set   ->  let rec loop set =
				if SetKEPos.is_empty set then SetBexps.empty
				else 
				  let min_elt = SetKEPos.min_elt set in
				  SetBexps.add (e min_elt) (loop (SetKEPos.remove min_elt set))
			      in let newSet = loop set in
				 (match (SetBexps.cardinal newSet) with
				     0  ->  raise (ERROR "empty set in e")
				   | 1  ->  SetBexps.min_elt newSet
				   | _  ->  MKE.Disj newSet)       (* *)
	| MKEPos.Conc ls  ->  buildConj (List.map e ls)
	  
    let rec firstBSKE exp =
      match exp with
	  MKEPos.Prog (p,i)       ->  SetPairsActBe.singleton ((p,i),MKE.One)
	| MKEPos.Bexp b           ->  SetPairsActBe.empty
	| MKEPos.Star exp'        ->  firstBSKE exp'
	| MKEPos.Conc (exp1::ls)  ->  let exp2 = buildConcPos ls in
				      let z = e exp1 in
				      if (z=MKE.Zero) then firstBSKE exp1
				      else SetPairsActBe.union (firstBSKE exp1) 
					(MPairsActBe.map (fun (x,y) -> (x,buildConj [z;y])) (firstBSKE exp2))
	| MKEPos.Conc []          ->  raise (ERROR "function firstBSKE received an unexpected pattern: Conc []")
	| MKEPos.Or set           ->  let rec loop set =
					if SetKEPos.is_empty set then SetPairsActBe.empty
					else 
					  let min_elt = SetKEPos.min_elt set in
					  SetPairsActBe.union (firstBSKE min_elt) 
					    (loop (SetKEPos.remove min_elt set))
				      in loop set

    (* fBSKE exp set build the follow of exp. *)
    let rec fBSKE exp (set:actBePairSet) : pairsFBSKESet =
      match exp with
	  MKEPos.Prog (p,i)       ->  SetPairsFBSKE.singleton ((p,i),set)
	| MKEPos.Bexp b           ->  SetPairsFBSKE.empty
	| MKEPos.Star exp'        ->  fBSKE exp' (SetPairsActBe.union (firstBSKE exp') set)
	| MKEPos.Conc (exp1::ls)  ->  let exp2 = buildConcPos ls in
				      let z = e exp2 in
				      let set1 = 
					if (z=MKE.Zero) then firstBSKE exp2
					else 
					  SetPairsActBe.union (firstBSKE exp2) 
					    (MPairsActBe.map (fun (x,y) -> (x,buildConj [z;y])) set) in
				      let set2 = fBSKE exp1 set1 in
				      let set3 = fBSKE exp2 set in
				      SetPairsFBSKE.union set2 set3
	| MKEPos.Conc []          ->  raise (ERROR "function fBSKE received an unexpected pattern: Conc []")
	| MKEPos.Or set'          ->  let rec loop set'' =
					if SetKEPos.is_empty set'' then SetPairsFBSKE.empty
					else 
					  let min_elt = SetKEPos.min_elt set'' in
					  SetPairsFBSKE.union (fBSKE min_elt set) 
					    (loop (SetKEPos.remove min_elt set''))
				      in loop set'
				      
    let buildOutputsBSKE lsFirst lsF initialState =
      (* ls1 = the outputs of the initial state *)
      let ls1 = [? List: (initialState,be) | ((p,_),be) <- List: lsFirst ; p=finSymbIndx ?] in
      (* ls2 = the outputs of the other states; we make not(p'=finSymbIndx) to avoid infinite loops 
	 from the state -1 to itself *)
      let ls2 = [? List: (i,be) | ((p',i),set) <- List: lsF ; ((p,_),be) <- List: SetPairsActBe.elements set ; 
		 p=finSymbIndx ; not(p'=finSymbIndx) ?] in  
      let ls3 = if ls1=[] then ([(initialState,MKE.Zero)]@ls2) else (ls1@ls2) in
      (* all other states for which not output was found, have output=0 *)
      let ls4 = 
	[? List: (i,MKE.Zero) | ((p,i),_) <- List: lsF ; (List.mem_assoc i ls3)=false ; not(p=finSymbIndx) ?] in
      (ls3@(ls4@[(bottomSt,MKE.Zero)]))  (* by convention, the bottom state has output 0 *)

    (* Builds a list of the transitions *)
    let buildTransBSKE lsFirst lsF initialState =
      (* firstTrans = the transitions from the initial state *)
      let firstTrans = 
	[? List: ((initialState,p),(j,be)) | ((p,j),be) <- List: lsFirst ; not(p=finSymbIndx) ?] in
      let otherTrans = 
	[? List: ((i,p),(j,be)) | ((_,i),set) <- List: lsF ; 
	 ((p,j),be) <- List: SetPairsActBe.elements set ; not(p=finSymbIndx) ?]
      in (firstTrans@otherTrans)
      
    let buildAutUsingBSKE exp i =
      let exp = MKEPos.Conc [exp; finSymbol] in
      let lsFirst = SetPairsActBe.elements (firstBSKE exp) in                       (* firstBSKE *)
      let setAux = SetPairsActBe.singleton ((finSymbIndx,finSymbPos), MKE.One) in
      let lsF = SetPairsFBSKE.elements (fBSKE exp setAux) in                        (* follow *)
      let initialState = if i=1 then initialState1 else initialState2 in
      let outputs = buildOutputsBSKE lsFirst lsF initialState in
      let trans = 
	buildTransBSKE lsFirst lsF initialState in 
      (if !debugging then 
	  debugAutomataBSKE (outputs,trans,i)) ; 
      TransProd (initialState,outputs,trans)

    (* Builds a pair of NFAGS from KAT expressions exp1 and exp2 *)
    let buildAutomataUsingBS exp1 exp2 = 
      let a1 = buildAutUsingBS exp1 1 in 
      let a2 = buildAutUsingBS exp2 2 in
      (a1,a2)
	
    let buildAutomataUsingBSKE exp1 exp2 = 
      let a1 = buildAutUsingBSKE exp1 1 in 
      let a2 = buildAutUsingBSKE exp2 2 in
      (a1,a2)

  end
