
(* a module of functions that apply several rewrite rules based
   on axioms of KA and BA to katexps, thus transforming them to 
   sintacticaly different but still equivalent expressions.
   For more info on these rules see 'Ricardo Almeida. MsC theses.
   Decision Algorithms for Kleene Algebra with Tests and Hoare 
   Logic. 2012' *)

open Module_KatExpressions

exception ERROR of string
exception MaxLenExceded of int

module rec MRewriteRules : sig 
  type t = MKE.ke
  val applyComplements : MKE.ke  -> (MKE.ke * int)
  val applyComplementsBexp : MKE.be  -> (MKE.be * int)
  val applyRewriteRules : MKE.ke -> int -> int -> (MKE.ke * ((string * int) list))
  val applyLeftDistrConjOnce : MKE.ke ->  (MKE.ke * int)
  val applyLeftDistrConj : MKE.ke -> int -> (MKE.ke * int)
  val applyLeftDistrDisj : MKE.ke -> int -> (MKE.ke * int)
  val applyLeftDistrDisjOnce : MKE.ke  -> (MKE.ke * int)
  val applyDenesting : MKE.ke ->  (MKE.ke * int)
  val applyTwiceStar : MKE.ke ->  (MKE.ke * int)
  val applyLeftDistrConcOnce : MKE.ke -> (MKE.ke * int)
  val applyLeftDistrConc : MKE.ke -> int -> (MKE.ke * int)
end 
  = struct
      
    type t = MKE.ke

    let rec remove lst x = match lst with
	(l::ls)  ->  if x=l then ls else l::(remove ls x)
      | []       ->  raise (ERROR "function remove reached the empty list")

    (* application of BA axioms starts here  *)
	
    (* substitutes 
         - Zero for tnott and vice-versa
         - One for t+nott and vice-versa *)
    let rec applyComplements exp =  
      let aux n exp =
	match exp with
	    MKE.Bexp bexp   ->   let (newBexp,newN) = applyComplementsBexp bexp 
				 in (MKE.Bexp newBexp, newN)
	  | MKE.Prog i      ->   (MKE.Prog i, 0)
	  | MKE.Star exp    ->   let (newExp,newN) = applyComplements exp
				 in (MKE.Star newExp, newN)
	  | MKE.Or set      ->   let (newSet,newN) = transformSet set
				 in (MKE.Or newSet, newN)
	  | MKE.Conc lst    ->   let (newLst,newN) = transformLst lst
				 in (MKE.Conc newLst, newN)
      in aux 0 exp
    and applyComplementsBexp bexp =
      let rec aux n exp =
	match exp with
	    MKE.Zero  ->  (MKE.(Conj (MKE.setBexpsfromList [MKE.(Test 1);MKE.(NegBexp (Test 1))])), 1)
	  | MKE.One   ->  (MKE.(Disj (MKE.setBexpsfromList [MKE.(Test 1);MKE.(NegBexp (Test 1))])), 1)
	  | MKE.Disj set  ->  let (newSet,newN) = transformBSet set
			      in (MKE.Disj newSet, newN)
	  | MKE.Conj set  ->  let (newLst,newN) = transformBLst (SetBexps.elements set)
			      in (MKE.Conj (MKE.setBexpsfromList newLst), newN)
	  | MKE.NegBexp bexp     ->  let (newBexp,newN) = applyComplementsBexp bexp
				     in (MKE.NegBexp newBexp, newN)
	  | MKE.Test t     ->  (MKE.Test t,0)
      in aux 0 bexp
    and transformBSet set = 
      if (SetBexps.is_empty set) then (set,0) else 
	let fst_elt = SetBexps.min_elt set
	in match fst_elt with
	    MKE.Test i when (SetBexps.mem (MKE.NegBexp (MKE.Test i)) set) ->
		let set1 = SetBexps.remove (MKE.Test i) (SetBexps.remove (MKE.NegBexp (MKE.Test i)) set)
		in let (set', n') = transformBSet set1
		   in (SetBexps.add (MKE.One) set', n'+1) 
	  | MKE.NegBexp (MKE.Test i) when ((SetBexps.mem (MKE.Test i)) set)  ->   
	      let set1 = SetBexps.remove (MKE.Test i) (SetBexps.remove (MKE.NegBexp (MKE.Test i)) set)
	      in let (set', n') = transformBSet set1
		 in (SetBexps.add (MKE.Zero) set', n'+1)
	  | otherwise  ->  let (exp',n') = applyComplementsBexp fst_elt
			   in let (exp'',n'') = transformBSet (SetBexps.remove fst_elt set)
			      in (SetBexps.union (SetBexps.singleton exp') exp'', n'+n'')
    and transformBLst lst = match lst with
	(MKE.Test i)::ls when (List.mem (MKE.NegBexp (MKE.Test i)) ls)   -> 
	  let ls' = remove ls (MKE.NegBexp (MKE.Test i))
	  in let (lst',n') = transformBLst ls'
	     in ((MKE.Zero)::lst',n'+1)
      | (MKE.NegBexp (MKE.Test i))::ls when (List.mem (MKE.Test i) ls)  -> 
	let ls' = remove ls (MKE.Test i)
	in let (lst',n') = transformBLst ls'
	   in ((MKE.Zero)::lst',n'+1) 
      | l::ls        ->  let (l',n') = applyComplementsBexp l
			 in let (ls',n'') = transformBLst ls
			    in (l'::ls', n'+n'')
      | []           ->  ([],0)
    and transformSet set = 
      if (SetKE.is_empty set) then (set,0) else 
	let fst_elt = SetKE.min_elt set
	in let (exp',n') = applyComplements fst_elt
	   in let (exp'',n'') = transformSet (SetKE.remove fst_elt set)
	      in (SetKE.union (SetKE.singleton exp') exp'', n'+n'')
    and transformLst lst = 
      match lst with
	  []       ->  (lst,0)
	| (l::ls)  ->  let (hd,n') = applyComplements l
		       in let (tl,n'') = transformLst ls
			  in (hd::tl,n'+n'')

    (* substitutes bexp1bexp2+..+bexp1bexpn for bexp1(bexp2+..+bexpn) *)
    let rec applyLeftDistrConjOnce exp =  
      let aux n exp =
	match exp with
	    MKE.Bexp bexp   ->   let (newBexp,newN) = applyAxiomInBexp bexp
				 in (MKE.Bexp newBexp, newN)
	  | MKE.Prog i      ->   (MKE.Prog i, 0)
	  | MKE.Star exp    ->   let (newExp,newN) = applyLeftDistrConjOnce exp
				 in (MKE.Star newExp, newN)
	  | MKE.Or set      ->   let (newSet,newN) = transformSet set
				 in (MKE.Or newSet, newN)
	  | MKE.Conc lst    ->   let (newLst,newN) = transformLst lst
				 in (MKE.Conc newLst, newN)
      in aux 0 exp
    and applyAxiomInBexp bexp =
      let rec aux n exp =
	match exp with
	    MKE.Conj set      ->  transformBSet [] set
	  | MKE.NegBexp bexp  ->  let (newBexp,newN) = applyAxiomInBexp bexp
				  in (MKE.NegBexp newBexp, newN)
	  | MKE.Disj set      ->  let (newSet,newN) = iterateInBSet set
				  in (MKE.Disj newSet, newN)
	  | otherwise         ->  (exp,0)
      in aux 0 bexp
    and transformBSet accum set = 
      if (SetBexps.is_empty set) then (if accum=[] then (MKE.Zero,0) else MKE.Conj (MKE.setBexpsfromList accum),0) else 
	let fst_elt = SetBexps.min_elt set
	in match fst_elt with
	    MKE.Disj set_d  ->  
	      if accum!=[] then
		let newDisj = distribute accum set_d
		in let (exp',n') = transformBSet [] (SetBexps.remove fst_elt set)
		   in match exp' with
		       MKE.Disj s -> (MKE.Conj (SetBexps.add (MKE.Disj newDisj) (SetBexps.singleton (MKE.Disj s))), n'+1)
		     | MKE.Conj s -> (MKE.Conj (SetBexps.union s (SetBexps.singleton (MKE.Disj newDisj))), n'+1)
		     | MKE.Zero   -> (MKE.Disj newDisj, n'+1)
		     | otherwise  -> raise (ERROR "inner patter matching in transformBSet received a pattern different from MKE.Disj _ and MKE.Conj _ and MKE.Zero")
	      else 
		transformBSet [fst_elt] (SetBexps.remove fst_elt set)
	  | otherwise       ->  let (exp',n') = applyAxiomInBexp fst_elt
				in let (exp'',n'') = transformBSet (accum@[exp']) (SetBexps.remove fst_elt set)
				   in (exp'',n''+n')
    (* distribute receives a list of bexps [e1,e2,..,en], 'accum', and 
       builds the conjunction c = e1e2...en. Then every element di in
       set_d becomes cdi. The return value is (MKE.Disj set_d) *)
    and distribute accum set_d = 
      if accum=[] then raise (ERROR "function distribute in applyLeftDistrConjOnce received [] as accum")
      else 
	let buildConj' lst elt = 
	  match elt with
	      MKE.Conj set_c  ->   MKE.Conj (SetBexps.union set_c (MKE.setBexpsfromList lst))
	    | otherwise       ->   MKE.Conj (MKE.setBexpsfromList  (lst@[elt]))
in let newSet = MKE.setBexpsMap (buildConj' accum) set_d
	in newSet
    and iterateInBSet set = 
      if (SetBexps.is_empty set) then (set,0) else 
	let fst_elt = SetBexps.min_elt set
	in let (exp',n') = applyAxiomInBexp fst_elt
	   in let (set'',n'') = iterateInBSet (SetBexps.remove fst_elt set)
in (SetBexps.union (SetBexps.singleton exp') set'', n'+n'')
    and transformSet set = 
      if (SetKE.is_empty set) then (set,0) else 
	let fst_elt = SetKE.min_elt set
	in let (exp',n') = applyLeftDistrConjOnce fst_elt
	   in let (exp'',n'') = transformSet (SetKE.remove fst_elt set)
	      in (SetKE.union (SetKE.singleton exp') exp'', n'+n'')
    and transformLst lst = 
      match lst with
	  []       ->  (lst,0)
	| (l::ls)  ->  let (hd,n') = applyLeftDistrConjOnce l
		       in let (tl,n'') = transformLst ls
			  in (hd::tl,n'+n'')
			  
    let rec applyLeftDistrConj exp maxLen = 
      let (exp',n') = applyLeftDistrConjOnce exp in
      if (((MKE.sizeKE exp') > maxLen) || (MKE.katExp2String exp = MKE.katExp2String exp')) then 
	(exp,0)
      else
	let (exp'',n'') = applyLeftDistrConj exp' maxLen
	in (exp'',n''+n')

    (* substitutes (bexp1+bexp2)(bexp1+bexp3)..(bexp1+bexpn) for bexp1+bexp2bexp3..bexpn *)
    let rec applyLeftDistrDisjOnce exp =  
	match exp with
	    MKE.Bexp bexp   ->   let (newBexp,newN) = applyAxiomInBexp bexp
				 in (MKE.Bexp newBexp, newN)
	  | MKE.Prog i      ->   (MKE.Prog i, 0)
	  | MKE.Star exp    ->   let (newExp,newN) = applyLeftDistrDisjOnce exp
				 in (MKE.Star newExp, newN)
	  | MKE.Or set      ->   let (newSet,newN) = transformSet set
				 in (MKE.Or newSet, newN)
	  | MKE.Conc lst    ->   let (newLst,newN) = transformLst lst
				 in (MKE.Conc newLst, newN)
    and applyAxiomInBexp bexp =
	match bexp with
	    MKE.Conj set  ->  let (newSet,newN) = iterateInBSet set
			      in (MKE.Conj newSet, newN)
	  | MKE.NegBexp bexp'  ->  let (newBexp,newN) = applyAxiomInBexp bexp'
				  in (MKE.NegBexp newBexp, newN)
	  | MKE.Disj set  ->  transformBSet [] set
	  | otherwise  ->  (bexp,0)
    and transformBSet accum set = 
      if (SetBexps.is_empty set) then (if accum=[] then (MKE.Zero,0) else MKE.Disj (MKE.setBexpsfromList accum),0) else 
	let fst_elt = SetBexps.min_elt set
	in match fst_elt with
	    MKE.Conj set_d  ->  
	      if accum!=[] then
		let newConj = distribute accum set_d
		in let (exp',n') = transformBSet [] (SetBexps.remove fst_elt set)
		   in match exp' with
		       MKE.Disj s -> (MKE.Disj (SetBexps.add (MKE.Conj newConj) s), n'+1)
		     | MKE.Conj s -> (MKE.Disj (SetBexps.union (SetBexps.singleton exp') (SetBexps.singleton (MKE.Conj newConj))), n'+1)
		     | MKE.Zero   -> (MKE.Conj newConj, n'+1)
		     | otherwise  -> raise (ERROR "inner patter matching in transformBSet of applyLeftDistrDisj received a pattern different from MKE.Disj _ and MKE.Conj _ and MKE.Zero")
	      else 
		transformBSet [fst_elt] (SetBexps.remove fst_elt set)
	  | otherwise       ->  let (exp',n') = applyAxiomInBexp fst_elt
				in let (exp'',n'') = transformBSet (accum@[exp']) (SetBexps.remove fst_elt set)
				   in (exp'',n''+n')
    and distribute accum set_c = 
      if accum=[] then raise (ERROR "function distribute in applyLeftDistrDisjOnce received [] as accum")
      else 
	let buildDisj' lst elt = 
	  match elt with
	      MKE.Disj set_d  ->   MKE.Disj (SetBexps.union set_d (MKE.setBexpsfromList lst))
	    | otherwise       ->   MKE.Disj (MKE.setBexpsfromList  (lst@[elt]))
	in let newSet = MKE.setBexpsMap (buildDisj' accum) set_c
	   in newSet
    and iterateInBSet set = 
      if (SetBexps.is_empty set) then (set,0) else 
	let fst_elt = SetBexps.min_elt set
	in let (exp',n') = applyAxiomInBexp fst_elt
	   in let (set'',n'') = iterateInBSet (SetBexps.remove fst_elt set)
	      in (SetBexps.union (SetBexps.singleton exp') set'', n'+n'')
    and transformSet set = 
      if (SetKE.is_empty set) then (set,0) else 
	let fst_elt = SetKE.min_elt set
	in let (exp',n') = applyLeftDistrDisjOnce fst_elt
	   in let (exp'',n'') = transformSet (SetKE.remove fst_elt set)
	      in (SetKE.union (SetKE.singleton exp') exp'', n'+n'')
    and transformLst lst = 
      match lst with
	  []       ->  (lst,0)
	| (l::ls)  ->  let (hd,n') = applyLeftDistrDisjOnce l
		       in let (tl,n'') = transformLst ls
			  in (hd::tl,n'+n'')

    let rec applyLeftDistrDisj exp maxLen = 
      let (exp1,n1) = applyLeftDistrDisjOnce exp in
      if (((MKE.sizeKE exp1) > maxLen) || (MKE.katExp2String exp1 = MKE.katExp2String exp)) then 
	(exp,0) 
      else
	let (exp3,n3) = applyLeftDistrDisj exp1 maxLen
	in (exp3,n3+n1)
	
	
    (* application of KA axioms starts here  *) 

    (* substitutes ( expn*..exp2*exp1* )*expn* for ( exp1+exp2+...+expn )*    *)
    let rec applyDenesting exp =  
      match exp with
	  MKE.Star exp'  ->  let foo exp = 
			       match exp with
				   MKE.Or set               ->  transformExp set
				 | MKE.Bexp (MKE.Disj set)  ->  transformBExp set
				 | otherwise                ->  let (exp2,n2) = applyDenesting exp
								in (MKE.Star exp2, n2)
			     in foo exp'
	| MKE.Or set     ->   let (newSet,newN) = transformSet set
			      in (MKE.Or newSet, newN)
	| MKE.Conc lst   ->   let (newLst,newN) = transformLst lst
			      in (MKE.buildConcat newLst, newN)
	| otherwise      ->   (exp, 0)
    and transformExp set =
      let lst_elts = List.rev (SetKE.elements set)
      in let applyStar exp = MKE.Star exp
	 in let rec foo lst = match lst with
	     l::ls  ->  let (ls',n') = foo ls
			in let (l',n) = applyDenesting l
			   in (l'::ls',n+n')
	   | []     ->  ([],1)
	    in let (lst',n') = foo lst_elts
	       in let inner_conc = MKE.buildConcat (List.map applyStar lst')
		  in (MKE.Conc [MKE.Star inner_conc; MKE.Star (List.hd lst')],n')
    and transformBExp set =
      let lst_elts = List.rev (SetBexps.elements set)
      in let applyStar exp = MKE.Star (MKE.Bexp exp)
	 in let inner_conc = MKE.buildConcat (List.map applyStar lst_elts)
	    in (MKE.Conc [MKE.Star inner_conc; MKE.Star (MKE.Bexp (List.hd lst_elts))],1)
    and transformSet set = 
      if (SetKE.is_empty set) then (set,0) else 
	let fst_elt = SetKE.min_elt set
	in let (exp',n') = applyDenesting fst_elt
	   in let (exp'',n'') = transformSet (SetKE.remove fst_elt set)
	      in (SetKE.union (SetKE.singleton exp') exp'', n'+n'')
    and transformLst lst = 
      match lst with
	  []       ->  (lst,0)
	| (l::ls)  ->  let (hd,n') = applyDenesting l
		       in let (tl,n'') = transformLst ls
			  in (hd::tl,n'+n'')

    (*   substitutes exp*exp* for exp*   *)
    let rec applyTwiceStar exp =  
      match exp with
	  MKE.Star exp'  ->  let (exp2,n2) = applyTwiceStar exp'
			       in (MKE.Conc [MKE.Star exp2; MKE.Star exp2], n2+1)
	| MKE.Or set     ->   let (newSet,newN) = transformSet set
			       in (MKE.Or newSet, newN)
	| MKE.Conc lst   ->   let (newLst,newN) = transformLst lst
			       in (MKE.Conc newLst, newN)
	| otherwise      ->   (exp, 0)
    and transformSet set = 
      if (SetKE.is_empty set) then (set,0) else 
	let fst_elt = SetKE.min_elt set
	in let (exp',n') = applyTwiceStar fst_elt
	   in let (exp'',n'') = transformSet (SetKE.remove fst_elt set)
	      in (SetKE.union (SetKE.singleton exp') exp'', n'+n'')
    and transformLst lst = 
      match lst with
	  []       ->  (lst,0)
	| (l::ls)  ->  let (hd,n') = applyTwiceStar l
		       in let (tl,n'') = transformLst ls
			  in match hd with
			      MKE.Conc lst  ->  
				(match lst with 
				    [MKE.Star e;MKE.Star e'] when e=e'  ->  (lst@tl,n'+n'')
				  | otherwise                           ->  ((MKE.Conc lst)::tl,n'+n''))
			    | otherwise   ->   (hd::tl,n'+n'')	    

    (* substitutes 
       - exp1exp2+..+exp1expn for exp1(exp2+..+expn)
       - exp1bexp2+..+exp1bexpn for exp1(bexp2+..+bexpn) *)
    let rec applyLeftDistrConcOnce exp =  
      match exp with
	| MKE.Star exp    ->   let (newExp,newN) = applyLeftDistrConcOnce exp
			       in (MKE.Star newExp, newN)
	| MKE.Or set      ->   let (newSet,newN) = transformSet set
			       in (MKE.Or newSet, newN)
	| MKE.Conc lst    ->   let (newExp,newN) = transformLst [] lst
			       in (newExp, newN)
	| otherwise       ->   (exp,0)
    and transformLst accum lst = 
      if lst=[] then (if accum=[] then (MKE.Bexp MKE.Zero,0) else (MKE.buildConcat accum),0) else 
	let fst_elt = List.hd lst
	in match fst_elt with
	    MKE.Or set_or  ->  
	      if accum!=[] then
		let newOr = distribute accum (SetKE.elements set_or)
		in aux newOr lst
	      else 
		transformLst [fst_elt] (List.tl lst)
	  | MKE.Bexp (MKE.Disj set_d)  ->  
	    if accum!=[] then
	      let newOr = distributeB accum (SetBexps.elements set_d)
	      in aux newOr lst
	    else 
	      transformLst [fst_elt] (List.tl lst)
	  | otherwise       ->  let (exp',n') = applyLeftDistrConcOnce fst_elt
				in let (exp'',n'') = transformLst (accum@[exp']) (List.tl lst)
				   in (exp'',n''+n')
    and aux newOr lst = 
      let (exp',n') = transformLst [] (List.tl lst)
      in match exp' with
	  MKE.Conc ls -> (MKE.Conc ((MKE.Or newOr)::ls), n'+1)
	| MKE.Or set  -> (MKE.Conc [MKE.Or newOr; MKE.Or set], n'+1)
	| MKE.Bexp MKE.Zero   -> (MKE.Or newOr, n'+1)
	| otherwise   -> (MKE.Conc [MKE.Or newOr; exp'], n'+1)
    and distributeB accum lst_c = 
      if accum=[] then raise (ERROR "function distribute in applyLeftDistrConcOnce received [] as accum")
      else 
	let buildConc' lst elt = MKE.Conc (lst@[MKE.Bexp elt])
	in MKE.setKEfromList (List.map (buildConc' accum) lst_c)
    (* distribute receives a list of exps [e1,e2,..,en], 'accum', and 
       builds the concatenation c = e1e2...en. Then every element di in
       lst_c becomes cdi. The return value is MKE.Or (setFrom lst_c) *)
    and distribute accum lst_c = 
      if accum=[] then raise (ERROR "function distributeB in applyLeftDistrConcOnce received [] as accum")
      else 
	let buildConc' lst elt = 
	  match elt with
	      MKE.Conc ls  ->  MKE.Conc (accum@ls)
	    | otherwise    ->  MKE.Conc (lst@[elt])
	in MKE.setKEfromList (List.map (buildConc' accum) lst_c)
    and transformSet set = 
      if (SetKE.is_empty set) then (set,0) else 
	let fst_elt = SetKE.min_elt set
	in let (exp',n') = applyLeftDistrConcOnce fst_elt
	   in let (exp'',n'') = transformSet (SetKE.remove fst_elt set)
	      in (SetKE.union (SetKE.singleton exp') exp'', n'+n'')
			  
    let rec applyLeftDistrConc exp maxLen = 
      let (exp',n') = applyLeftDistrConcOnce exp in
      if (((MKE.sizeKE exp') > maxLen) || (MKE.katExp2String exp' = MKE.katExp2String exp)) then 
	(exp,0) 
      else 
	let (exp'',n'') = applyLeftDistrConc exp' maxLen
	in (exp'',n''+n')
	
	
    (* main function - applies the axiom functions sequentially *)
    let applyRewriteRules exp minLen maxLen = 
      let (exp0,n0) = applyDenesting exp in
      if ((MKE.sizeKE exp0) > maxLen || (MKE.sizeKE exp0 < minLen)) then 
	(exp, [("denesting",0);("twice star",0);("complements",0);("left distr. of disj.",0);("left distr. of conj.",0);("left distr. of conc.",0)]) 
      else 
	let (exp1,n1) = applyTwiceStar exp0 in
	if ((MKE.sizeKE exp1) > maxLen || (MKE.sizeKE exp1) < minLen) then
	  (exp0,[("denesting",n0);("twice star",0);("complements",0);("left distr. of disj.",0);("left distr. of conj.",0);("left distr. of conc.",0)])
	else 
	  let (exp2,n2) = applyComplements exp1 in
	  if ((MKE.sizeKE exp2) > maxLen || (MKE.sizeKE exp2) < minLen) then
	    (exp1,[("denesting",n0);("twice star",n1);("complements",0);("left distr. of disj.",0);("left distr. of conj.",0);("left distr. of conc.",0)])
	  else 
	    let (exp3,n3) = applyLeftDistrDisj exp2 maxLen in
	    let (exp4,n4) = applyLeftDistrConj exp3 maxLen in
	    let (exp5,n5) = applyLeftDistrConc exp4 maxLen in
	    (exp5, [("denesting",n0);("twice star",n1);("complements",n2);("left distr. of disj.",n3);("left distr. of conj.",n4);("left distr. of conc.",n5)])
	      
  end
