
(* This module works on relations of type 2^((elt set) x (elt set)) between elements of type elt.
   The function ?: given two elements x and y and a relation R, it verifies if x and y
   are related in R up-to-congruence *)

open Module_SetBuilder
open Module_Utils

exception Error of string

let debugging = ref false

let print str = print_endline str; flush stdout
let print_str str = print_string str; flush stdout

module Up2Congr (M: sig type t val fromElt2String: t -> string end):
sig
  module S : Set.S with type elt = M.t and type t = S(M).t
  module R : Set.S with type elt = S.t*S.t and type t = R(M).t
  type elt = M.t
  type eltSet = S.t
  type rel = R.t
  val inCongr: eltSet -> eltSet -> rel -> bool
end
  =
struct
  type elt = M.t
  module S = S (M)    (* builds a set of elts *)
  module R = R (M)
  type eltSet = S.t
  type rel = R.t
      
  (* Type for Forms *)
  type form =                  
      Complete of eltSet     (* in this case, eltSet is a Normal Form *)
    | Incomplete of eltSet*(eltSet list)     

  (* Function Aliases *)
  let lsRemoveRepet = MUtils.lsRemoveRepetitions

  (* Helpful Debugging Printers *)
  let fromEltSet2String eltSet = "{"^ S.fold (fun e a -> (M.fromElt2String e)^a^",") eltSet "" ^"}"

  let fromRel2String rel = 
    "{"^ R.fold (fun (x,y) a -> "("^fromEltSet2String x^","^fromEltSet2String y^")"^a^" ") rel "" ^"}"

  let debugInCongrEntryPoint (x,y,rel) = 
    print ("I've entered inCongr with x="^ fromEltSet2String x ^" y="^ fromEltSet2String y ^" rel="^fromRel2String rel)

  let buildLst (h:eltSet) (rel:rel) : (eltSet list) =      
    (* returns a list of all X s.t. X is related to H in rel *)
    List.append 
      [? List : x | (x,h') <- List : (R.elements rel) ; h'=h (*; not(x=h)*) ?] 
      [? List : x | (h',x) <- List : (R.elements rel) ; h'=h (*; not(x=h)*) ?]

  let expand (form:form) (rel:rel) : (form) = 
    match form with
	Incomplete (set,ls) ->
	  let (hd,tl) = (List.hd ls,List.tl ls) in
	  let newSet = S.union set hd in
	  let newLst = 
	    lsRemoveRepet (tl@[? List:(x:eltSet) | (h1:elt) <- List:S.elements hd; (* ie, h1 is contained in hd *)
		(x:eltSet) <- List: buildLst (S.singleton h1) rel ;   (* ie, x is related to h1 in rel *)
		not(S.subset x set) 
		  ?])
	  in if newLst=[] then (Complete newSet) else (Incomplete (newSet,newLst))
      | Complete _             ->  raise (Error "Up2Congr.expand received a Complete form")
  
  let rec inCongr (x:eltSet) (y:eltSet) (rel:R.t) : bool = 
    if !debugging then debugInCongrEntryPoint (x,y,rel) ;
    let form1 = expand (Incomplete (S.empty, [x])) rel in
    let form2 = expand (Incomplete (S.empty, [y])) rel in
    foo form1 form2 rel
  and foo (form1:form) (form2:form) (rel:rel) = 
    (* OCaml doesn't deal so well with set equality when elt type is rather complex, so a user-defined *)
    (* function to test equality, in replacement of S.equal, may be advisable in such a case *)
    match (form1,form2) with
	(Incomplete (set1,ls1), Incomplete (set2,ls2))  -> 
	  if (S.equal set1 set2) then true 
	  else foo (expand form1 rel) (expand form2 rel) rel
      | (Incomplete (set1,_), Complete set2)  ->  
	if (S.equal set1 set2) then true 
	else foo (expand form1 rel) form2 rel
      | (Complete set1, Incomplete (set2,_))  ->  
	if !debugging then 
	  print ("foo has received 1 normal form:"^fromEltSet2String set1^" and an incomplete (set:"^fromEltSet2String set2^")") ;
	if (S.equal set1 set2) then true 
	else foo form1 (expand form2 rel) rel
      | (Complete set1, Complete set2)  ->  
	let answer = S.equal set1 set2 in
	if !debugging then print ("foo has received 2 normal forms:"^fromEltSet2String set1^" and "^fromEltSet2String set2) ; answer

end
