
open Module_KatExpressions
open Module_KatExpressionsPos
open Module_KatExpressionsPos2
open Lexer
open Parser
open LexerPosition
open ParserPosition
open LexerPolishPosition
open Module_SetBuilder

exception ERROR of string

module MUtils : sig
  val lsDiff : 'a list -> 'a list -> 'a list ;;
  val lsDiff2 : 'a list -> 'a list -> 'a list ;;
  val lsFindIndex : 'a -> 'a list -> int ;;
  val lsIntPrint : int list -> unit ;;
  val lsRemove : 'a -> 'a list -> 'a list ;;
  val lsRemoveIndex : int -> 'a list -> 'a list ;;
  val lsRemoveIndexes : int list -> 'a list -> 'a list ;;
  val lsRemoveRepetitions : 'a list -> 'a list ;;
  val parse : string -> MKE.ke ;;
  val parsePolish : string -> MKE.ke ;;
  val parseBexp : string -> MKE.be ;;
  val parsePosition : string -> MKEPos.ke ;;
  val parsePolishPosition : string -> MKEPos.ke ;;
  val powersetLists : 'a list -> 'a list list ;;
  val powersetListsWithoutEmpty : 'a list -> 'a list list ;;
  val powersetListsKeepIndex : 'a list -> (int list * 'a list) list ;;
  val powersetListsKeepIndexWithoutEmpty : 'a list -> (int list * 'a list) list ;;
  val printHInt : (SetInt.t*SetInt.t) list -> unit ;;
  val printListPairsStrings : (string*string) list -> unit ;;
  val shuffleList: 'a list -> unit -> 'a list ;;
  val subArray: 'a array -> int -> 'a array ;;
  val tryListHd: 'a list -> string -> 'a ;;
end
  = struct

    (* Keeps in ls1 all elements which do not appear in ls2 *)
    let lsDiff ls1 ls2 = List.filter (fun x -> not(List.mem x ls2)) ls1

    (* Returns all elements which are in one of the lists but not in the other *)
    let lsDiff2 ls1 ls2 = (lsDiff ls1 ls2)@(lsDiff ls2 ls1)

    let lsFindIndex a ls = 
      let rec loop n a ls = match ls with
	  h::tl  ->  if h=a then n else loop (n+1) a tl
	| []     ->  raise (ERROR "MUtils.lsFindIndex attempted to find the index of an element which doesn't exist in the list")
      in loop 0 a ls

    let lsIntPrint ls = print_string "[" ; List.iter (fun x -> print_string ((string_of_int x) ^ ";")) ls ; print_string "] "

    let lsRemove a ls = 
      if List.mem a ls then
	List.filter (fun x -> not(x=a)) ls
      else raise (ERROR "MUtils.lsRemove attempted to remove an element from a list where it doesn't exist")

    let lsRemoveIndex index ls = 
      let rec loop n index ls = 
	match ls with
	  (h::tl)  ->  if n=index then tl else h::(loop (n+1) index tl)
	| []       ->  []
      in loop 0 index ls

    let rec lsRemoveRepetitions ls = 
      match ls with
	  h::tl  ->  if List.mem h tl then (lsRemoveRepetitions tl) else h::(lsRemoveRepetitions tl)
	| []     ->  []
    
    let lsRemoveIndexes indexes ls = 
      let rec loop n indexes ls = 
	match (indexes,ls) with
	  (i::is,h::tl)  ->  if n=i then loop (n+1) is tl else h::(loop (n+1) (i::is) tl)
	| ([]   ,h::tl)  ->  h::tl
	| (i::is,[])     ->  raise (ERROR "MUtils.lsRemoveIndexes attempted to remove index non-existent in a list")
	| ([]   ,[])     ->  []
      in loop 0 (List.fast_sort compare (lsRemoveRepetitions indexes)) ls

    let parse str =
      let linebuf = Lexing.from_string (str ^ "\nEOF") in
      List.hd (Parser.data Lexer.token linebuf)
	
    let parsePolish str =
      let linebuf = Lexing.from_string (str ^ "\nEOF") in
      List.hd (ParserPolish.data LexerPolish.token linebuf)

    let parsePosition str =
      let linebuf = Lexing.from_string (str ^ "\nEOF") in
      List.hd (ParserPosition.data LexerPosition.token linebuf)

    let parsePolishPosition str =
      let linebuf = Lexing.from_string (str ^ "\nEOF") in
      List.hd (ParserPolishPosition.data LexerPolishPosition.token linebuf)

    let parseBexp str = MKE.convertKe2Be (parse str)

    let rec powersetLists ls = 
      match ls with
	[]    -> [[]]
      | h::tl -> List.fold_left (fun xs t -> (h::t)::t::xs) [] (powersetLists tl);;

    let powersetListsWithoutEmpty ls = lsRemove [] (powersetLists ls)

    (* like powersetLists, but keeps the index of the elements of ls used
       in each subset; its output is thus not of type [[x;y;z];[x;z]] but 
       of type [([0;1;2],[x;y;z]); ([0;2],[x;z])] for ls=[x;y;z] *)
    let powersetListsKeepIndex ls =
      let rec loop lst =
	match lst with
	  []    -> [([],[])]
	| h::tl -> List.fold_left (fun xs (i,t) -> ((lsFindIndex h ls)::i,h::t)::(i,t)::xs) [] (loop tl)
      in loop (lsRemoveRepetitions ls)

    let powersetListsKeepIndexWithoutEmpty ls = lsRemove ([],[]) (powersetListsKeepIndex ls)

    let printHInt ls = List.iter (fun (x,y) -> print_string "({"; MInt.printSetInt x; print_string "},{"; MInt.printSetInt y ; print_string "}), ") ls
      
    let printListPairsStrings ls = List.iter (fun (a,b) -> print_string (a^" & " ); print_endline b) ls

    let tryListHd ls (funcName:string) = 
      try 
	List.hd ls
      with 
	Failure "hd" -> raise (ERROR ("I tried to take the head of an empty list in function "^funcName))

    let shuffleList d () =
      let nd = List.map (fun c -> (Random.bits (), c)) d in
      let sond = List.sort compare nd in
      let d = List.map snd sond in
      let nd = List.map (fun c -> (Random.bits (), c)) d in
      let sond = List.sort compare nd in
      List.map snd sond

    (* Returns an array with the elements from i till the end of arr *)
    let subArray arr i = Array.sub arr i ((Array.length arr)-i)
	
  end
